@@ -380,6 +380,25 @@ flash_types[] = {
 		},
 		flags: FLASH_IS_CFI
 	},	
+	{
+		name:	"AM29LV128ML",
+		size:	    16*1024*1024,
+		min_sectorsize: 64*1024,	
+		querystr: { 'Q','R','Y', 0x02,0,0x40,0,0,0,0,0,0x27,0x36,0,0, 7, 
+			7,0xa,0,1,5,4,0,0x18, 2,0,5,0,1, 0xff,0,0,
+			1, 0,0,0,0, 0,0,0,0, 0,0,0,0,  0,0,0,
+			'P','R','I', 0x31,0x33,8,2,1,1,4,0,0,1,0xb5,0xc5,4
+			,1
+		},
+		auto_sel: { 
+			{0x00,0x4f,0x0001},	/* Manufacturer ID */
+			{0x01,0x4f,0x227e},	/* Device ID */
+			{0x03,0x4f,0x0008},	/* SecSi lowest not locked */
+			{0x0e,0x4f,0x2212},	/* Ex Device ID */
+			{0x0f,0x4f,0x2200},	/* Ex2 Device ID */
+		},
+		flags: FLASH_IS_CFI
+	},	
 	{
 		name:	"AM29LV256ML",
 		size:	    32*1024*1024,
@@ -720,7 +739,12 @@ second_cycle_write(AMD_Flash *flash,uint16_t value,uint32_t dev_addr) {
 		if((reg==0x2aa) && (value==0x55)) {
 			flash->cycle++;
 			return;
-		}  
+		} else if(((reg & 0xff) ==  0) && (value == 0x70)) {
+			/* ignore sector unlocks */
+			flash->cycle = CYCLE_FIRST;
+			flash->cmd=CMD_NONE;
+			return;
+		}
 		fprintf(stderr,"Unknown: value %08x to reg %08x in second cycle\n",value,reg);
 		return;
 	} else if(flash->cmd==CMD_UNLOCK_BYPASS_PGM) {
@@ -1543,7 +1567,7 @@ AMDFlashBank_New(const char *flash_name) {
 		fprintf(stderr,"Flash type %s not found. Available types:\n",flash_type);
 		for(i=0;i<nr_types;i++) {
 			ftype=&flash_types[i];	
-			fprintf(stderr,"\t%s\n", ftype->name);
+			fprintf(stderr," %-15s - size %dk\n", ftype->name,ftype->size/1024);
 		}
 		exit(1734);
 	}
