@@ -36,6 +36,7 @@
 #include <arm9cpu.h>
 #include <instructions.h>
 #include <idecode.h>
+#include <thumb_decode.h>
 #include <bus.h>
 #include <mmu.h>
 #include <fio.h>
@@ -53,6 +54,8 @@
 #define dbgprintf(x...)
 #endif
 
+ARM9 gcpu;
+
 uint32_t debugflags=0;
 //uint32_t debugflags = DEBUG_LOG_PC;
 uint32_t mmu_high_vectors=0;
@@ -113,6 +116,33 @@ gdb_setreg(void *clientData,const uint8_t *data,int index,int len)
 	return; 
 }
 
+/*
+ * ---------------------------------------------------------------------
+ * get_bkpt_ins
+ * 	returns the operation code of the breakpoint instruction.
+ *	Needed by the debugger to insert breakpoints
+ * ---------------------------------------------------------------------
+ */
+static void 
+gdb_get_bkpt_ins(void *clientData,uint8_t *ins,uint64_t addr,int len) 
+{
+	if(len == 2) {
+		uint16_t value = 0xbe00;
+		if(MMU_Byteorder() == TARGET_BYTEORDER) {
+			*((uint16_t*)ins) = host16_to_target(value);
+		} else {
+			*((uint16_t*)ins) = swap16(host16_to_target(value));
+		}
+	} else if(len == 4) {
+		uint32_t value = 0xe1200070;
+		if(MMU_Byteorder() == TARGET_BYTEORDER) {
+			*((uint32_t*)ins) = host32_to_target(value);
+		} else {
+			*((uint32_t*)ins) = swap32(host32_to_target(value));
+		}
+	}
+}
+
 static int 
 gdb_stop(void *clientData) 
 {
@@ -500,8 +530,6 @@ hello_proc(void *cd) {
 //	CycleTimer_Add(&htimer,10000000000LL,hello_proc,NULL);
 }
 
-ARM9 gcpu;
-
 static int
 irq_change(SigNode *node,int value,void *clientData)
 {
@@ -562,6 +590,7 @@ ARM9_New() {
 	arm->gdbops.getmem = gdb_getmem;
 	arm->gdbops.setmem = gdb_setmem;
 	arm->gdbops.step = gdb_step;
+	arm->gdbops.get_bkpt_ins = gdb_get_bkpt_ins;
 	arm->gserv = GdbServer_New(&arm->gdbops,arm);
 	gcpu.signal_mask |= ARM_SIG_RESTART_LOOP;
 	return arm;
@@ -611,6 +640,13 @@ ARM9_Exception(int exception,int pc_offset) {
 	uint32_t old_cpsr=REG_CPSR;
 	int new_mode = EX_TO_MODE(exception);
 	uint32_t new_pc = EX_TO_PC(exception);
+
+#if 0
+	fprintf(stderr,"EXCEPTION %08x at %08x\n",exception,GET_REG_PC-8+pc_offset);
+	exit(1);
+	//ARM_Break();
+#endif
+
 	if(mmu_high_vectors) {
 		new_pc|=0xffff0000;
 	}
@@ -627,6 +663,7 @@ ARM9_Exception(int exception,int pc_offset) {
 	SET_REG_PC(new_pc);
 }
 
+
 /*
  * -----------------------------------------------------------------------
  * Check Signals
@@ -650,7 +687,7 @@ CheckSignals() {
 			}
 			if(unlikely(gcpu.signals & ARM_SIG_RESTART_LOOP)) {
 				if(likely(gcpu.dbg_mode== DBG_MODE_OFF)) { 
-					longjmp(gcpu.restart_loop_jump,1);
+					ARM_RestartLoop();
 				}  else if(gcpu.dbg_mode == DBG_MODE_STEP) {
 					gcpu.dbg_steps--;
 					if(gcpu.dbg_steps==0) {
@@ -658,7 +695,7 @@ CheckSignals() {
 						if(gcpu.gserv) {
 							GdbServer_Signal(gcpu.gserv,SIGINT);
 						}
-						longjmp(gcpu.restart_loop_jump,1);
+						ARM_RestartLoop();
 					} else {
 						mainloop_event_pending = 1;
 					}
@@ -667,14 +704,20 @@ CheckSignals() {
 					if(gcpu.gserv) {
 						GdbServer_Signal(gcpu.gserv,SIGINT);
 					}
-					longjmp(gcpu.restart_loop_jump,1);
+					ARM_RestartLoop();
 				} else if(gcpu.dbg_mode == DBG_MODE_BREAK) {
 					if(gcpu.gserv) {
 						if(GdbServer_Signal(gcpu.gserv,SIGINT)>0) {
 							gcpu.dbg_mode = DBG_MODE_STOPPED;
-							longjmp(gcpu.restart_loop_jump,1);
+							ARM_RestartLoop();
+#if 1
+						}   else {
+						// jk
+							gcpu.dbg_mode = DBG_MODE_STOPPED;
+							ARM_RestartLoop();
+#endif
 						}
-					} 
+					}
 					ARM9_Exception(EX_PABT,8);
 					gcpu.dbg_mode = DBG_MODE_OFF;
 				} else {
@@ -769,6 +812,21 @@ log_uint32(uint32_t data) {
 #endif
 }
 
+static void
+Thumb_Loop() {
+	ThumbInstructionProc *iproc;
+	fprintf(stderr,"Entering Thumb loop\n");
+	setjmp(gcpu.abort_jump);
+	while(1) {
+		ICODE=MMU_IFetch16(REG_PC);
+		REG_PC+=2;
+		iproc=ThumbInstructionProc_Find(ICODE);
+		iproc();
+		CycleCounter+=2;
+		CycleTimers_Check();
+	}
+}
+
 /*
  * ---------------------------------------------
  * The main loop for 32Bit instruction set
@@ -811,10 +869,19 @@ ARM9_Loop32() {
 void
 ARM9_Run() {
 	uint32_t addr=0;
+	uint32_t gdbwait;
 	if(Config_ReadUInt32(&addr,"global","start_address")<0) {
 		addr=0;
 	}
-	fprintf(stderr,"Starting CPU at %08x\n",addr);
+	if(Config_ReadUInt32(&gdbwait,"global","gdbwait")<0) {
+		gdbwait=0;
+	}
+	if(gdbwait) {
+		fprintf(stderr,"CPU is waiting for gdb connection at %08x\n",addr);
+		gcpu.dbg_mode = DBG_MODE_STOPPED;
+	} else {
+		fprintf(stderr,"Starting CPU at %08x\n",addr);
+	}
 	gettimeofday(&gcpu.starttime,NULL);
 	REG_PC=addr;
 	/* A long jump to this label redecides which main loop is used  */
@@ -828,7 +895,7 @@ ARM9_Run() {
                         FIO_WaitEventTimeout(&tout);
 		} else {
 			if(REG_CPSR & FLAG_T) {
-				fprintf(stderr,"Thumb not implemented\n");
+				Thumb_Loop();
 			} else {
 				ARM9_Loop32(); 
 			}
