@@ -35,6 +35,7 @@
 #include <fcntl.h>
 #include <bus.h>
 
+Bus *MainBus;
 /*
  * ------------------------------------
  * Hash for single IO-Ports
@@ -110,7 +111,7 @@ twolevel_translate_w(uint32_t addr)
 	index = (addr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
 	base = slvl_map[index];
 	if(likely(base)) {
-		if(unlikely(((int)base) & PG_TRACED)) {
+		if(unlikely(((unsigned long)base) & PG_TRACED)) {
 			base -= PG_TRACED;
 			slvl_map[index] = base;
 			IO_Write8(0,addr); 
@@ -263,7 +264,7 @@ Bus_Write8(uint8_t value,uint32_t addr) {
  */
 
 void
-Bus_Write(uint32_t addr,uint8_t *buf,int count) 
+Bus_Write(uint32_t addr,uint8_t *buf,uint32_t count) 
 {
 	while(count)  {
 		Bus_Write8(*buf++,addr++);
@@ -279,8 +280,9 @@ Bus_WriteSwap32(uint32_t addr,uint8_t *buf,int count)
 		count--;
 	}
 }
+
 void
-Bus_Read(uint8_t *buf,uint32_t addr,int count) 
+Bus_Read(uint8_t *buf,uint32_t addr,uint32_t count) 
 {
 	while(count)  {
 		*buf=Bus_Read8(addr++);
@@ -509,8 +511,8 @@ Mem_SplitLargePage(uint32_t pgaddr) {
 		flags |= MEM_FLAG_WRITABLE;
 		hva = whva;
 	}
-	if(Mem_UnMapBlock(addr) > 0) {
-		fprintf(stderr,"Unmapped at %08x successfully\n",addr);	
+	if(!Mem_UnMapBlock(addr)) {
+		fprintf(stderr,"Failed to unmap block at %08x\n",addr);	
 	}
 	for(count=0;count < MEM_MAP_BLOCKSIZE;count+=slvl_size) {
 		Mem_Map2LvlBlock(addr+count,hva+count,flags); 
@@ -531,28 +533,62 @@ Mem_TracePage(uint32_t pgaddr)
 	}
 	index = (pgaddr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
 	hva = slvl_map[index];
-	if(hva && !((int)hva & PG_TRACED)) {
+	if(hva && !((unsigned long)hva & PG_TRACED)) {
 		
 		slvl_map[index] += PG_TRACED;
-	//	fprintf(stderr,"make page at %08x traced\n",pgaddr); // jk
 	} else {
-		fprintf(stderr,"can not trace %08x hva %p %08x\n",pgaddr,hva,!((int)hva & PG_TRACED));
+		fprintf(stderr,"Page is already traced %08x hva %p %08x\n",pgaddr,hva,!((unsigned long)hva & PG_TRACED));
 	}
 	if(InvalidateProc) {
 		InvalidateProc();
 	}
 }
 
+void
+Mem_UntracePage(uint32_t pgaddr) 
+{
+	int index;
+	uint8_t *hva;
+	uint8_t **slvl_map;
+	index = pgaddr >> twoLevelMMap.frst_lvl_shift; 
+	if(unlikely(!(slvl_map = twoLevelMMap.flvlmap_write[index]))) {
+		fprintf(stderr,"no slvl map for addr %08x\n",pgaddr); // jk
+		return;
+	}
+	index = (pgaddr & twoLevelMMap.scnd_lvl_mask) >> twoLevelMMap.scnd_lvl_shift;
+	hva = slvl_map[index];
+	if(hva && ((unsigned long)hva & PG_TRACED)) {
+		slvl_map[index] -= PG_TRACED;
+		//fprintf(stderr,"Untraced page at %08x, hva %p\n",pgaddr,slvl_map[index]);
+	} else {
+		fprintf(stderr,"Page was not traced %08x, hva %p\n",pgaddr,slvl_map[index]);
+	}
+	if(InvalidateProc) {
+		InvalidateProc();
+	}
+}
+
+
 void 
 Mem_TraceRegion(uint32_t start,uint32_t length) 
 {
-	int i;
+	uint32_t i;
 	int slvl_size =  twoLevelMMap.scnd_lvl_blocksize;
 	fprintf(stderr,"Trace region from %08x length %08x\n",start,length); // jk
 	for(i=0;i<length;i+=slvl_size) {
 		Mem_TracePage(start+i);
 	} 
 }
+void 
+Mem_UntraceRegion(uint32_t start,uint32_t length) 
+{
+	uint32_t i;
+	int slvl_size =  twoLevelMMap.scnd_lvl_blocksize;
+	fprintf(stderr,"Untrace region from %08x length %08x\n",start,length); // jk
+	for(i=0;i<length;i+=slvl_size) {
+		Mem_UntracePage(start+i);
+	} 
+}
 
 /*
  * --------------------------------------------------------------------
@@ -1173,6 +1209,17 @@ ckalloc(size_t size)
 	memset(mem,0,size);
 	return mem;
 }
+
+static struct Bus mainBus = {
+	.read32 = Bus_Read32,
+        .read16 = Bus_Read16,
+       	.read8 = Bus_Read8,
+        .readblock = Bus_Read,
+	.write32 = Bus_Write32,
+	.write16 = Bus_Write16,
+	.write8 = Bus_Write8,
+	.writeblock = Bus_Write
+};
 /*
  * --------------------------------------------------------------
  * Create and Initialize memory and IO-Handler maps and Hashes
@@ -1210,5 +1257,6 @@ Bus_Init(InvalidateCallback *invalidate,uint32_t min_memblocksize)
 
 	iohandlerMap=ckalloc(sizeof(IOHandler *)*IOH_MAP_ENTRIES);
 	iohandlerFlvlMap=ckalloc(sizeof(IOHandler **)*IOH_FLVL_SZ);
+	MainBus = &mainBus; 
 	fprintf(stderr,"MemMap and IO-Handler Hash initialized\n");
 }
