@@ -4,8 +4,8 @@
 
 typedef struct ArmCoprocessor {
         void *owner;
-        uint32_t (*mrc)(struct ArmCoprocessor *,int opcode_1,int crn,int crm,int opcode_2);
-        void (*mcr)(struct ArmCoprocessor *,int opcode_1,uint32_t value,int crn,int crm,int opcode_2);
+        uint32_t (*mrc)(struct ArmCoprocessor *,uint32_t icode);
+        void (*mcr)(struct ArmCoprocessor *,uint32_t icode,uint32_t value);
         void (*cdp)(struct ArmCoprocessor *,uint32_t icode);
         void (*ldc)(struct ArmCoprocessor *,uint32_t icode);
         void (*stc)(struct ArmCoprocessor *,uint32_t icode);
