@@ -55,6 +55,7 @@
 #define		NCR_FDX		(1<<3)
 #define		NCR_LBK_MASK	(3<<1)
 #define		NCR_LBK_SHIFT	(1)
+#define 	NCR_RST		(1<<0)
 #define DM_NSR 		(1)
 #define 	NSR_SPEED 	(1<<7)
 #define 	NSR_LINKST 	(1<<6)
@@ -172,6 +173,17 @@
 #define		SMCR_FB1	(1<<1)	
 #define		SMCR_FB0	(1<<0)	
 
+/* Only in newer revisions of the chip */
+#define DM_ETXCSR	(0x30)
+#define DM_TCSCR	(0x31)
+#define	DM_RCSCSR	(0x32)
+#define DM_MPAR		(0x33)
+#define DM_LEDCR	(0x34)
+#define	DM_BUSCR	(0x38)
+#define	DM_INTCR	(0x39)
+#define DM_SCCR		(0x50)
+#define DM_RSCCR	(0x51)
+
 #define DM_MRCMDX	(0xf0)
 #define DM_MRCMD	(0xf2)
 #define DM_MRRL		(0xf4)
@@ -221,12 +233,15 @@ typedef void DMWriteProc(DM9000 *dm,uint32_t value,uint8_t index,int rqlen);
 
 struct DM9000 {
 	int register_spacing; /* (bytes) for io-register spacing */
-	int buswidth;
+	int buswidth;	/* unit is bytes */
 	M93C46 *eprom;
 	int eeprom_valid;
 	uint8_t randmac[6]; /* random mac for case of invalid eeprom */
 	PHY_Device *extphy[MAX_PHYS];
 	PHY_Device iphy;
+	/* The signal levels for the interrupts */	
+	int int_active;
+	int int_inactive;
 	SigNode *irqNode;
 	int interrupt_posted;
 	BusDevice bdev;
@@ -237,6 +252,7 @@ struct DM9000 {
 	DMReadProc *read_reg[256];
 	DMWriteProc *write_reg[256];
 	uint8_t index_reg;
+	uint64_t index_chg_cnt; /* some shity registers do not update when index is not changed */
 	CycleTimer epcr_busyTimer; /* eeprom or Phy access in progress timer */
 
 	int tx_pktindex;
@@ -278,7 +294,20 @@ struct DM9000 {
 	uint8_t pidh;
 	uint8_t chipr;
 	uint8_t smcr;
-	uint8_t mrcmdx;
+
+	/* only in new revision */
+	uint8_t etxcsr;
+	uint8_t tcscr;
+	uint8_t rcscsr;
+	uint8_t mpar;
+	uint8_t ledcr;
+	uint8_t buscr;
+	uint8_t intcr;
+	uint8_t sccr;
+	uint8_t rsccr;
+
+	uint32_t mrcmd_latch; /* Maximum size for 32Bit bus */
+	uint64_t mrcmdx_idx_tag;
 	uint8_t mrcmd;
 	uint8_t mwcmdx;
 	uint8_t mwcmd;
@@ -349,12 +378,12 @@ update_interrupts(DM9000 *dm)
 	/* Only lower 4 Bits are interupt bits */
 	if((dm->isr & dm->imr) & 0xf) {
 		if(!dm->interrupt_posted) {
-			SigNode_Set(dm->irqNode,SIG_LOW);
+			SigNode_Set(dm->irqNode,dm->int_active);
 			dm->interrupt_posted=1;
 		}
 	} else {
 		if(dm->interrupt_posted) {
-			SigNode_Set(dm->irqNode,SIG_HIGH); // ? SIG_OPEN or SIG_PULLUP ?
+			SigNode_Set(dm->irqNode,dm->int_inactive); // ? SIG_OPEN or SIG_PULLUP ?
 			dm->interrupt_posted=0;
 		}
 	}
@@ -589,11 +618,11 @@ soft_reset(DM9000 *dm)
 	dm->rcr = 0;
 	dm->rsr = 0;	
 	dm->rocr = 0;
-	dm->bptr = 0x37;
-	dm->fctr = 0x38;
+	dm->bptr = 0x37; /* Register 8 */
+	dm->fctr = 0x38; /* Register 9 */
 	dm->fcr = 0;	
 	dm->epcr = 0;
-	dm->epar = 0x40;
+	dm->epar = 0x40; /* Register 0xc */
 	/* dm->epdrl */
 	/* dm->epdrh */
 	dm->wcr = 0;
@@ -610,10 +639,10 @@ soft_reset(DM9000 *dm)
 	dm->pidh = 0x90;
 	dm->chipr = 0;
 	dm->smcr = 0;
-	dm->rxfifo_rp = 0xc00;
 	dm->rxfifo_count = 0;
-	dm->rxfifo_wp=  0xc00; 
-	dm->txfifo_wp = 0;
+	dm->rxfifo_rp = dm->rxfifo_wp=  0xc00; 
+	dm->txfifo_rp = dm->txfifo_wp = 0;
+	dm->sram[dm->rxfifo_rp] = 0; /* Is this set zeroed on RXEN ? jk */
 	switch(dm->buswidth) {
 		case 1:
 			dm->isr = 0x80;
@@ -708,6 +737,8 @@ static int
 reload_eeprom(DM9000 *dm)  {
 	uint8_t data[128];
 	uint8_t *mac = data;
+	uint16_t pin_control;
+	uint16_t auto_load_control;
 	int i;
 	int result = 0;
 	for(i=0;i<128;i++) {
@@ -734,6 +765,43 @@ reload_eeprom(DM9000 *dm)  {
 			dm->par[i] = dm->randmac[i]; 
 		}
 	}
+	auto_load_control = data[6] + (data[7]<<8);
+	if((auto_load_control & 0xc) == 4) { 
+		pin_control = data[12] + (data[13]<<8);
+		switch(pin_control & 0x18) {
+			case 0:
+				dm->int_active = SIG_HIGH;
+				dm->int_inactive = SIG_LOW;
+				dm->intcr = 0;
+				break;
+			case 8:
+				dm->int_active = SIG_LOW;
+				dm->int_inactive = SIG_HIGH;
+				dm->intcr = 1;
+				break;
+
+			case 0x10:
+				dm->int_active = SIG_OPEN;
+				dm->int_inactive = SIG_LOW;
+				dm->intcr = 2;
+				break;
+
+			case 0x18:
+				dm->int_active = SIG_LOW;			
+				dm->int_inactive = SIG_OPEN;
+				dm->intcr = 3;
+				break;
+		} 
+	} else {
+		dm->int_active = SIG_HIGH;
+		dm->int_inactive = SIG_LOW;
+		dm->intcr = 0;
+	}
+	if(dm->interrupt_posted) {
+		SigNode_Set(dm->irqNode,dm->int_active);
+	} else {
+		SigNode_Set(dm->irqNode,dm->int_inactive);
+	}	
 	return result;
 }
 
@@ -747,6 +815,12 @@ static void
 ncr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
 	dm->ncr = value &  ~1;
+#if 1
+	if(value & NCR_RST) {
+		// should auto clear after 10 uS
+		soft_reset(dm); // doesn't work anymore with this !
+	}	
+#endif
 	if(value & NCR_EXT_PHY) {
 		fprintf(stderr,"DM9000 ncr: External Phy not supported\n");
 	}
@@ -760,7 +834,10 @@ ncr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 static uint32_t
 nsr_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
-        return dm->nsr;
+	uint8_t clearbits = NSR_WAKEST | NSR_TX2_END | NSR_TX1_END;	
+	uint8_t retval = dm->nsr;
+	dm->nsr &= ~clearbits;
+        return retval;
 }
 
 static void
@@ -820,6 +897,7 @@ rcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 {
 	uint32_t diff = dm->rcr ^ value;
 	if(diff & value & RCR_RXEN) {
+		/* Is sram[wp] zeroed on RXEN ? */
 		clear_pktsrc(dm);
 		enable_receiver(dm);
 	} else {
@@ -875,6 +953,13 @@ bptr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 	dm->bptr = value;	
 }
 
+/*
+ * -------------------------------------------------------------------------
+ * Flow control Threshold register
+ * Bits 0-3 LWOT: Watermark in kbytes when a Pause 0 packet is sent
+ * Bits 4-7 HWOT: Watermark in kbytes when a pause 0xffff packet is sent
+ * -------------------------------------------------------------------------
+ */
 static uint32_t
 fctr_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
@@ -887,6 +972,19 @@ fctr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 	dm->fctr = value;
 }
 
+/*
+ * ------------------------------------------------------------------------
+ * FCR Flow Control register
+ * Bit 7: TXP0: Send a pause packet with 0x0000 (no pause)
+ * Bit 6: TXPF: Send a pause packet with 0xffff
+ * Bit 5: TXPEN: Force Enable transmitting of pause packets
+ * Bit 4: BKPA: Backpressure mode for all packets (JAM, Half Duplex)
+ * Bit 3: BKPM: Backpressure mode for matching packets (JAM, Half Duplex)
+ * Bit 2: RXPS: Rx pause packet status latched
+ * Bit 1: RXPCS: Pause packet current status of receiver (1=pause)
+ * Bit 0: FLCE Enable Flow control (listen on pause packets)
+ * ------------------------------------------------------------------------
+ */
 static uint32_t
 fcr_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
@@ -898,7 +996,7 @@ fcr_read(DM9000 *dm,uint8_t index,int rqlen)
 static void
 fcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
-	dm->fcr = (dm->fcr & (FCR_RXPS | FCR_RXPCS)) | (value & ~(FCR_RXPS | FCR_RXPCS));
+	dm->fcr = (dm->fcr & (FCR_RXPS | FCR_RXPCS)) | (value & ~(FCR_RXPS | FCR_RXPCS | FCR_TXP0 | FCR_TXPF));
 }
 
 static uint32_t
@@ -927,6 +1025,9 @@ epcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 			dprintf("DM9000 EPCR: Phy not found\n");
 			return;
 		}
+		/* This is not true in real device ! On
+		 * write ERRE is cleared before it is ready 
+		 */	
 		if(dm->epcr & EPCR_ERRE) {
 			return;
 		}
@@ -961,9 +1062,13 @@ epcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 			CycleTimer_Mod(&dm->epcr_busyTimer,MicrosecondsToCycles(20));
 		} else if (diff & value & EPCR_ERPRW) {
 			uint16_t val = dm->epdrl | (dm->epdrh<<8);
-			m93c46_writew(dm->eprom,val,addr);
-			dm->epcr |= EPCR_ERRE;
-			CycleTimer_Mod(&dm->epcr_busyTimer,MicrosecondsToCycles(20));
+			if(value & EPCR_WEP) {
+				m93c46_writew(dm->eprom,val,addr);
+				dm->epcr |= EPCR_ERRE;
+				CycleTimer_Mod(&dm->epcr_busyTimer,MicrosecondsToCycles(20));
+			} else {
+				fprintf(stderr,"DM9000 emu: eeprom is writeprotected\n");
+			}
 		}
 	}
 	if(diff & value & EPCR_REEP) {
@@ -997,18 +1102,13 @@ epar_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 static uint32_t
 epdrl_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
-	return dm->epdrl | (dm->epdrh << 8);
+	return dm->epdrl;
 }
 
 static void
 epdrl_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
-	if(rqlen==1) {
-		dm->epdrl = value;
-	} else {
-		dm->epdrl = value & 0xff;
-		dm->epdrh = (value >> 8) & 0xff;
-	}
+	dm->epdrl = value;
 }
 
 static uint32_t
@@ -1052,18 +1152,12 @@ static uint32_t
 par_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
 	unsigned int n = index - 0x10;	
-	int i;
 	uint32_t value=0;
-	if((n+rqlen)>6) {
+	if(n>6) {
 		fprintf(stderr,"DM9000 emulator bug in par_read\n");
 		return 0;
 	}
-	for(i=0;i<rqlen;i++) {
-		value |= dm->par[n+i] << (i*8);
-	}
-	if(rqlen>1) {
-		fprintf(stderr,"DM9000 PAR byteorder unknown\n");
-	}
+	value |= dm->par[n];
 	return value;
 }
 
@@ -1071,18 +1165,13 @@ static void
 par_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
 	unsigned int n = index - 0x10;	
-	int i;
 	dprintf("PAR write 0x%02x to n %d,rqlen %d\n",value,n,rqlen);
-	if((n+rqlen)>6) {
+	if(n>6) {
 		fprintf(stderr,"DM9000 emulator bug in par_write\n");
 		return;
 	}
-	if(rqlen>1) {
-		fprintf(stderr,"DM9000 wide par write byteorder\n");
-	}
-	for(i=0;i<rqlen;i++) {
-		dm->par[n+i] = value >> (i*8);
-	}
+	/* fprintf(stderr,"Write PAR %d val 0x%02x\n",index,value); */
+	dm->par[n] = value;
 }
 
 /*
@@ -1095,18 +1184,12 @@ static uint32_t
 mar_read(DM9000 *dm,uint8_t index,int rqlen) 
 {
 	unsigned int n = index - 0x16;	
-	int i;
-	uint32_t value = 0;
-	if(rqlen>1) {
-		fprintf(stderr,"DM9000 wide mar read not implemented\n");
-	}
-	if(n+rqlen>8) {
+	uint32_t value;
+	if(n>8) {
 		fprintf(stderr,"DM9000 emulator bug in mar_read, index %d\n",index);
 		return 0;
 	}
-	for(i=0;i<rqlen;i++) {
-		value |= dm->mar[n+i] << (i*8);
-	}
+	value = dm->mar[n];
         return value;
 }
 
@@ -1114,17 +1197,11 @@ static void
 mar_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
 	unsigned int n = index - 0x16;	
-	int i;
-	if(rqlen>1) {
-		fprintf(stderr,"DM9000 wide par write not implemented\n");
-	}
-	if(n+rqlen>8) {
+	if(n>8) {
 		fprintf(stderr,"DM9000 emulator bug in mar_write, index %d\n",index);
 		return;
 	}
-	for(i=0;i<rqlen;i++) {
-		dm->mar[n+i] = value >> (i*8);
-	}
+	dm->mar[n] = value;
 }
 
 static uint32_t
@@ -1268,21 +1345,196 @@ smcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
         dprintf("Not implemented 0x%02x\n",index);
 }
 
+/*
+ * -------------------------------------------
+ * Early transmit control/status register
+ * -------------------------------------------
+ */
+static uint32_t
+etxcsr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+etxcsr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+	dm->etxcsr = value & 0x83;
+        fprintf(stderr,"DM9000: ETXCSR has no effect\n");
+}
+
+/*
+ * ---------------------------------------
+ * Check sum control register
+ * ---------------------------------------
+ */
+static uint32_t
+tcscr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+tcscr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Checksum control register not implemented\n",index);
+}
+
+/*
+ * -------------------------------------------------------------
+ * Receive Check sum status Register
+ * -------------------------------------------------------------
+ */
+static uint32_t
+rcscsr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+rcscsr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Not implemented 0x%02x\n",index);
+}
+/*
+ * ----------------------------------------------------------
+ * MII PHY Adress Register
+ * ----------------------------------------------------------
+ */
+static uint32_t
+mpar_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+mpar_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Not implemented 0x%02x\n",index);
+}
+
+/*
+ * ---------------------------------------------------------
+ * Led Pin control register. Switch between GPIO mode and
+ * SMII mode 
+ * ---------------------------------------------------------
+ */
+static uint32_t
+ledcr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+ledcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Not implemented 0x%02x\n",index);
+}
+static uint32_t
+buscr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        return dm->buscr;
+}
+static void
+buscr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+	dm->buscr = value;
+}
+static uint32_t
+intcr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        return dm->intcr;
+}
+static void
+intcr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+	
+	switch(value & 3) {
+		case 0:
+			dm->int_active = SIG_HIGH;
+			dm->int_inactive = SIG_LOW;
+			break;
+		case 1:
+			dm->int_active = SIG_LOW;
+			dm->int_inactive = SIG_HIGH;
+			break;
+
+		case 2:
+			dm->int_active = SIG_OPEN;
+			dm->int_inactive = SIG_LOW;
+			break;
+
+		case 3:
+			dm->int_active = SIG_LOW;			
+			dm->int_inactive = SIG_OPEN;
+			break;
+	}
+	dm->intcr = value & 3;
+}
+static uint32_t
+sccr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+sccr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Not implemented 0x%02x\n",index);
+}
+static uint32_t
+rsccr_read(DM9000 *dm,uint8_t index,int rqlen)
+{
+        fprintf(stderr,"DM9000: Not implemented 0x%02x\n",index);
+        return 0;
+}
+static void
+rsccr_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
+{
+        dprintf("DM9000: Not implemented 0x%02x\n",index);
+}
+
+
+/*
+ * -------------------------------------------------------------------
+ * MRCMDX returns the value read from SRAM on previous Cycle !
+ * It does not update the cached value from SRAM when previous cycle
+ * had the same index
+ * -------------------------------------------------------------------
+ */
 static uint32_t
 mrcmdx_read(DM9000 *dm,uint8_t index,int rqlen)
 {
 	int i;
 	uint32_t rp=dm->rxfifo_rp;
 	uint32_t value=0;
-	for(i=0;i<rqlen;i++) {
+	uint32_t retval;
+	/* 
+         * -------------------------------------------
+	 * If index did not change since last read of
+	 * mrcmdx an outdated value is returned
+         * -------------------------------------------
+ 	 */
+	if(dm->mrcmdx_idx_tag == dm->index_chg_cnt) {
+		return dm->mrcmd_latch;
+	}
+	for(i=0;i<dm->buswidth;i++) {
 		value = value | (dm->sram[rp] << (i*8));
 		rp=(rp+1);
 		if(rp>=16384) {
 			rp=3*1024;
 		}
 	}
-	//fprintf(stderr,"pre-read %04x from mem len %d, rp %d\n",value,rqlen,rp);
-        return value;
+	dm->mrcmdx_idx_tag = dm->index_chg_cnt;
+	/* 
+	 * ------------------------------------------------------
+	 * Return the value which was latched by previous read
+	 * and store the new version in the latch
+	 * ------------------------------------------------------
+	 */
+	retval = dm->mrcmd_latch;
+	dm->mrcmd_latch = value;
+        return retval;
 }
 
 static void
@@ -1291,24 +1543,41 @@ mrcmdx_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
         fprintf(stderr,"DM9000: mrcmdx register is not writeable\n");
 }
 
+/*
+ * ---------------------------------------------------------------------------
+ * MRCMD: Read from rx fifo 
+ * ---------------------------------------------------------------------------
+ */
 static uint32_t
 mrcmd_read(DM9000 *dm,uint8_t index,int rqlen)
 {
 	int i;
-	uint32_t rp=dm->rxfifo_rp;
+	uint32_t rp = dm->rxfifo_rp;
 	uint32_t value=0;
-	for(i=0;i<rqlen;i++) {
+	uint32_t retval;
+	rp = rp + dm->buswidth;
+	if(rp >= 16384)  {
+		rp = 3*1024 + (rp & (dm->buswidth -1));
+	}
+	dm->rxfifo_rp = rp;
+	for(i=0;i<dm->buswidth;i++) {
 		value = value | (dm->sram[rp] << (i*8));
 		rp=(rp+1);
 		dm->rxfifo_count--;
-		if(rp>=16384)  {
+		if(rp >= 16384)  {
 			rp=3*1024;
 		}
 	}
-	dm->rxfifo_rp = rp;
+	/* 
+ 	 * ---------------------------------------------------------
+	 * Return the value which was latched on a previous read 
+ 	 * and store the next value from fifo in the latch.
+ 	 * ---------------------------------------------------------
+	 */
+	retval = dm->mrcmd_latch;
+	dm->mrcmd_latch = value;
 	update_receiver_status(dm);
-//	fprintf(stderr,"read %04x from mem len %d\n",value,rqlen);
-        return value;
+        return retval;
 }
 
 static void
@@ -1325,11 +1594,6 @@ mrrl_read(DM9000 *dm,uint8_t index,int rqlen)
 static void
 mrrl_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
-	fprintf(stderr,"MRRL write %02x\n",value);
-	if(rqlen>1) {
-		dm->rxfifo_rp = value;
-		fprintf(stderr,"MRRL 16 Bit write not implemented\n");
-	}
 	dm->rxfifo_rp = (dm->rxfifo_rp & 0xff00) | value;
 	update_rxfifo_count(dm);
 	update_receiver_status(dm);
@@ -1344,10 +1608,6 @@ mrrh_read(DM9000 *dm,uint8_t index,int rqlen)
 static void
 mrrh_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
-	fprintf(stderr,"Mrrh write %02x\n",value);
-	if(rqlen>1) {
-		fprintf(stderr,"MRRH 16 Bit write not possible\n");
-	}
 	dm->rxfifo_rp = (dm->rxfifo_rp & 0xff) | (value <<8);
 	update_rxfifo_count(dm);
 	update_receiver_status(dm);
@@ -1365,7 +1625,7 @@ mwcmdx_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen)
 {
 	int i;
 	uint32_t wp=dm->txfifo_wp;
-	for(i=0;i<rqlen;i++) {
+	for(i=0;i<dm->buswidth;i++) {
 		dm->sram[wp] = (value >> (i*8)) & 0xff;
 		wp=(wp+1) % (3*1024);
 	}
@@ -1390,7 +1650,7 @@ static void
 mwcmd_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
 	int i;
-	for(i=0;i<rqlen;i++) {
+	for(i=0;i<dm->buswidth;i++) {
 		dm->sram[dm->txfifo_wp] = (value >> (i*8)) & 0xff;
 		dm->txfifo_wp=(dm->txfifo_wp+1) % (3*1024);
 	}
@@ -1405,11 +1665,7 @@ mwrl_read(DM9000 *dm,uint8_t index,int rqlen)
 static void
 mwrl_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
-	if(rqlen<2) {
-		dm->txfifo_wp = (dm->txfifo_wp & 0xff) | value;
-	} else {
-		dm->txfifo_wp = value % (3*1024);
-	}
+	dm->txfifo_wp = (dm->txfifo_wp & 0xff) | value;
 }
 
 static uint32_t
@@ -1422,7 +1678,6 @@ static void
 mwrh_write(DM9000 *dm,uint32_t value,uint8_t index,int rqlen) 
 {
 	dm->txfifo_wp = (dm->txfifo_wp & 0xff)  | (value << 8);
-	dm->txfifo_wp = dm->txfifo_wp % 3*1024; /* not verified with real device */
 }
 
 static uint32_t
@@ -1534,6 +1789,18 @@ create_registers(DM9000 *dm) {
 	DMReg_New(DM_PID1,pidh_read,pidh_write,dm);
 	DMReg_New(DM_CHIPR,chipr_read,chipr_write,dm);
 	DMReg_New(DM_SMCR,smcr_read,smcr_write,dm);
+
+	/* Only new revision */
+	DMReg_New(DM_ETXCSR,etxcsr_read,etxcsr_write,dm);
+	DMReg_New(DM_TCSCR,tcscr_read,tcscr_write,dm);
+	DMReg_New(DM_RCSCSR,rcscsr_read,rcscsr_write,dm);
+	DMReg_New(DM_MPAR,mpar_read,mpar_write,dm);
+	DMReg_New(DM_LEDCR,ledcr_read,ledcr_write,dm);
+	DMReg_New(DM_BUSCR,buscr_read,buscr_write,dm);
+	DMReg_New(DM_INTCR,intcr_read,intcr_write,dm);
+	DMReg_New(DM_SCCR,sccr_read,sccr_write,dm);
+	DMReg_New(DM_RSCCR,rsccr_read,rsccr_write,dm);
+
 	DMReg_New(DM_MRCMDX,mrcmdx_read,mrcmdx_write,dm);
 	DMReg_New(DM_MRCMD,mrcmd_read,mrcmd_write,dm);
 	DMReg_New(DM_MRRL,mrrl_read,mrrl_write,dm);
@@ -1552,7 +1819,10 @@ static void
 index_write(void *clientData,uint32_t value,uint32_t address,int rqlen) 
 {
 	DM9000 *dm = (DM9000 *) clientData;
-	dm->index_reg = value;
+	if(dm->index_reg != value) {
+		dm->index_reg = value;
+		dm->index_chg_cnt++;
+	}
 }
 
 static uint32_t
@@ -1666,7 +1936,7 @@ iphy_write(PHY_Device *phy,uint16_t value,int reg)
 	switch(reg) {
 		case	DMP_CONTROL:
 			dm->control=value;
-			fprintf(stderr,"write to phy reg 0x%02x value 0x%04x\n",reg,value);
+			dprintf("write to phy reg 0x%02x value 0x%04x\n",reg,value);
 			break;	
 
 		case	DMP_STATUS:
@@ -1727,7 +1997,7 @@ epy_access_done(void *clientData)
 BusDevice *
 DM9000_New(const char *devname,int register_spacing) {
         DM9000 *dm = (DM9000*)malloc(sizeof(DM9000));
-	char *epromname = alloca(strlen(devname) + 20);
+	char *epromname = (char*)alloca(strlen(devname) + 20);
 	char *irqrequest_name = (char*)alloca(strlen(devname)+20);
         if(!dm) {
                 fprintf(stderr,"Out of memory \n");
@@ -1760,7 +2030,8 @@ DM9000_New(const char *devname,int register_spacing) {
         dm->an_lp_ability = 0x45e1;
         dm->an_expansion = 0;
 	dm->dscr=0x410;
-	dm->dscsr=0xf010;
+	dm->buscr = 0x61;
+	dm->dscsr=0x8010;
 	dm->btcsr10=0x7800;
 	sprintf(epromname,"%s.eeprom",devname);
 	dm->eprom = m93c46_New(epromname);
