@@ -26,23 +26,39 @@
 
 #ifndef DISKIMAGE_H
 #define DISKIMAGE_H
+
+#include <stdio.h>
+#include <stdint.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <sys/file.h>
+#include <sys/mman.h>
 #include <sys/types.h>
 
+#include <stdio.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <stdint.h>
+
 #define DI_RDWR		(16)
 #define DI_CREAT_FF	(1)
 #define	DI_CREAT_00	(2)
+#define DI_SPARSE	(4)
 typedef struct DiskImage {
 	int fd;
-	size_t size;
+	uint64_t size;
 	void *map;
 } DiskImage;
 
 void DiskImage_Close(DiskImage *di);
-DiskImage * DiskImage_Open(const char *name,size_t size,int flags);
+DiskImage * DiskImage_Open(const char *name,uint64_t size,int flags);
 
 void * DiskImage_Mmap(DiskImage *di); 
 
-int DiskImage_Read(DiskImage *di,off_t ofs,uint8_t *buf,int count);
-int DiskImage_Write(DiskImage *di,off_t ofs,const uint8_t *buf,int count);
+int DiskImage_Read(DiskImage *di,off64_t ofs,uint8_t *buf,int count);
+int DiskImage_Write(DiskImage *di,off64_t ofs,const uint8_t *buf,int count);
 
 #endif
