@@ -1,7 +1,7 @@
 /*
  * -------------------------------------------------------------------------
  *
- * Emulation of Dallas DS1305 realtime clock
+ * Emulation of Dallas DS1305 realtime clock with SPI Interface
  *
  * (C) 2006  Jochen Karrer 
  *
@@ -191,7 +191,7 @@ static void
 update_time(DS1305 *ds) {
 	struct timeval host_tv; 
         time_t time;
-        struct tm *tm;
+        struct tm tm;
 	if(ds->control_reg & CONTROL_NEOSC) {
 		return;
 	}
@@ -208,29 +208,29 @@ update_time(DS1305 *ds) {
 		ds->useconds += 1000000;
 	}
 	//fprintf(stderr,"seconds %ld\n",time);
-        tm = gmtime(&time);
-        //fprintf(stderr,"host: tm.hour %d tm.tm_isdst %d\n",tm->tm_hour,tm->tm_isdst);
-	if((tm->tm_year < 100) || (tm->tm_year >= 200)) {
-		fprintf(stderr,"DS1305: Illegal year %d\n",tm->tm_year+1900); 
+        gmtime_r(&time,&tm);
+        //fprintf(stderr,"host: tm.hour %d tm.tm_isdst %d\n",tm.tm_hour,tm.tm_isdst);
+	if((tm.tm_year < 100) || (tm.tm_year >= 200)) {
+		fprintf(stderr,"DS1305: Illegal year %d\n",tm.tm_year+1900); 
 	}
-	ds->year = i_to_bcd((tm->tm_year-100) % 100);
-	ds->month = i_to_bcd(tm->tm_mon+1);
-	ds->day = i_to_bcd(tm->tm_wday+1);
-	ds->date = i_to_bcd(tm->tm_mday);
+	ds->year = i_to_bcd((tm.tm_year-100) % 100);
+	ds->month = i_to_bcd(tm.tm_mon+1);
+	ds->day = i_to_bcd(tm.tm_wday+1);
+	ds->date = i_to_bcd(tm.tm_mday);
 	if(ds->hours & HOURS_12H) {
 		int hour;
 		int pm = 0;
-		if(tm->tm_hour == 0) {
+		if(tm.tm_hour == 0) {
 			hour = 12;
 			pm = 0;
-		} else if(tm->tm_hour < 12) {
-			hour = tm->tm_hour;
+		} else if(tm.tm_hour < 12) {
+			hour = tm.tm_hour;
 			pm = 0;
-		} else if(tm->tm_hour < 13) {
+		} else if(tm.tm_hour < 13) {
 			hour = 12;
 			pm = 1;	
 		} else {
-			hour = tm->tm_hour - 12; 
+			hour = tm.tm_hour - 12; 
 			pm = 1;
 		}
 		ds->hours = (ds->hours & 0xe0) | i_to_bcd(hour);
@@ -238,11 +238,11 @@ update_time(DS1305 *ds) {
 			ds->hours |= HOURS_PM;
 		}
 	} else {
-		ds->hours = i_to_bcd(tm->tm_hour);
+		ds->hours = i_to_bcd(tm.tm_hour);
 	}
-	ds->minutes = i_to_bcd(tm->tm_min);
-	ds->seconds = i_to_bcd(tm->tm_sec);
-	fprintf(stderr,"UDT: Year %d, hour %d minute %d\n",tm->tm_year+1900,tm->tm_hour,tm->tm_min);
+	ds->minutes = i_to_bcd(tm.tm_min);
+	ds->seconds = i_to_bcd(tm.tm_sec);
+	fprintf(stderr,"UDT: Year %d, hour %d minute %d\n",tm.tm_year+1900,tm.tm_hour,tm.tm_min);
 	return;
 }
 
@@ -983,7 +983,7 @@ DS1305_New(const char *name)
         } else {
 		ds->time_offset = 0;
         }
-	fprintf(stderr,"DS1305 created with time offset %lld\n",ds->time_offset);
+	fprintf(stderr,"DS1305 created with time offset %lld\n",(long long)ds->time_offset);
 	update_time(ds);
 	return &ds->spidev;
 }
