@@ -53,6 +53,13 @@
 
 #define CMDBUF_SIZE (512)
 
+typedef struct BreakPoint {
+	uint64_t addr;
+	int len;
+	uint8_t backup[8];
+	struct BreakPoint *next;
+} BreakPoint;
+
 typedef struct GdbSession {
 	int fd;
 	int rfh_is_active;
@@ -65,6 +72,7 @@ typedef struct GdbSession {
 	int cmdbuf_wp;
 	uint8_t csum;
 	void *backend;
+	BreakPoint *bkpt_head;
 	struct GdbSession *next;
 } GdbSession;
 
@@ -141,6 +149,60 @@ gsess_reply(GdbSession *gsess,const char *format,...) {
 	return 0;
 }
 
+/*
+ * ---------------------------------------------------
+ * Find a breakpoint by address/length pair
+ * ---------------------------------------------------
+ */
+static BreakPoint *
+find_breakpoint(GdbSession *gsess,uint64_t addr,int len) 
+{
+	BreakPoint *cursor;
+	for(cursor=gsess->bkpt_head;cursor; cursor=cursor->next) {
+		if((cursor->addr == addr) && (cursor->len == len)) {
+			return cursor;
+		}
+	}
+	return NULL;
+}
+
+/*
+ * ---------------------------------------------------------------------------------
+ * Unlink a breakpoint identified by a address/length pair from the linked list
+ * ---------------------------------------------------------------------------------
+ */
+static BreakPoint *
+unlink_breakpoint(GdbSession *gsess,uint64_t addr,int len) 
+{
+	BreakPoint *cursor,*prev;
+	for(prev=NULL,cursor=gsess->bkpt_head;cursor; prev=cursor,cursor=cursor->next) {
+		if((cursor->addr == addr) && (cursor->len == len)) {
+			if(prev) {
+				prev->next = cursor->next;
+			} else {
+				gsess->bkpt_head = cursor->next;
+			}
+			return cursor;
+		}
+	}
+	return NULL;
+}
+static void 
+remove_breakpoints(GdbSession *gsess) 
+{
+	BreakPoint *cursor;
+	GDebugOps *dbgops = gsess->dbgops;
+	if(!dbgops->setmem)
+		return;
+	while(gsess->bkpt_head) {
+		cursor = gsess->bkpt_head;
+		dbgops->setmem(gsess->backend,cursor->backup,cursor->addr,cursor->len);
+		gsess->bkpt_head = cursor->next;
+		free(cursor);
+
+	}
+}
+
 void
 gsess_del(GdbSession *gsess) {
         uint8_t c;
@@ -167,6 +229,7 @@ gsess_del(GdbSession *gsess) {
 	if(gserv->first_gsess == gsess) {
 		gserv->first_gsess = NULL;
 	}
+	remove_breakpoints(gsess);
         free(gsess);
 }
 
@@ -379,11 +442,11 @@ gsess_setmem(GdbSession *gsess,char *data,int maxlen)
 		return;
 	}
 	while(readp < maxlen) {
-		if(data[readp] == ':') {
+		if(data[readp++] == ':') {
 			break;
 		}	
 	}
-	if((readp+2*len)>=maxlen) {
+	if((readp+2*len)>maxlen) {
 		gsess_reply(gsess,"E00");
 		return;
 	}
@@ -391,9 +454,11 @@ gsess_setmem(GdbSession *gsess,char *data,int maxlen)
 		uint8_t value[4];
 		readlen = (len >= 4) ? 4 : len;	
 		if(hexparse(data+readp,value,readlen) <= 0) {
+			fprintf(stderr,"setmem: Parse hex string %s failed\n",data+readp);
 			gsess_reply(gsess,"E00");
 			return;
 		}
+		fprintf(stderr,"setmem value %02x, addr %08x\n",value[0],addr);
 		dbgops->setmem(gsess->backend,value,addr,readlen);
 		len-=readlen;
 		readp+=readlen;
@@ -439,12 +504,127 @@ gsess_setreg(GdbSession *gsess,char *data)
 	gsess_reply(gsess,"OK");
 }
 
+/*
+ * -----------------------------------------------------------------------
+ * add break- and watchpoints
+ * Z0 set memory breakpoint
+ * Z1 set hardware breakpoint
+ * Z2 Insert a write watchpoint
+ * Z3 Insert a read watchpoint
+ * Z4 Insert an access watchpoint
+ * -----------------------------------------------------------------------
+ */
+static void
+gsess_add_breakpoint(GdbSession *gsess,char *cmd,int cmdlen) 
+{
+	GDebugOps *dbgops = gsess->dbgops;
+	BreakPoint *bkpt;
+	int type;
+	uint32_t addr; 
+	uint8_t bkpt_ins[8];
+	unsigned int len;
+	if(sscanf(cmd,"Z%d,%x,%x",&type,&addr,&len) != 3) {
+		fprintf(stderr,"gdebug: parse error cmd \"%s\"\n",cmd);
+		gsess_reply(gsess,"E00");
+		return;
+	}	
+	if(len > 8) {
+		fprintf(stderr,"gdebug: bkpt instruction to long (%d)\n",len);
+		gsess_reply(gsess,"E00");
+		return;
+	}
+	if(!dbgops->get_bkpt_ins || !dbgops->setmem || !dbgops->getmem) {
+		fprintf(stderr,"gdebug backend does not support breakpoints\n");
+		gsess_reply(gsess,"");
+	}
+	dbgops->get_bkpt_ins(gsess->backend,bkpt_ins,addr,len);
+	switch(type) {
+		case 0:
+			bkpt = find_breakpoint(gsess,addr,len);
+			if(bkpt)  {
+				fprintf(stderr,"Breakpoint already exists\n");
+				gsess_reply(gsess,"E00");
+				return;
+			}
+			bkpt = malloc(sizeof(*bkpt));
+			memset(bkpt,0,sizeof(*bkpt));
+			bkpt->addr = addr;
+			bkpt->len = len;
+			bkpt->next = gsess->bkpt_head;
+			gsess->bkpt_head = bkpt;
+			dbgops->getmem(gsess->backend,bkpt->backup,addr,len);
+			dbgops->setmem(gsess->backend,bkpt_ins,addr,len);
+			gsess_reply(gsess,"OK");
+			break;
+		default:
+			gsess_reply(gsess,"");
+			break;
+	}
+	fprintf(stderr,"softgun gdb interface does not support breakpoints\n");
+	fprintf(stderr,"cmd %s\n",cmd);
+}
+
+/*
+ * -------------------------------------------------------------------------
+ * remove break- and watchpoints
+ * z0 ... z4
+ * -------------------------------------------------------------------------
+ */
+static void
+gsess_remove_breakpoint(GdbSession *gsess,char *cmd,int cmdlen) 
+{
+	GDebugOps *dbgops = gsess->dbgops;
+	BreakPoint *bkpt;
+	int type;
+	uint32_t addr; 
+	unsigned int len;
+	if(sscanf(cmd,"z%d,%x,%x",&type,&addr,&len) != 3) {
+		fprintf(stderr,"gdebug: parse error cmd \"%s\"\n",cmd);
+		gsess_reply(gsess,"E00");
+		return;
+	}	
+	if(len > 8) {
+		fprintf(stderr,"gdebug: bkpt instruction to long (%d)\n",len);
+		gsess_reply(gsess,"E00");
+		return;
+	}
+	if(!dbgops->setmem) {
+		fprintf(stderr,"gdebug backend does not support breakpoints\n");
+		gsess_reply(gsess,"");
+	}
+	switch(type) {
+		case 0:
+			bkpt = unlink_breakpoint(gsess,addr,len);
+			if(!bkpt) {
+				fprintf(stderr,"Removing nonexistent breakpoint\n");
+				gsess_reply(gsess,"E00");
+				return;
+			}
+			dbgops->setmem(gsess->backend,bkpt->backup,addr,len);
+			free(bkpt);
+			gsess_reply(gsess,"OK");
+			break;
+			
+		default:
+			gsess_reply(gsess,"");
+			break;
+	}
+	fprintf(stderr,"softgun gdb interface does not support breakpoints\n");
+	fprintf(stderr,"cmd %s\n",cmd);
+}
+
+/*
+ * -------------------------------------------------------
+ * gsess_execute_cmd
+ * 	The interpreter for commands from gdb 
+ * -------------------------------------------------------
+ */
 static int 
 gsess_execute_cmd(GdbSession *gsess) 
 {
 	char *cmd = gsess->cmdbuf;
 	int result=0;
-	dprintf("Das Kommando wird hingerichted\n");	
+	//printf("The gdb command is executed %s\n",cmd);	
 	switch(cmd[0]) {
 		/* Status query */
 		case '?':
@@ -511,6 +691,14 @@ gsess_execute_cmd(GdbSession *gsess)
 			gsess_setmem(gsess,cmd+1,gsess->cmdbuf_wp-1);
 			break;
 
+		/* Insert/Remove Breakpoints */
+		case 'z':
+			gsess_remove_breakpoint(gsess,cmd,gsess->cmdbuf_wp);
+			break;
+		case 'Z':
+			gsess_add_breakpoint(gsess,cmd,gsess->cmdbuf_wp);
+			break;
+			
 		default:
 			dprintf("unknown cmd \'%c\'\n",cmd[0]);
 			gsess_output(gsess,"$#00",4); 
@@ -519,6 +707,12 @@ gsess_execute_cmd(GdbSession *gsess)
 	return result;
 }
 
+/*
+ * -------------------------------------------------------------------
+ * feed_state_machine
+ * 	The data sink for characters from gdb
+ * -------------------------------------------------------------------
+ */
 static int
 feed_state_machine (GdbSession *gsess,uint8_t c) {
 	int result = 0;
@@ -549,7 +743,7 @@ feed_state_machine (GdbSession *gsess,uint8_t c) {
 				gsess->cmdbuf[gsess->cmdbuf_wp] = 0;
 			} else {
 				if(gsess->cmdbuf_wp >= CMDBUF_SIZE) {
-					dprintf("Message from gdb to long, ignoring\n");
+					fprintf(stderr,"Message from gdb to long, ignoring\n");
 				} else {
 					gsess->cmdbuf[gsess->cmdbuf_wp++] = c;
 					gsess->csum += c;
