@@ -26,4 +26,4 @@
  */
 
 typedef int XY_IHexDataHandler(uint32_t addr,uint8_t *buf,int len,void *clientData);
-int XY_LoadIHexFile(const char *filename, XY_IHexDataHandler *cb,void *clientData);
+int64_t XY_LoadIHexFile(const char *filename, XY_IHexDataHandler *cb,void *clientData);
