@@ -41,7 +41,8 @@
 
 //#define CRM_BASE_ADDR (0x10027000)
 #define CRM_CSCR(base)        ((base) + 0x0)
-#define 	CSCR_CLKO_SEL(n)  (((n) & 0x07) << 29) /* Clock signal source for the CLKO pin. */
+#define 	CSCR_PRESC_POS   (29)
+#define 	CSCR_PRESC(n)     (((n) & 0x03) << CSCR_PRESC_POS) /* MPU PLL clock prescaler */
 #define 	CSCR_USB_DIV(n)   (((n) & 0x07) << 26) /* Divider for CLK48M generation */
 #define 	CSCR_SD_CNT(n)    (((n) & 0x03) << 24) /* Shut-Down count */
 #define 	CSCR_SPLL_RESTART (1 << 22)            /* Restarts the SPLL */
@@ -51,24 +52,13 @@
 #define 	CSCR_FIR_SEL      (1 << 18)            /* FIR and MIR clock select */
 #define 	CSCR_SP_SEL       (1 << 17)            /* SPLL clock source */
 #define 	CSCR_MCU_SEL      (1 << 16)            /* MPLL clock source */
-#define 	CSCR_PRESC_POS    14
-#define 	CSCR_PRESC(n)     (((n) & 0x03) << CSCR_PRESC_POS) /* MPU PLL clock prescaler */
 #define 	CSCR_BCLKDIV_POS  10
 #define 	CSCR_BCLKDIV(n)   (((n) & 0x0f) << CSCR_BCLKDIV_POS) /* System Bus Clock Divider */
 #define 	CSCR_IPDIV(n)     (((n) & 0x01) << 9)  /* Peripheral Clock Divider */
-#define 	CSCR_OSC26M_EN    (1 << 3)             /* Enables the 26 MHz oscillator */
+#define 	CSCR_OSC26M_DIS    (1 << 3)             /* Enables the 26 MHz oscillator */
 #define 	CSCR_FPM_EN       (1 << 2)             /* Frequency Premultiplier Enable */
 #define 	CSCR_SPEN         (1 << 1)             /* Serial Peripheral PLL Enable */
 #define 	CSCR_MPEN         (1 << 0)             /* MPLL Enable */
-/* CLKO pin selects to be used with CSCR_CLKO_SEL(n) */
-#define 	CSCR_CLKO_HCLK     0       /* HCLK               */
-#define 	CSCR_CLKO_SPLLREF  1       /* SPLL reference */
-#define 	CSCR_CLKO_CLK48M   2       /* CLK48M             */
-#define 	CSCR_CLKO_CLK26M   3       /* CLK26M             */
-#define 	CSCR_CLKO_FPMCLK   4       /* FPMCLK             */
-#define 	CSCR_CLKO_FCLK     5       /* FCLK               */
-#define		CSCR_CLKO_CLK32    6       /* CLK32              */
-#define 	CSCR_CLKO_IPGCLK   7       /* IPG_CLK            */
 #define CRM_MPCTL0(base)      ((base) + 0x4)
 #define CRM_MPCTL1(base)      ((base) + 0x8)
 #define		MPCTL1_BRMO 	(1<<6)
@@ -78,6 +68,14 @@
 #define		SPCTL1_BRMO	(1<<6)
 #define		SPCTL1_LF	(1<<15)
 #define CRM_OSC26MCTL(base)   ((base) + 0x14)
+#define		OSC26MCTL_AGC_SHIFT	(8)
+#define		OSC26MCTL_AGC_MASK	(0x3f<<8)
+#define		OSC26MCTL_PEAK_MASK	(3<<16)
+#define		OSC26MCTL_PEAK_SHIFT	(16)
+#define		OSC26M_PEAK_OK		(0<<16)
+#define		OSC26M_PEAK_LOW		(1<<16)
+#define		OSC26M_PEAK_HIGH	(2<<16)
+#define		OSC26M_PEAK_INVALID	(3<<16)
 #define CRM_PCDR0(base)       ((base) + 0x18)
 #define CRM_PCDR1(base)       ((base) + 0x1c)
 #define CRM_PCCR0(base)       ((base) + 0x20)
@@ -130,44 +128,43 @@
 #define		CCSR_CLKO_SEL_MASK	(0x1f)
 #define		CCSR_CLKO_SEL_SHIFT	(0)
 #define		CCSR_CLKO_SEL(x)	((x) & 0x1f)
+#if 1
+/* CLKO pin selects to be used with CCSR_CLKO_SEL(n) */
+#define 	CCSR_CLK0_CLK32		(0)	
+#define 	CCSR_CLKO_PREMCLK 	(1)
+#define 	CCSR_CLKO_CLK26M	(2)
+#define 	CCSR_CLKO_MPLLREFCLK	(3)
+#define 	CCSR_CLKO_SPLLREFCLK	(4)
+#define 	CCSR_CLKO_MPLLCLK	(5)
+#define		CCSR_CLKO_SPLLCLK	(6)
+#define 	CCSR_CLKO_FCLK		(7)
+#define 	CCSR_CLKO_HCLK		(8)
+#define		CCSR_CLKO_IPG_CLK	(9)
+#define		CCSR_CLK0_PERCLK1	(10)
+#define		CCSR_CLK0_PERCLK2	(11)
+#define		CCSR_CLK0_PERCLK3	(12)
+#define		CCSR_CLK0_PERCLK4	(13)
+#define		CCSR_CLKO_SSI1BAUD	(14)
+#define		CCSR_CLKO_SSI2BAUD	(15)
+#define		CCSR_CLKO_NFCBAUD	(16)
+#define		CCSR_CLKO_FIRI		(17)
+#define		CCSR_CLK0_CLK48MALW	(18)
+#define		CCSR_CLK0_CLK32KALW	(19)
+#define		CCSR_CLK0_CLK48M	(20)
+#define		CCSR_CLK0_CLK48DIV_CLKO	(21)
+#endif
 #define CRM_PMCTL(base)       ((base) + 0x2c)
 #define CRM_PMCOUNT(base)     ((base) + 0x30)
 #define CRM_WKGDCTL(base)     ((base) + 0x34)
 
-#define 	CSCR_CLKO_SEL(n)  (((n) & 0x07) << 29) /* Clock signal source for the CLKO pin. */
-#define 	CSCR_USB_DIV(n)   (((n) & 0x07) << 26) /* Divider for CLK48M generation */
-#define 	CSCR_SD_CNT(n)    (((n) & 0x03) << 24) /* Shut-Down count */
-#define 	CSCR_SPLL_RESTART (1 << 22)            /* Restarts the SPLL */
-#define 	CSCR_MPLL_RESTART (1 << 21)            /* Restarts the MPLL */
-#define 	CSCR_SSI2_SEL     (1 << 20)            /* SSI2 Baud Source Select */
-#define 	CSCR_SSI1_SEL     (1 << 19)            /* SSI1 Baud Source Select */
-#define 	CSCR_FIR_SEL      (1 << 18)            /* FIR and MIR clock select */
-#define 	CSCR_SP_SEL       (1 << 17)            /* SPLL clock source */
-#define 	CSCR_MCU_SEL      (1 << 16)            /* MPLL clock source */
-#define 	CSCR_PRESC_POS    14
-#define 	CSCR_PRESC(n)     (((n) & 0x03) << CSCR_PRESC_POS) /* MPU PLL clock prescaler */
-#define 	CSCR_BCLKDIV_POS  10
-#define 	CSCR_BCLKDIV(n)   (((n) & 0x0f) << CSCR_BCLKDIV_POS) /* System Bus Clock Divider */
-#define 	CSCR_IPDIV(n)     (((n) & 0x01) << 9)  /* Peripheral Clock Divider */
-#define 	CSCR_OSC26M_EN    (1 << 3)             /* Enables the 26 MHz oscillator */
-#define 	CSCR_FPM_EN       (1 << 2)             /* Frequency Premultiplier Enable */
-#define 	CSCR_SPEN         (1 << 1)             /* Serial Peripheral PLL Enable */
-#define 	CSCR_MPEN         (1 << 0)             /* MPLL Enable */
-/* CLKO pin selects to be used with CSCR_CLKO_SEL(n) */
-#define 	CSCR_CLKO_HCLK     0       /* HCLK               */
-#define 	CSCR_CLKO_SPLLREF  1       /* SPLL reference */
-#define 	CSCR_CLKO_CLK48M   2       /* CLK48M             */
-#define 	CSCR_CLKO_CLK26M   3       /* CLK26M             */
-#define 	CSCR_CLKO_FPMCLK   4       /* FPMCLK             */
-#define 	CSCR_CLKO_FCLK     5       /* FCLK               */
-#define 	CSCR_CLKO_CLK32    6       /* CLK32              */
-#define 	CSCR_CLKO_IPGCLK   7       /* IPG_CLK            */
-
 /* 
  * System controll registers from Chapter 9 are included in the CRM module according to the
  * memory map
  */
-#define	SYSCTL_SIDR(base)	((base)+0x804)
+#define	SYSCTL_SIDR0(base)	((base)+0x804)
+#define	SYSCTL_SIDR1(base)	((base)+0x808)
+#define	SYSCTL_SIDR2(base)	((base)+0x80c)
+#define	SYSCTL_SIDR3(base)	((base)+0x810)
 #define	SYSCTL_FMCR(base)	((base)+0x814)
 #define SYSCTL_GPCR(base)	((base)+0x818)
 #define	SYSCTL_WBCR(base)	((base)+0x81c)
@@ -187,6 +184,7 @@
 
 typedef struct IMX_Crm {
 	BusDevice bdev;
+	char *name;
 	Clock_t *osc32;
 	Clock_t *osc26m;
 	Clock_t *premclk;
@@ -288,11 +286,20 @@ cscr_read(void *clientData,uint32_t address,int rqlen)
 static void
 update_mpll_clk(IMX_Crm *crm) {
 	uint32_t value = crm->mpctl0;
-	unsigned int pd = (value >> 26) & 0xf;
-	unsigned int mfd = (value >> 16) & 0x3ff;
-	unsigned int mfi = (value >> 10) & 0xf;
-	unsigned int mfn = (value & 0x3ff);
+	unsigned int pd;
+	unsigned int mfd;
+	unsigned int mfi;
+	unsigned int mfn;
 	uint64_t mul,div;
+	float fractional_part;
+	int brmo;
+	pd = (value >> 26) & 0xf;
+	mfd = (value >> 16) & 0x3ff;
+	mfi = (value >> 10) & 0xf;
+	mfn = (value & 0x3ff);
+	if(mfi < 5) {
+		mfi = 5;
+	}
 	mul = 2 * (mfi * (mfd+1) + mfn);
 	div = (mfd+1) * (pd+1);
 	if(crm->cscr & CSCR_MPEN) {
@@ -300,7 +307,19 @@ update_mpll_clk(IMX_Crm *crm) {
 	} else {
 		Clock_MakeDerived(crm->mpll_clk,crm->mpll_refclk,0,div);
 	}
-	fprintf(stderr,"MPCTL0 : multiplier %f, freq %f\n",(double)mul/div,Clock_Value(crm->mpll_clk));
+	/* fractional part between ]0.1..0.9[ requires second order BRMO */
+	fractional_part = mfn / (mfd+1);
+	if((fractional_part > 0.9) || (fractional_part < 0.1)) {
+		brmo = 0;
+	} else {
+		brmo = 1;
+	}
+	if(brmo != !!(crm->mpctl1 & MPCTL1_BRMO))  {
+		fprintf(stderr,"MPCTL warning: fractional part is %f, BRMO bit should be %d\n"
+			,fractional_part,brmo);
+	}
+	//fprintf(stderr,"MPCTL0 mul %lld, div %lld, mfn %d, mfd %d\n",mul,div,mfn,mfd);
+	//fprintf(stderr,"MPCTL0 : multiplier %f, freq %f\n",(double)mul/div,Clock_Freq(crm->mpll_clk));
 }
 static void
 update_spll_clk(IMX_Crm *crm) {
@@ -310,9 +329,14 @@ update_spll_clk(IMX_Crm *crm) {
 	unsigned int mfi;
 	unsigned int mfn;
 	uint64_t mul,div;
+	float fractional_part;
+	int brmo;
 	pd = (value >> 26) & 0xf;
 	mfd = (value >> 16) & 0x3ff;
 	mfi = (value >> 10) & 0xf;
+	if(mfi < 5) {
+		mfi = 5;
+	}
 	mfn = (value >> 0) & 0x3ff;
 	mul = 2 * (mfi * (mfd+1) + mfn);
 	div = (mfd+1) * (pd+1);
@@ -321,7 +345,17 @@ update_spll_clk(IMX_Crm *crm) {
 	} else {
 		Clock_MakeDerived(crm->spll_clk,crm->spll_refclk,0,div);
 	}
-	fprintf(stderr,"SPCTL0 : multiplier %f, freq %f\n",(double)mul/div,Clock_Value(crm->spll_clk));
+	fractional_part = mfn / (mfd+1);
+	if((fractional_part > 0.9) || (fractional_part < 0.1)) {
+		brmo = 0;
+	} else {
+		brmo = 1;
+	}
+	if(brmo != !!(crm->spctl1 & SPCTL1_BRMO))  {
+		fprintf(stderr,"SPCTL warning: fractional part is %f, BRMO bit should be %d\n"
+			,fractional_part,brmo);
+	}
+	//fprintf(stderr,"SPCTL0 : multiplier %f, freq %f\n",(double)mul/div,Clock_Freq(crm->spll_clk));
 }
 static void
 cscr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
@@ -330,8 +364,8 @@ cscr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	int presc = (value >> 29) & 7;
 	int usb_div = (value >> 26) & 7;
 	//int sd_cnt = (value >> 24) & 3;
-	//int spll_restart = (value >> 22) & 1;	
-	//int mpll_restart = (value >> 21) & 1;	
+	int spll_restart = (value >> 22) & 1;	
+	int mpll_restart = (value >> 21) & 1;	
 	int ssi2_sel = (value >> 20) & 1;
 	int ssi1_sel = (value >> 19) & 1;
 	int fir_sel = (value >> 18) & 1;
@@ -343,12 +377,16 @@ cscr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	int osc26m_dis = (value >> 3) & 1;
 	int fpm_en = (value >> 2) & 1;	
 	crm->cscr = value & 0xff7f3e1f;
-	/* Should by delayed by 1-2 CLK32 cycles */
+	/* Should be delayed by 1-2 CLK32 cycles */
 	crm->cscr &= ~(CSCR_MPLL_RESTART | CSCR_SPLL_RESTART);
 	if(osc26m_dis) {
-		Clock_Set(crm->osc26m,0);
+		Clock_SetFreq(crm->osc26m,0);
 	} else {
-		Clock_Set(crm->osc26m,26*1000*1000);
+		uint32_t freq;
+		if(Config_ReadUInt32(&freq,crm->name,"osc26") < 0) {
+			freq = 0;
+		}
+		Clock_SetFreq(crm->osc26m,freq);
 	}
 	if(fpm_en) {
 		Clock_MakeDerived(crm->premclk,crm->osc32,512,1);
@@ -374,22 +412,32 @@ cscr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	Clock_MakeDerived(crm->fclk,crm->mpll_clk,1,presc+1);
 	Clock_MakeDerived(crm->hclk,crm->fclk,1,bclkdiv+1);
 	if(fir_sel) {
+		Clock_Decouple(crm->firiinclk);
 		Clock_MakeDerived(crm->firiinclk,crm->spll_clk,1,1);
 	} else {
+		Clock_Decouple(crm->firiinclk);
 		Clock_MakeDerived(crm->firiinclk,crm->mpll_clk,1,1);
 	}
 	if(ssi1_sel) {
+		Clock_Decouple(crm->ssi1inclk);
 		Clock_MakeDerived(crm->ssi1inclk,crm->mpll_clk,1,1);
 	} else {
+		Clock_Decouple(crm->ssi1inclk);
 		Clock_MakeDerived(crm->ssi1inclk,crm->spll_clk,1,1);
 	}
 	if(ssi2_sel) {
+		Clock_Decouple(crm->ssi2inclk);
 		Clock_MakeDerived(crm->ssi2inclk,crm->mpll_clk,1,1);
 	} else {
+		Clock_Decouple(crm->ssi2inclk);
 		Clock_MakeDerived(crm->ssi2inclk,crm->spll_clk,1,1);
 	}
 	Clock_MakeDerived(crm->perclk,crm->hclk,1,ipdiv+1);
 	Clock_MakeDerived(crm->clk48m,crm->spll_clk,1,usb_div+1);
+	if(spll_restart || mpll_restart) {
+	//	Clock_DumpTree(crm->osc26m);
+	//	Clock_DumpTree(crm->osc32);
+	}
         return;
 }
 
@@ -404,6 +452,7 @@ static void
 mpctl0_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
         IMX_Crm *crm = (IMX_Crm *) clientData;
+	//fprintf(stderr,"MPCTL0: %08x\n",value);
 	crm->mpctl0 = value & 0xbfff3fff;
 	update_mpll_clk(crm);
         return;
@@ -421,6 +470,7 @@ mpctl1_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
         IMX_Crm *crm = (IMX_Crm *) clientData;
 	crm->mpctl1 = value & (MPCTL1_BRMO);
+	//fprintf(stderr,"MPCTL1: %08x\n",value);
         return;
 }
 
@@ -437,6 +487,7 @@ spctl0_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
         IMX_Crm *crm = (IMX_Crm *) clientData;
 	crm->spctl0 = value & 0xbfff3fff;
 	update_spll_clk(crm);
+	//fprintf(stderr,"SPCTL0: %08x\n",value);
         return;
 }
 
@@ -459,7 +510,21 @@ static uint32_t
 osc26mctl_read(void *clientData,uint32_t address,int rqlen)
 {
         IMX_Crm *crm = (IMX_Crm *) clientData;
-	crm->osc26mctl &= ~(3<<16); /* always in desired range */
+	int agc = (crm->osc26mctl & OSC26MCTL_AGC_MASK) >> OSC26MCTL_AGC_SHIFT;
+	uint32_t peak;
+	if((crm->cscr & CSCR_OSC26M_DIS) || (Clock_Freq(crm->osc26m) == 0)) {
+		/* verified with real board without osc26 + cscr disabled */
+		peak = OSC26M_PEAK_INVALID;
+	} else if(agc > 0x22) {
+		peak = OSC26M_PEAK_HIGH;
+	} else if (agc < 0x1a) {
+		peak = OSC26M_PEAK_LOW;
+	} else {
+		peak = OSC26M_PEAK_OK;
+	}
+	//fprintf(stderr,"agc %d, peak %04x\n",agc,peak);
+	crm->osc26mctl &= ~(3<<16);
+	crm->osc26mctl |= peak; /* always in desired range */
         return crm->osc26mctl; 
 }
 
@@ -467,7 +532,7 @@ static void
 osc26mctl_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
         IMX_Crm *crm = (IMX_Crm *) clientData;
-	crm->osc26mctl = value & 0x3f3f;
+	crm->osc26mctl = value & 0x3f00;
         return;
 }
 static void
@@ -526,8 +591,8 @@ update_perdivs(IMX_Crm *crm) {
 	} else {
 		Clock_MakeDerived(crm->ssi2clk,crm->ssi2inclk,0,1);
 	}
-	/*	fprintf(stderr,"** perclk1: freq %f\n",Clock_Value(crm->perclk1)); 	*/
-	/*	fprintf(stderr,"** mpll_clk: freq %f\n",Clock_Value(crm->mpll_clk)); 	*/
+	/*	fprintf(stderr,"** perclk1: freq %f\n",Clock_Freq(crm->perclk1)); 	*/
+	/*	fprintf(stderr,"** mpll_clk: freq %f\n",Clock_Freq(crm->mpll_clk)); 	*/
 }
 
 static uint32_t
@@ -784,7 +849,90 @@ static void
 ccsr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
         IMX_Crm *crm = (IMX_Crm *) clientData;
+	uint32_t clko_sel;
 	crm->ccsr = (value & 0x1f) | (3<<8);
+	clko_sel = value & 0x1f;
+	Clock_Decouple(crm->clko);
+	switch(clko_sel) {	
+		case CCSR_CLK0_CLK32:
+			Clock_MakeDerived(crm->clko,crm->osc32,1,1);
+			break;
+
+		case CCSR_CLKO_PREMCLK:
+			Clock_MakeDerived(crm->clko,crm->premclk,1,1);
+			break;
+
+		case CCSR_CLKO_CLK26M:
+			Clock_MakeDerived(crm->clko,crm->osc26m,1,1);
+			break;
+
+		case CCSR_CLKO_MPLLREFCLK:
+			Clock_MakeDerived(crm->clko,crm->mpll_refclk,1,1);
+			break;
+		case CCSR_CLKO_SPLLREFCLK:
+			Clock_MakeDerived(crm->clko,crm->spll_refclk,1,1);
+			break;
+		case CCSR_CLKO_MPLLCLK:
+			Clock_MakeDerived(crm->clko,crm->mpll_clk,1,1);
+			break;
+		case CCSR_CLKO_SPLLCLK:
+			Clock_MakeDerived(crm->clko,crm->spll_clk,1,1);
+			break;
+		case CCSR_CLKO_FCLK:
+			Clock_MakeDerived(crm->clko,crm->fclk,1,1);
+			break;
+		case CCSR_CLKO_HCLK:
+			Clock_MakeDerived(crm->clko,crm->hclk,1,1);
+			break;
+		case CCSR_CLKO_IPG_CLK:
+			Clock_MakeDerived(crm->clko,crm->perclk,1,1);
+			break;
+
+		case CCSR_CLK0_PERCLK1:
+			Clock_MakeDerived(crm->clko,crm->perclk1,1,1);
+			break;
+		case CCSR_CLK0_PERCLK2:
+			Clock_MakeDerived(crm->clko,crm->perclk2,1,1);
+			break;
+		case CCSR_CLK0_PERCLK3:
+			Clock_MakeDerived(crm->clko,crm->perclk3,1,1);
+			break;
+		case CCSR_CLK0_PERCLK4:
+			Clock_MakeDerived(crm->clko,crm->perclk4,1,1);
+			break;
+		case CCSR_CLKO_SSI1BAUD:
+			Clock_MakeDerived(crm->clko,crm->ssi1clk,1,1);
+			break;
+		case CCSR_CLKO_SSI2BAUD:
+			Clock_MakeDerived(crm->clko,crm->ssi2clk,1,1);
+			break;
+		case CCSR_CLKO_NFCBAUD:
+			Clock_MakeDerived(crm->clko,crm->nfcclk,1,1);
+			break;
+		case CCSR_CLKO_FIRI:
+			Clock_MakeDerived(crm->clko,crm->firiclk,1,1);
+			break;
+		
+		case CCSR_CLK0_CLK48MALW:
+			/* What is always ??? */
+			Clock_MakeDerived(crm->clko,crm->clk48m,1,1);
+			break;
+		case CCSR_CLK0_CLK32KALW:
+			/* What is always ??? */
+			Clock_MakeDerived(crm->clko,crm->osc32,1,1);
+			break;
+		case CCSR_CLK0_CLK48M:
+			/* What is not always ??? */
+			Clock_MakeDerived(crm->clko,crm->clk48m,1,1);
+			break;
+
+		case CCSR_CLK0_CLK48DIV_CLKO:
+			Clock_MakeDerived(crm->clko,crm->clk48div_clko,1,1);
+			break;
+		default:
+			fprintf(stderr,"Unknown CLKO_SEL %d\n",clko_sel);
+			break;
+	}
         return;
 }
 
@@ -839,33 +987,42 @@ wkgdctl_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 /*
  * ------------------------------------------------------------------------------
  * System control registers from Chapter 9 
- * SIDR register: manual tells wrong default states of fuse bits:
- * Todo: check with real CPU
+ * SIDR register: Manual is bad. Use mask revision difference pdf. 
+ * Values are taken from imx21ads with 0M55B CPU
  * ------------------------------------------------------------------------------
  */
 static uint32_t
 sidr0_read(void *clientData,uint32_t address,int rqlen)
 {
-	fprintf(stderr,"Warning: SIDR0 is not well documented\n");
-        return 0; 
+	/* GT4 Board version 1 with 1MB55 0x67837a72 */
+	/* GT4 Board version 2 with 0MB55 0xf3476f6b */
+	/* ADS Board version 1 with 0MB55 0xe60558ec */
+        return  0xe60558ec; 
 }
 static uint32_t
 sidr1_read(void *clientData,uint32_t address,int rqlen)
 {
-	fprintf(stderr,"Warning: SIDR1 is not well documented\n");
-        return 0; 
+	/* GT4 Board version 1 with 1MB55 0x00006001 */
+	/* GT4 Board version 2 with 0MB55 0x00006000 */
+	/* ADS Board version 1 with 0MB55 0x0000e000 */
+	
+	return 0x0000e000;
 }
 static uint32_t
 sidr2_read(void *clientData,uint32_t address,int rqlen)
 {
-	fprintf(stderr,"Warning: SIDR2 is not well documented\n");
-        return 0x0000101d; 
+	/* GT4 Board version 1 with 1MB55 0x201d101d */
+	/* GT4 Board version 2 with 0MB55 0x101d101d */
+	/* ADS Board version 1 with 0MB55 0x101d101d */
+        return 0x101d101d; 
 }
 static uint32_t
 sidr3_read(void *clientData,uint32_t address,int rqlen)
 {
-	fprintf(stderr,"Warning: SIDR3 is not well documented\n");
-        return 1<<30; 
+	/* FS3 Board version 1 with 1MB55 0xd0080000 */
+	/* FS3 Board version 2 with 0MB55 0x04008000 */
+	/* ADS Board version 1 with 0MB55 0x04008000 */
+	return 0x40080000;
 }
 
 static void
@@ -941,6 +1098,7 @@ dscr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 		return;
 	}
 	crm->dscr[index] = value & 0x7fff7fff;
+	fprintf(stderr,"DSCR%d: %08x\n",index,value);
         return;
 }
 
@@ -996,10 +1154,10 @@ IMXCrm_Map(void *owner,uint32_t base,uint32_t mask,uint32_t flags)
 	IOH_New32(CRM_WKGDCTL(base),wkgdctl_read,wkgdctl_write,crm);
 
 	/* SIDR: 128 bit register */ 
-	IOH_New32(SYSCTL_SIDR(base),sidr0_read,sidr_write,crm);
-	IOH_New32(SYSCTL_SIDR(base)+4,sidr1_read,sidr_write,crm);
-	IOH_New32(SYSCTL_SIDR(base)+8,sidr2_read,sidr_write,crm);
-	IOH_New32(SYSCTL_SIDR(base)+0xc,sidr3_read,sidr_write,crm);
+	IOH_New32(SYSCTL_SIDR0(base),sidr0_read,sidr_write,crm);
+	IOH_New32(SYSCTL_SIDR1(base),sidr1_read,sidr_write,crm);
+	IOH_New32(SYSCTL_SIDR2(base),sidr2_read,sidr_write,crm);
+	IOH_New32(SYSCTL_SIDR3(base),sidr3_read,sidr_write,crm);
 	IOH_New32(SYSCTL_FMCR(base),fmcr_read,fmcr_write,crm);
 	IOH_New32(SYSCTL_GPCR(base),gpcr_read,gpcr_write,crm);
 	IOH_New32(SYSCTL_WBCR(base),wbcr_read,wbcr_write,crm);
@@ -1036,7 +1194,10 @@ IMXCrm_UnMap(void *owner,uint32_t base,uint32_t mask)
 	IOH_Delete32(CRM_PMCTL(base));
 	IOH_Delete32(CRM_PMCOUNT(base));
 	IOH_Delete32(CRM_WKGDCTL(base));
-	IOH_Delete32(SYSCTL_SIDR(base));
+	IOH_Delete32(SYSCTL_SIDR0(base));
+	IOH_Delete32(SYSCTL_SIDR1(base));
+	IOH_Delete32(SYSCTL_SIDR2(base));
+	IOH_Delete32(SYSCTL_SIDR3(base));
 	IOH_Delete32(SYSCTL_FMCR(base));
 	IOH_Delete32(SYSCTL_GPCR(base));
 	IOH_Delete32(SYSCTL_WBCR(base));
@@ -1066,11 +1227,17 @@ BusDevice *
 IMX21Crm_New(const char *name) 
 {
 	IMX_Crm *crm = (IMX_Crm*)malloc(sizeof(IMX_Crm));
+	uint32_t osc32_freq;
 	if(!crm) {
 		fprintf(stderr,"Out of memory allocating crm\n");
 		exit(1);
 	}
 	memset(crm,0,sizeof(IMX_Crm));
+	crm->name = strdup(name);
+	if(!crm->name) {
+		fprintf(stderr,"Out of memory for CRM name\n");
+		exit(1);
+	}
 	crm->cscr = 0x77180607;
 	crm->mpctl0 = 0x00071c07;
 	crm->mpctl1 = 0x00008000;
@@ -1149,8 +1316,11 @@ IMX21Crm_New(const char *name)
 	crm->rtic_perclk = Clock_New("%s.rtic_perclk",name);
 	crm->rnga_perclk = Clock_New("%s.rnga_perclk",name);
 
-	Clock_Set(crm->osc32,32768);
-//	Clock_Set(crm->osc26m,26*1000*1000); //in cscr
+	/* The 32k Osc is always running, the 26MHz is controlled by CSCR */
+	if(Config_ReadUInt32(&osc32_freq,crm->name,"osc32")<0) {
+		osc32_freq = 32768;
+	}
+	Clock_SetFreq(crm->osc32,osc32_freq);
 
 	crm->fmcr = 0xffffffcb;
 	crm->gpcr = 0x0000000c;
@@ -1163,8 +1333,16 @@ IMX21Crm_New(const char *name)
         crm->bdev.owner=crm;
         crm->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
 
+	cscr_write(crm,crm->cscr,0,4);
+	pccr0_write(crm,crm->pccr0,0,4);
+	pccr1_write(crm,crm->pccr1,0,4);
+	pcdr0_write(crm,crm->pcdr0,0,4);
+	pcdr1_write(crm,crm->pcdr1,0,4);
+	ccsr_write(crm,crm->ccsr,0,4);
+
         fprintf(stderr,"IMX21 Clock and Reset module \"%s\" created\n",name);
 
-	CycleTimer_Add(&crm->dbgtimer,285000000,dump_proc,crm);
+	//CycleTimer_Add(&crm->dbgtimer,(uint64_t)285000000*10,dump_proc,crm);
+	CycleTimer_Add(&crm->dbgtimer,(uint64_t)3,dump_proc,crm);
         return &crm->bdev;
 }
