@@ -34,6 +34,13 @@
 #include <cycletimer.h>
 #include <clock.h>
 
+#if 0
+#define dprintf(x...) { fprintf(stderr,x); }
+#else
+#define dprintf(x...)
+#endif
+
+
 #define CSPI_RXDATA_REG(base) 	((base)+0)
 #define CSPI_TXDATA_REG(base) 	((base)+0x4)
 #define CSPI_CTRL_REG(base) 	((base)+0x8)
@@ -199,8 +206,8 @@ update_spiclk(IMX21Cspi *cspi)
 {
 	int datarate = (cspi->controlreg & CSPI_CTRL_DATARATE_MASK) >> CSPI_CTRL_DATARATE_SHIFT;
 	int divider;
-	if(datarate > 0x12) {
-		fprintf(stderr,"i.MX21 CSPI: Illegal data rate in control register\n");
+	if((datarate == 0) || (datarate > 0x12)) {
+		fprintf(stderr,"i.MX21 CSPI: Illegal data rate %d in control register\n",datarate);
 	}
 	if((datarate & 1)) {
 		divider  = 3 * (1<<((datarate-1)/2));
@@ -225,7 +232,7 @@ rxdata_read(void *clientData,uint32_t address,int rqlen)
 	int rxcount;
 	uint32_t intreg;
 	uint32_t dmareg;
-	fprintf(stderr,"RXDATA read 0x%08x\n",value);
+	dprintf("RXDATA read 0x%08x\n",value);
 	if(cspi->rxfifo_rp < cspi->rxfifo_wp) {
 		cspi->rxfifo_rp++;
 		rxcount = RXFIFO_COUNT(cspi); 
@@ -413,7 +420,7 @@ rxfifo_put(IMX21Cspi *cspi)
 	if(rxcount < RXFIFO_SIZE) {
 		cspi->rxfifo[cspi->rxfifo_wp % RXFIFO_SIZE] = cspi->in_shiftreg & mask;
 		cspi->rxfifo_wp++; rxcount ++;
-		fprintf(stderr,"rxfifo_put %08x\n", cspi->in_shiftreg & mask);
+		dprintf("rxfifo_put %08x\n", cspi->in_shiftreg & mask);
 		intreg = cspi->intreg;
 		dmareg = cspi->dmareg;
 		intreg |= CSPI_INT_RR; /* receive ready */
@@ -548,7 +555,13 @@ eval_one_cmd(IMX21Cspi *cspi)
 
 		case CMD_DCLK_DELAY: {
 			/* setup timer */
-			uint32_t nsecs = 1000; /* should be dynamic */	
+			uint32_t spihz = Clock_Freq(cspi->spiclk);	
+			uint32_t nsecs;
+			if(spihz) {
+				nsecs  = 1000000000 / spihz;
+			} else {
+				nsecs = 1; 
+			}
 			CycleCounter_t cycles = NanosecondsToCycles(nsecs);
 			CycleTimer_Mod(&cspi->nDelayTimer,cycles);
 			return PGM_SLEEP;
@@ -563,7 +576,7 @@ eval_one_cmd(IMX21Cspi *cspi)
 				/* CLK 32.768 khz */
 				nsecs = (cspi->periodreg & 0x7fff) * 30517;
 			} else {
-				uint64_t hz = Clock_Value(cspi->spiclk);
+				uint64_t hz = Clock_Freq(cspi->spiclk);
 				if(hz) {
 					nsecs = (uint64_t)1000000000 
 						* (uint64_t)(cspi->periodreg & 0x7fff) / hz;
@@ -729,10 +742,10 @@ run_programm(void *clientData)
 		if(result == PGM_DONE) {
 			/* Try to refill the shiftreg */
 			if(txfifo_to_shiftreg(cspi,1) <= 0) {
-				fprintf(stderr,"Nothing to refill\n");
+				dprintf("Nothing to refill\n");
 				break;
 			}
-			fprintf(stderr,"Refilled\n");
+			dprintf("Refilled\n");
 		} else if (result == PGM_SLEEP) {
 			break;
 		}
@@ -774,10 +787,10 @@ ctrl_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	uint32_t diff;
 	diff = cspi->controlreg ^ value;
 	cspi->controlreg = (cspi->controlreg & CSPI_CTRL_XCH) | (value & 0x00ffffff);
-	fprintf(stderr,"%s: Controlreg write %08x, modified to %08x\n",cspi->name,value,cspi->controlreg);
+	dprintf("%s: Controlreg write %08x, modified to %08x\n",cspi->name,value,cspi->controlreg);
 	update_spiclk(cspi);
 	if(diff & value  & CSPI_CTRL_XCH) {
-		fprintf(stderr,"Start XCH\n");
+		dprintf("Start XCH\n");
 		txfifo_to_shiftreg(cspi,0);
 		run_programm(cspi);
 	}
@@ -978,7 +991,7 @@ IMX21_CspiNew(const char *name)
         cspi->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
 	cspi->clk = Clock_New("%s.clk",name);
 	cspi->spiclk = Clock_New("%s.spiclk",name);
-	Clock_Set(cspi->clk,33250006); 	/* Should be connected to perclk2 and not fixed here */
+	Clock_SetFreq(cspi->clk,33250006); 	/* Should be connected to perclk2 and not fixed here */
 	reset_cspi(cspi);
 	fprintf(stderr,"i.MX21 CSPI module \"%s\" created\n",name);
 	return &cspi->bdev;
