@@ -144,7 +144,7 @@ typedef struct DMAChan {
 	int nr; 	/* my own number */
 	SigNode * irqNode;
 	SigNode *currReqLine;
-	uint8_t xfer_buf[4];
+	uint8_t xfer_buf[64]; 
 	unsigned int xferbuf_rp;
 	unsigned int xferbuf_wp;
 	SigTrace *dmaReqTrace;
@@ -168,6 +168,7 @@ struct IMX21Dmac {
 	BusDevice bdev;
 	DMAChan dmachan[16];
 	SigNode *dmaReqNode[32];
+	SigNode *dummyHighNode;
 	uint16_t dcr;
 	uint16_t disr;
 	uint16_t dimr;
@@ -207,6 +208,8 @@ dmac_reset(IMX21Dmac *dmac)
 		chan->cntr = 0;
 		chan->ccr = 0;
 		chan->rssr = 0;
+		/* Should untrace line and stop dma also */
+		chan->currReqLine = dmac->dummyHighNode;
 		chan->blr = 0;
 		chan->rtor = 0;
 		chan->bucr = 0;
@@ -271,14 +274,14 @@ dma_read(DMAChan *chan,int smod,int ssiz)
 	}
 	switch(ssiz) {
 		case 8:
-			if(chan->xferbuf_wp > 3) {
+			if(chan->xferbuf_wp > 63) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
 			chan->xfer_buf[chan->xferbuf_wp++] = Bus_Read8(addr);
 			break;
 		case 16:
-			if(chan->xferbuf_wp > 2) {
+			if(chan->xferbuf_wp > 62) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
@@ -287,7 +290,7 @@ dma_read(DMAChan *chan,int smod,int ssiz)
 			chan->xfer_buf[chan->xferbuf_wp++] = (value>>8) & 0xff;
 			break;
 		case 32:
-			if(chan->xferbuf_wp > 0) {
+			if(chan->xferbuf_wp > 60) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
@@ -337,14 +340,14 @@ dma_write(DMAChan *chan,int dmod,int dsiz)
 	} 
 	switch(dsiz) {
 		case 8:
-			if(chan->xferbuf_rp > 3) {
+			if(chan->xferbuf_rp > 63) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
 			Bus_Write8(chan->xfer_buf[chan->xferbuf_rp++],addr);
 			break;
 		case 16:
-			if(chan->xferbuf_rp > 2) {
+			if(chan->xferbuf_rp > 62) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
@@ -353,7 +356,7 @@ dma_write(DMAChan *chan,int dmod,int dsiz)
 			Bus_Write16(value,addr);
 			break;
 		case 32:
-			if(chan->xferbuf_rp > 0) {
+			if(chan->xferbuf_rp > 60) {
 				fprintf(stderr,"Emulator bug in %s line %d\n",__FILE__,__LINE__);
 				return;
 			}
@@ -375,6 +378,7 @@ do_dma(DMAChan *chan)
 	uint32_t smod,dmod,dsiz,ssiz;
 	chan->stop_dma = 0;
 	int count = 0;
+	//int burst_len = chan->blr & 0x3f; 
 	dprintf("Entered DO-DMA\n");
 	smod = (chan->ccr & CCR_SMOD_MASK);
 	dmod = (chan->ccr & CCR_DMOD_MASK);
@@ -411,10 +415,7 @@ do_dma(DMAChan *chan)
 		fprintf(stderr,"i.MX21 DMAC: downwards DMA direction not implemented\n");
 		return;
 	}
-	if(!chan->currReqLine) {
-		return;
-	}
-	while((chan->ccnr < chan->cntr) && (SigNode_Val(chan->currReqLine) == SIG_LOW)) {
+	while((chan->ccnr < chan->cntr) && ((SigNode_Val(chan->currReqLine) == SIG_LOW) || !(chan->ccr & CCR_REN))) {
 		chan->xferbuf_wp = 0;
 		chan->xferbuf_rp = 0;
 		do { 
@@ -438,13 +439,9 @@ static void
 check_dma(DMAChan *chan) 
 {
 	int sigval;
-	if(!chan->currReqLine) {
-		return;
-	} else {
-		sigval = SigNode_Val(chan->currReqLine);
-		if((sigval == SIG_LOW) ||(sigval == SIG_PULLDOWN)) {
-			do_dma(chan);
-		}
+	sigval = SigNode_Val(chan->currReqLine);
+	if(sigval == SIG_LOW) {
+		do_dma(chan);
 	}
 }
 
@@ -453,7 +450,7 @@ change_dmareq(SigNode *node,int value,void *clientData)
 {
 	DMAChan *chan = (DMAChan *) clientData;
 	dprintf("DMAC: request line changed to state %d\n",value);
-	if((value == SIG_LOW) || (value == SIG_PULLDOWN)) {
+	if(value == SIG_LOW) {
 		do_dma(chan);
 	} else {
 		chan->stop_dma = 1;
@@ -473,7 +470,8 @@ disable_dmareq(DMAChan *chan)
 static void
 enable_dmareq(DMAChan *chan) 
 {
-	if(chan->currReqLine) {
+	IMX21Dmac *dmac = chan->dmac;
+	if(chan->currReqLine != dmac->dummyHighNode) {
 		dprintf("DMAC: Tracing dma request line\n");
 		chan->dmaReqTrace = SigNode_Trace(chan->currReqLine,change_dmareq,chan);
 		check_dma(chan);
@@ -828,8 +826,18 @@ ccr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 		chan->ccnr = 0;
 	} 
 	if(diff & (CCR_CEN | CCR_REN)) { 
+		/* 
+		 * ---------------------------------------
+		 * Check for dma_req line triggered DMA 
+		 * ---------------------------------------
+		 */
 		update_dmareq(chan);
 	}
+	/* 
+	 * -----------------------------------------------
+ 	 * Check for Manually triggered DMA 
+	 * -----------------------------------------------
+	 */
 	if((value & CCR_CEN) && !(value & CCR_REN)) {
 		do_dma(chan);
 	}
@@ -853,12 +861,11 @@ rssr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 		chan->rssr = value;
 		disable_dmareq(chan);
 		if(value == 0) {
-			chan->currReqLine = 0; 
+			chan->currReqLine = dmac->dummyHighNode; 
 		} else {	
 			chan->currReqLine = dmac->dmaReqNode[value];
 		}
 		update_dmareq(chan);
-		
 	}
         return;
 }
@@ -993,17 +1000,11 @@ IMX21_DmacNew(const char *name) {
                 exit(1);
         }
         memset(dmac,0,sizeof(IMX21Dmac));
-	for(i=0;i<16;i++) {
-		DMAChan *chan = &dmac->dmachan[i];
-		chan->dmac = dmac;
-		chan->nr = i;
-		chan->irqNode = SigNode_New("%s.irq%d",name,i);
-		if(!chan->irqNode) {
-			fprintf(stderr,"i.MX21 DMA controller: can not create irq node %d\n",i);
-			exit(1);
-		}
-		SigNode_Set(chan->irqNode,SIG_PULLUP);
-	}
+	dmac->dummyHighNode = SigNode_New("%s.high",name);
+	if(!dmac->dummyHighNode) {
+		fprintf(stderr,"Can not create dummy high node for DMAC\n");
+	}	
+	SigNode_Set(dmac->dummyHighNode,SIG_HIGH);
 	for(i=0;i<32;i++) {
 		dmac->dmaReqNode[i] = SigNode_New("%s.dma_req%d",name,i);
 		if(!dmac->dmaReqNode[i]) {
@@ -1011,6 +1012,18 @@ IMX21_DmacNew(const char *name) {
 			exit(1);
 		}
 	}
+	for(i=0;i<16;i++) {
+		DMAChan *chan = &dmac->dmachan[i];
+		chan->dmac = dmac;
+		chan->nr = i;
+		chan->irqNode = SigNode_New("%s.irq%d",name,i);
+		chan->currReqLine = dmac->dummyHighNode;
+		if(!chan->irqNode) {
+			fprintf(stderr,"i.MX21 DMA controller: can not create irq node %d\n",i);
+			exit(1);
+		}
+		SigNode_Set(chan->irqNode,SIG_PULLUP);
+	}
         dmac->bdev.first_mapping=NULL;
         dmac->bdev.Map=IMXDmac_Map;
         dmac->bdev.UnMap=IMXDmac_Unmap;
