@@ -9,6 +9,11 @@
  *      Primary and alternate functions not implemented 
  *	Interrupt on positive edge is tested 
  *
+ *	Docu does't tell how input from pad to Primary/Alternate function
+ *	is connected. I suspect that it is always connected, no matter
+ *	which mode. For example RTS Change interrupt on PB26 works
+ *	even if in gpio mode
+ *
  *  This program is free software; you can distribute it and/or modify it
  *  under the terms of the GNU General Public License (Version 2) as
  *  published by the Free Software Foundation.
@@ -289,7 +294,7 @@ PinInfo pinTable[] = {
 	{5,0,"gpioF.0","imx21.m12"},
 	{5,1,"gpioF.1","imx21.t15"},
 	{5,2,"gpioF.2","imx21.l12"},
-	{5,3,"gpioF.3","imx21.u16"},
+	{5,3,"gpioF.3","imx21.u15"},
 	{5,4,"gpioF.4","imx21.u14"},
 	{5,5,"gpioF.5","imx21.w13"},
 	{5,6,"gpioF.6","imx21.t13"},
@@ -320,6 +325,21 @@ PinInfo pinTable[] = {
 	{5,31,"gpioF.31",NULL},
 	{-1,-1,NULL,NULL}
 };
+
+#include <arm9cpu.h>
+static inline void
+check_interrupts() {
+#if 1
+        if(!(REG_CPSR & FLAG_I)) {
+                static int count=0;
+                if(count<10) {
+                        count++;
+                        fprintf(stderr,"Emulator: Warning, interrupts not disabled: possible nonatomic gpio write\n");
+                }
+        }
+#endif
+}
+
 /*
  * ------------------------------------------------------
  * return the node value (SIG_LOW or SIG_HIGH)
@@ -485,6 +505,9 @@ update_icr(GpioPort *port)
 	int i;
 	int ic;
 	for(i=0;i<32;i++) {
+		if(!port->cpuPinNode[i]) {
+			continue;
+		}
 		ic = (icr >> (i<<1)) & 3;
 		int interrupt = -1;
 		switch(ic) {
@@ -492,13 +515,13 @@ update_icr(GpioPort *port)
 			case 1: /* no falling edge on icr change */
 				break;
 			case 2: /* high level */
-				value =  SigNode_Val(port->gpioNode[i]);
+				value =  SigNode_Val(port->cpuPinNode[i]);
 				if((value == SIG_HIGH) || (value == SIG_PULLUP)) {
 					interrupt = 1;
 				}	
 				break;
 			case 3: /* low level */
-				value =  SigNode_Val(port->gpioNode[i]);
+				value =  SigNode_Val(port->cpuPinNode[i]);
 				if((value == SIG_LOW) || (value == SIG_PULLDOWN)) {
 					interrupt = 1;
 				}	
@@ -519,7 +542,7 @@ reset_port(GpioPort *port)
 	port->iconfa1 = port->iconfa2 = 0xffffffff;
 	port->iconfb1 = port->iconfb2 = 0xffffffff;
 	port->dr = 0;	
-	//port->gius = gius_reset /* fetch from configfile */
+	port->gius = 0; /* Should be fetched from configfile */
 	port->ssr = 0;
 	port->icr1 = port->icr2 = 0;
 	port->imr = 0;
@@ -566,6 +589,7 @@ ocr1_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->ocr1 = value;
+	check_interrupts(); 
 	update_gpioouts(port);
         return;
 }
@@ -582,6 +606,7 @@ ocr2_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->ocr2 = value;
+	check_interrupts(); 
 	update_gpioouts(port);
         return;
 }
@@ -604,6 +629,7 @@ iconfa1_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->iconfa1 = value;
+	check_interrupts(); 
 	update_aouts(port);
         return;
 }
@@ -620,6 +646,7 @@ iconfa2_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->iconfa2 = value;
+	check_interrupts(); 
 	update_aouts(port);
         return;
 }
@@ -641,6 +668,7 @@ iconfb1_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->iconfb1 = value;
+	check_interrupts(); 
 	update_bouts(port);
         return;
 }
@@ -657,6 +685,7 @@ iconfb2_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->iconfb2 = value;
+	check_interrupts(); 
 	update_bouts(port);
         return;
 }
@@ -680,6 +709,7 @@ dr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->dr = value;
+	check_interrupts(); 
 	update_gpioouts(port);
         return;
 }
@@ -706,10 +736,14 @@ gius_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	uint32_t diff = port->gius ^ value;
 	int i;
 	port->gius = value;
+	check_interrupts(); 
 	for(i=0;i<32;i++) {
 		if(!(diff & (1<<i))) {
 			continue;
 		}
+		if(!port->cpuPinNode[i]) {
+			continue;
+		}
 		if(value & (1<<i)) {
 			SigNode_Link(port->cpuPinNode[i],port->gpioNode[i]);
 		} else {
@@ -771,6 +805,7 @@ icr1_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->icr1 = value;
+	check_interrupts(); 
 	update_icr(port);
         return;
 }
@@ -787,6 +822,7 @@ icr2_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
 	port->icr2 = value;
+	check_interrupts(); 
 	update_icr(port);
         return;
 }
@@ -808,8 +844,11 @@ static void
 imr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	GpioPort *port = (GpioPort *) clientData;
+	uint32_t diff = port->imr ^ value;
 	port->imr = value;
-	update_interrupt(port->gpio);
+	if(diff) {
+		update_interrupt(port->gpio);
+	}
         return;
 }
 
@@ -903,6 +942,7 @@ puen_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	GpioPort *port = (GpioPort *) clientData;
 	port->puen = value;
 	update_gpioouts(port);
+	//fprintf(stderr,"PUEN%c %08x\n",'a'+port->index,value);
         return;
 }
 
@@ -1001,7 +1041,7 @@ gin_trace_proc(struct SigNode * node,int value, void *clientData)
        	GpioPort *port = ti->port;
 	int index = ti->index;
 	uint64_t icr = ((uint64_t)port->icr2 <<32) | port->icr1;	
-	int ic = icr >> (index<<1);
+	int ic = (icr >> (index<<1)) & 3;
 	switch(ic) {
 		case 0: /* rising */
 		case 2: /* high */
@@ -1021,6 +1061,7 @@ gin_trace_proc(struct SigNode * node,int value, void *clientData)
 	return 0;
 }
 
+
 static PinInfo * 
 find_pininfo(int portindex,int bit) 
 {
@@ -1057,6 +1098,8 @@ create_port(GpioPort *port,int index,const char *modulename)
 			fprintf(stderr,"GPIO pin %d of port %d is missing in table\n",i,index);
 			exit(1);
 		}
+		ti->port = port;
+		ti->index = i;
 		if(pin_info->pin_name) {
 			port->cpuPinNode[i] = SigNode_New(pin_info->pin_name);
 			port->ginTrace[i] = SigNode_Trace(port->cpuPinNode[i],gin_trace_proc,ti);
@@ -1067,11 +1110,16 @@ create_port(GpioPort *port,int index,const char *modulename)
 		port->cinNode[i] = SigNode_New("%s.cin%d",portname,i);
 		port->aoutNode[i] = SigNode_New("%s.aout%d",portname,i);
 		port->boutNode[i] = SigNode_New("%s.bout%d",portname,i);
-		ti->port = port;
-		ti->index = i;
 	}
 }
 
+UNUSED static int
+pin_trace(struct SigNode * node,int value, void *clientData)
+{
+	fprintf(stderr,"Pin changed to %d\n",value);
+	return 0;
+}
+
 BusDevice *
 IMX21_GpioNew(const char *modulename)
 {
@@ -1105,6 +1153,7 @@ IMX21_GpioNew(const char *modulename)
         gpio->bdev.owner=gpio;
         gpio->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
         update_interrupt(gpio);
+//	SigNode_Trace(gpio->port[2].cpuPinNode[11],pin_trace,NULL);
         fprintf(stderr,"IMX21 GPIO module \"%s\" created\n",modulename);
         return &gpio->bdev;
 }
