@@ -207,7 +207,7 @@ static int
 execute_instruction(IMX21I2c *i2c) {
         uint32_t icode;
         if(i2c->ip >= CODE_MEM_SIZE) {
-                fprintf(stderr,"NS9xxx I2C: corrupt I2C script\n");
+                fprintf(stderr,"i.MX21 I2C: corrupt I2C script\n");
                 return RET_INTERP_ERROR;
         }
         icode = i2c->code[i2c->ip++];
@@ -323,7 +323,7 @@ execute_instruction(IMX21I2c *i2c) {
                         break;
 
                 default:
-                        fprintf(stderr,"NS9xxx I2C: Unknode instruction code %08x\n",icode);
+                        fprintf(stderr,"i.MX21 I2C: Unknode instruction code %08x\n",icode);
                         return RET_INTERP_ERROR;
                         break;
 
