@@ -4,7 +4,7 @@
  * (C) 2006 Jochen Karrer
  *   Author: Jochen Karrer
  *
- * state: nothing works
+ * state: incomplete but basically working 
  *
  *  This program is free software; you can distribute it and/or modify it
  *  under the terms of the GNU General Public License (Version 2) as
@@ -38,6 +38,13 @@
 #include <configfile.h>
 #include <cycletimer.h>
 #include <fbdisplay.h>
+#include <clock.h>
+
+#if 0
+#define dbgprintf(x...) { fprintf(stderr,x); }
+#else
+#define dbgprintf(x...)
+#endif
 
 #define LCDC_BASE_ADDR  0x10021000
 #define LCDC_LSSAR(base)   ((base) + 0x00)  //  32bit lcdc screen start addr reg
@@ -220,6 +227,7 @@ typedef struct IMXLcdc {
 	BusDevice bdev;
 	RfbServer *rfbserv;
 	int interrupt_posted;
+	Clock_t *clk;	
         SigNode *irqNode;
 	uint32_t lssar;
 	uint32_t lsr;
@@ -251,7 +259,6 @@ typedef struct IMXLcdc {
 
 	CycleTimer updateTimer;
 	int page_size;
-	//int traced_pages; 
 	uint32_t dirty_addr[UPDATE_FIFO_SIZE];
 	uint64_t update_fifo_wp;
 	uint64_t update_fifo_rp;
@@ -291,14 +298,28 @@ update_fbformat(IMXLcdc *lcdc)
 	FbFormat fbf;
 	uint32_t bpix = lcdc->lpcr &  LPCR_BPIX_MASK;
 	FbDisplay *display = lcdc->display;
+	int endsel = lcdc->lpcr & LPCR_END_SEL; 
+	if(endsel)
+		fprintf(stderr,"LCDC warning: big endian mode not implemented\n");
 	switch(bpix) {
 		case LPCR_BPIX_1:
 		case LPCR_BPIX_2:
 		case LPCR_BPIX_4:
 		case LPCR_BPIX_8:
-		case LPCR_BPIX_18:
 			fprintf(stderr,"LCDC Warn: unsupported bits per pixel\n");
 			return;
+
+		case LPCR_BPIX_18:
+			fbf.red_bits = 6;
+			fbf.red_shift = 2;
+			fbf.green_bits = 6;
+			fbf.green_shift = 10;
+			fbf.blue_bits = 6;
+			fbf.blue_shift = 18;
+			fbf.bits_per_pixel = 32;
+			fbf.depth = 18;
+			break;
+
 		case LPCR_BPIX_12: 
 			fbf.red_bits = 4; 
 			fbf.green_bits = 4; 
@@ -306,7 +327,10 @@ update_fbformat(IMXLcdc *lcdc)
 			fbf.red_shift = 8;
 			fbf.green_shift = 4;
 			fbf.blue_shift = 0;
+			fbf.bits_per_pixel = 16;
+			fbf.depth = 12;
 			break;
+
 		case LPCR_BPIX_16:
 			fbf.red_bits = 5; 
 			fbf.green_bits = 6; 
@@ -314,6 +338,8 @@ update_fbformat(IMXLcdc *lcdc)
 			fbf.red_shift = 11;
 			fbf.green_shift = 5;
 			fbf.blue_shift = 0;
+			fbf.bits_per_pixel = 16;
+			fbf.depth = 16;
 			break;
 			
 	}
@@ -340,25 +366,38 @@ update_memory_traces(IMXLcdc *lcdc)
 {
 	uint32_t start,end,length;
 	int vpw,height;
-	if(lcdc->trace_length) {
-		IOH_DeleteRegion(lcdc->trace_start,lcdc->trace_length) ;
-		lcdc->trace_length = 0;
-	}
-	if(!lcdc->display) {
-		return;
-	}
+	int i;
+
 	vpw = lcdc->lvpwr & LVPWR_VPW_MASK; 
 	height = ((lcdc->lsr & LSR_YMAX_MASK)); 
 	start = lcdc->lssar;
 	length = (vpw << 2) * height;
 	end = start + length - 1;	
+	if((lcdc->trace_length == length) && (lcdc->trace_start == start)) {
+		return;
+	}
+	if(lcdc->trace_length) {
+		IOH_DeleteRegion(lcdc->trace_start,lcdc->trace_length) ;
+		Mem_UntraceRegion(lcdc->trace_start,lcdc->trace_length);
+	}
+	if(!lcdc->display) {
+		return;
+	}
+	/* Clear all outstanding updates */
+	lcdc->update_fifo_rp=lcdc->update_fifo_wp;
+
 	if((start >= 0xc0000000) && (end <=  0xc7ffffff) && (length > 0)) {
-		fprintf(stderr,"updating traced page list %08x to %08x\n",start,end);
+		dbgprintf("updating traced page list %08x to %08x\n",start,end);
 		IOH_NewRegion(start,length,NULL,lcdc_mem_trace,0,lcdc);
-		//lcdc->traced_pages = (length+lcdc->page_size-1)/lcdc->page_size;
 		lcdc->trace_length = length;
 		lcdc->trace_start = start;
-		Mem_TraceRegion(start,length);
+		/* 
+		 * Trigger the trace a first time because all pages are dirty 
+		 * after address change 
+		 */
+		for(i=0;i<length;i+=lcdc->page_size) {
+			lcdc_mem_trace(lcdc,0,i+start,0);	
+		}
 	}
 }
 
@@ -469,7 +508,7 @@ lcpr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMXLcdc *lcdc = (IMXLcdc *) clientData;
 	lcdc->lcpr = value & 0xd7ff03ff;;
-	fprintf(stderr,"LCDC: LCPR Cursor position register write has no effect\n");
+	dbgprintf("LCDC: LCPR Cursor position register write has no effect\n");
 }
 
 /*
@@ -623,7 +662,7 @@ lvcr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMXLcdc *lcdc = (IMXLcdc *) clientData;
 	lcdc->lvcr = value & 0xfc00ffff;
-	fprintf(stderr,"LCDC: LVCR write has no effect\n");
+	dbgprintf("LCDC: LVCR write has no effect\n");
 }
 
 /*
@@ -719,7 +758,7 @@ static void
 ldcr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMXLcdc *lcdc = (IMXLcdc *) clientData;
-	lcdc->ldcr = 0x801f001f;
+	lcdc->ldcr = value & 0x800f000f;
 }
 
 /*
@@ -1076,12 +1115,14 @@ IMX21_LcdcNew(const char *name,FbDisplay *display)
 	}
 	lcdc->display = display;
 	lcdc->page_size = Mem_SmallPageSize();
+	lcdc->ldcr = 0x80080004;
 	CycleTimer_Init(&lcdc->updateTimer,update_display,lcdc);
 	lcdc->bdev.first_mapping=NULL;
         lcdc->bdev.Map=IMXLcdc_Map;
         lcdc->bdev.UnMap=IMXLcdc_UnMap;
         lcdc->bdev.owner=lcdc;
         lcdc->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
+	lcdc->clk = Clock_New("%s.clk",name);
         fprintf(stderr,"i.MX21 LCD-Controller \"%s\" created\n",name);
         return &lcdc->bdev;
 }
