@@ -36,12 +36,14 @@
 #include <configfile.h>
 #include <mmcard.h>
 #include <cycletimer.h>
+#include <clock.h>
 
 #if 0
 #define dprintf(x...) { fprintf(stderr,x); }
 #else
 #define dprintf(x...)
 #endif
+#define NEW
 
 
 #define UDELAY_CMD    (20)
@@ -60,13 +62,16 @@
 #define 	STATUS_SDIO_INT_ACTIVE          (1<<14)
 #define 	STATUS_END_CMD_RESP             (1<<13)
 #define 	STATUS_WRITE_OP_DONE            (1<<12)
+
 #define 	STATUS_READ_OP_DONE          	(1<<11) 
 #define 	STATUS_WR_CRC_ERROR_CODE_MASK   (3<<10)
 #define 	STATUS_WR_CRC_ERROR_CODE_SHIFT  (10)
 #define 	STATUS_CARD_BUS_CLK_RUN         (1<<8)
+
 #define 	STATUS_APPL_BUFF_FF             (1<<7)
 #define 	STATUS_APPL_BUFF_FE             (1<<6)
 #define 	STATUS_RESP_CRC_ERR             (1<<5)
+
 #define 	STATUS_CRC_READ_ERR             (1<<3)
 #define 	STATUS_CRC_WRITE_ERR            (1<<2)
 #define 	STATUS_TIME_OUT_RESP            (1<<1)
@@ -128,7 +133,10 @@
 #define MAX_BUF_FIFO_SIZE	(32)
 typedef struct IMX21Sdhc {
 	BusDevice bdev;
+	char *name;
 	MMCard *card;
+	Clock_t *inclk;	/* Input clock from Clock / PLL module */ 
+	Clock_t *sdclk;		/* Output clock from Clock to SD-Card  */ 
 	CycleTimer cmd_delay_timer;
 	CycleTimer write_delay_timer;
 	CycleTimer data_delay_timer;
@@ -145,6 +153,7 @@ typedef struct IMX21Sdhc {
 	uint16_t int_cntl;
 	uint16_t int_status;  /* Internal only, same bits as int_cntl */
 	uint16_t cmd;
+	uint16_t prev_cmd;    /* For debugging only ! */
 	uint16_t argh;
 	uint16_t argl;
 	uint32_t res_fifo[RES_FIFO_SIZE];
@@ -155,9 +164,48 @@ typedef struct IMX21Sdhc {
 	uint64_t buf_fifo_wp;
 	uint64_t buf_fifo_rp;
 	SigNode *irqNode;
-	SigNode *dmaReq;
+	SigNode *nDmaReq;
 } IMX21Sdhc;
 
+static unsigned int
+onecount(const uint32_t value)
+{
+        uint32_t val=value;
+        int ones=0;
+        while(val) {
+                if(val&1)
+                        ones++;
+                val>>=1;
+        }
+        return ones;
+}
+
+static uint32_t 
+update_clock(IMX21Sdhc *sdhc) 
+{
+	uint32_t clk_divider = sdhc->clk_rate & 0xf;
+	uint32_t clk_prescaler = (sdhc->clk_rate >> 4) & 0xfff;
+	uint32_t divider;
+	uint32_t prescaler;
+	uint32_t freq;
+	if((onecount(clk_prescaler) > 1)) {
+		fprintf(stderr,"%s: Illegal Clock Prescaler 0x%03x\n",sdhc->name,clk_prescaler);
+	}
+	if(clk_divider == 0) {
+		fprintf(stderr,"%s: Illegal Clock divider: 0\n",sdhc->name);
+	}
+	if(clk_prescaler) {
+		prescaler = (clk_prescaler * 2);
+	} else {
+		prescaler = 1;
+	}
+	divider = (clk_divider+1) * (prescaler);
+	Clock_MakeDerived(sdhc->sdclk,sdhc->inclk,1,divider);
+	freq = Clock_Freq(sdhc->sdclk);
+	//fprintf(stderr,"%s: DIVIDER %d value 0x%08x rate %d\n",sdhc->name,divider,sdhc->clk_rate,  freq);
+	return freq;
+}
+
 static void
 update_interrupt(IMX21Sdhc *sdhc) 
 {
@@ -243,6 +291,15 @@ sdhc_reset(IMX21Sdhc *sdhc)
 	sdhc->argl = 0;
 }
 
+/*
+ * ---------------------------------------------------------------------------------------
+ * Clock Control Register
+ * Bit 3:	Reset the MMC/SD host controller
+ * Bit 1:	Start Clock immediately or delayed, depending on transmission state
+ * Bit 0:	Stop Clock immediately or delayed, depending on transmission state
+ *
+ * ---------------------------------------------------------------------------------------
+ */
 static uint32_t
 str_stp_clk_read(void *clientData,uint32_t address,int rqlen)
 {
@@ -260,21 +317,52 @@ str_stp_clk_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	}
 	if(value & STR_STP_CLK_START_CLK) {
 		sdhc->status |= STATUS_CARD_BUS_CLK_RUN; 
-	}
-	if(value & STR_STP_CLK_STOP_CLK) {
+	} else if(value & STR_STP_CLK_STOP_CLK) {
+		/* Clock run should be disabled delayed (after transmission) */
 		sdhc->status &= ~(STATUS_CARD_BUS_CLK_RUN | STATUS_TIME_OUT_READ 
 		  | STATUS_TIME_OUT_RESP | STATUS_CRC_WRITE_ERR | STATUS_CRC_READ_ERR 
 		  | STATUS_RESP_CRC_ERR | STATUS_READ_OP_DONE | STATUS_WRITE_OP_DONE 
 		  | STATUS_END_CMD_RESP);
+	} else if((value & STR_STP_CLK_START_CLK) && (value & STR_STP_CLK_STOP_CLK)) {
+		fprintf(stderr,"%s: START and STOP clock at same time is prohibited\n",sdhc->name); 
 	}
         return;
 }
 
+/*
+ * ---------------------------------------------------------------------------------------------
+ * Status Register
+ *
+ * Bit 15: Card Presence: Warning: seems to work only outside transmission periods
+ *         because DAT[3] is used during transmission 
+ * Bit 14: SDIO_INT_ACTIVE Indicates if SDIO card sends interrupt condition. (w1c)
+ * Bit 13: END_CMD_RESP Indicates that cmd is transmitted to card and a response or timeout
+ *         was received. Check RESP_CRC_ERR and TIME_OUT_RESP ! (w1c)
+ * Bit 12: WRITE_OP_DONE: A write operation is finished (including writting card buffer to flash)
+ *	   Check WRITE_CRC_ERR also ! (w1 or stop mmcclk to clear)
+ * Bit 11: READ_OP_DONE: Read operation is finished. Check READ_CRC_ERR and TIME_OUT_READ 
+ *         (w1 or stop clock to clear)
+ * Bit 9,10: WR_CRC_ERROR_CODE: 00: good, 01: Transm. error 10: No crc 11: reserved
+ * Bit 8:  CARD_BUS_CLK_RUN: Indicates if the clock is running
+ * Bit 7:  APPL_BUFF_FF Data fifo is full
+ * Bit 6:  APPL_BUFF_FE Data fifo is empty
+ * Bit 5:  RESP_CRC_ERROR: 1=crc error in response fount (w1c or stop)
+ * Bit 3:  CRC Read error: CRC error on Dat line during read  (w1c or stop)
+ * Bit 2:  CRC_WRITE_ERROR: CRC Write error, see also error code
+ * Bit 1:  TIME_OUT_RESP: Reponse timed out (w1c or stop)
+ * Bit 0:  TIME_OUT_READ: no data received within READ_TO (w1c or stop)
+ * 
+ * ----------------------------------------------------------------------------------------------
+ */
 static uint32_t
 status_read(void *clientData,uint32_t address,int rqlen)
 {
 	IMX21Sdhc *sdhc = (IMX21Sdhc*) clientData;
-//	fprintf(stderr,"status_read: %08x\n",sdhc->status);
+	if(sdhc->card) {
+		sdhc->status |= STATUS_CARD_PRESENCE;
+	} else {
+		sdhc->status &= ~STATUS_CARD_PRESENCE;
+	}
         return sdhc->status;
 }
 
@@ -289,6 +377,15 @@ status_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
         return;
 }
 
+
+/*
+ * -----------------------------------------------------------------
+ * Clock Rate Register
+ * 
+ *	Bit 0-3:	CLK_DIVIDER
+ *	Bit 4-15:	CLK_PRESCALER
+ * -----------------------------------------------------------------
+ */
 static uint32_t
 clk_rate_read(void *clientData,uint32_t address,int rqlen)
 {
@@ -300,18 +397,8 @@ static void
 clk_rate_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMX21Sdhc *sdhc = (IMX21Sdhc*) clientData;
-	uint32_t clk_divider = value & 0xf;
-	uint32_t clk_prescaler = (value >> 4) & 0xfff;
-	uint32_t divider;
-	uint32_t prescaler;
 	sdhc->clk_rate = value;
-	if(clk_prescaler) {
-		prescaler = (clk_prescaler * 2);
-	} else {
-		prescaler = 1;
-	}
-	divider = (clk_divider+1) * (prescaler);
-	dprintf("ImxShdc: DIVIDER %d value 0x%08x rate %d\n",divider,value,  33247232 / divider);
+	update_clock(sdhc);
         return;
 }
 
@@ -358,7 +445,7 @@ do_transfer_write(void *clientData)
 		count = rqlen - sdhc->transfer_count;
 		if(count==0) {
 			fprintf(stderr,"Write complete shouldn't happen here, stop DMA\n");
-			SigNode_Set(sdhc->dmaReq,SIG_HIGH);
+			SigNode_Set(sdhc->nDmaReq,SIG_HIGH);
 			return;
 		}
 	}
@@ -384,7 +471,7 @@ do_transfer_write(void *clientData)
 		sdhc->int_status |= STATUS_APPL_BUFF_FE;
 		update_interrupt(sdhc);
 		if(sdhc->transfer_count < rqlen) {
-			SigNode_Set(sdhc->dmaReq,SIG_LOW);
+			SigNode_Set(sdhc->nDmaReq,SIG_LOW);
 		}
 	}
 	if(!sdhc->card)
@@ -394,73 +481,82 @@ do_transfer_write(void *clientData)
 	if(sdhc->transfer_count == rqlen) {
 		/* 
 		 * -----------------------------------------------------------
-		 * Real board sets this when board is nonbusy. This
- 		 * is about 6ms after write done.
+		 * Real board sets this when card is ready writing 
+		 * its internal buffer to the flash ???? MMC doku says that
+ 		 * that the dat line doesn't tell that the buffer is written, but
+		 * it tells that there is a free buffer. This
+ 		 * is about 6ms????6us after write done.
  		 * So this should be delayed and next data transaction has
-		 * to wait until it is done
+		 * to wait until it is done !!!! Do this soon now, causes
+		 * emulator to work with code where driver doesn't work
+		 * on real board
 		 * -----------------------------------------------------------
 		 */
-		sdhc->status |= STATUS_WRITE_OP_DONE;
-		sdhc->int_status |= STATUS_WRITE_OP_DONE;
-		update_interrupt(sdhc);
+		static int request_count = 0;
+		//if(request_count != 100) {
+			sdhc->status |= STATUS_WRITE_OP_DONE;
+			sdhc->int_status |= STATUS_WRITE_OP_DONE;
+			update_interrupt(sdhc);
+		//}
+		request_count++;
 	}
 	return;
 }
 
 /*
  * ----------------------------------------------------------------
- * do_transfer_read
- *	Read data from the SD-Card into buffer and set the
- *	dma request line to low. Called when a data read is
- *	started and whenever the fifo buffer becomes empty
+ * receive_data 
+ *	sink for data comming  from the SD-Card is 
+ * 	written into fifo and the dma request is initiated. 
+ *	Called by the MM/SD-Card when it sents data. 
  * ----------------------------------------------------------------
  */
-static void
-do_transfer_read(IMX21Sdhc *sdhc) 
+static int 
+receive_data(void *dev,uint8_t *buf,int len) 
 {
+	IMX21Sdhc *sdhc = (IMX21Sdhc *)dev;
 	int i;
 	/* Read direction */
-	uint8_t buf[64];
 	int fifobytes;
-	int count = 0;
+	int count;
 	int rqlen;
-	int result;
 	if(!sdhc->card) {
-		return;
-	}
-	if(sdhc->buf_fifo_wp != sdhc->buf_fifo_rp) {
-		sdhc->buf_fifo_rp = sdhc->buf_fifo_wp;
-		fprintf(stderr,"SDHC warning buffer fifo not empty when starting read\n");
+		fprintf(stderr,"Emulator Bug: Handler for non registered MMC card called\n");
+		return 0; 
 	}
 	/* Should check cmdat for read direction also here */
-	if((sdhc->cmd_dat_cont & CMD_DAT_CONT_BUS_WIDTH_MASK) == CMD_DAT_CONT_BUS_WIDTH_4) {
+	if((sdhc->cmd_dat_cont & CMD_DAT_CONT_BUS_WIDTH_MASK) == CMD_DAT_CONT_BUS_WIDTH_4)
+	{
 		fifobytes = 64;	
 	} else {
 		fifobytes = 16;
 	}
 	count = fifobytes - 2 * (sdhc->buf_fifo_wp - sdhc->buf_fifo_rp);	
 	rqlen = sdhc->nob * sdhc->blk_len;
+	if(count < len) {
+		fprintf(stderr,"SDHC: not enough room for data from mmcard: tfc %d, len %d\n",sdhc->transfer_count,len);
+	}
 	if(count + sdhc->transfer_count > rqlen) {
 		count = rqlen - sdhc->transfer_count;
 	}
-	if(count == 0) {
-		/* next read event will continue */
-		return;
-	} else if(count < 0 ) {
+	if(count < 0) {
 		fprintf(stderr,"Sdhc bug: less than zero room in transfer fifo\n");
+	} else if(count<len) {
+		/* Thats OK, because command might be of infinite length */
+		len = count;
 	} else if(count>64) {
 		fprintf(stderr,"Bug in %s line %d\n",__FILE__,__LINE__);
 		exit(1);
 	}
-	result = MMCard_Read(sdhc->card,buf,count);
 	dprintf("mmc read result %d bytes\n",result); 
-	if(result > 0 && (result <= 64)) {
+	if(len > 0 && (len <= fifobytes)) {
 		sdhc->status &= ~STATUS_APPL_BUFF_FE;
-		for(i=0;i<result;i+=2) {
+		for(i=0;i<len;i+=2) {
 			sdhc->buffer[sdhc->buf_fifo_wp % sdhc->buf_fifo_size] = buf[i] | (buf[i+1] << 8); 
 			sdhc->buf_fifo_wp++;
 		}
-		sdhc->transfer_count += result;
+		sdhc->transfer_count += len;
+		//fprintf(stderr,"transfer count is now %d, rqlen %d\n",sdhc->transfer_count,rqlen);
 		/* Is READ_OP_DONE set when data read from Card is done or when fifo dma is done ??? */
 		if(sdhc->transfer_count == rqlen) {
 			sdhc->status |= STATUS_READ_OP_DONE;
@@ -474,10 +570,12 @@ do_transfer_read(IMX21Sdhc *sdhc)
 		} else if ((sdhc->buf_fifo_wp - sdhc->buf_fifo_rp) > (fifobytes>>1)) {
 			fprintf(stderr,"Sdhc emulator bug: more than fifosize bytes in fifo\n");
 		}
-		SigNode_Set(sdhc->dmaReq,SIG_LOW);
+		SigNode_Set(sdhc->nDmaReq,SIG_LOW);
+	} else {
+		return 0;
 	}
+	return len;
 }
-
 /*
  * -----------------------------------------------------------------------
  * Do the data transaction some time after the CMD phase 
@@ -493,11 +591,7 @@ do_data_delayed(void *clientData) {
 		if((sdhc->cmd_dat_cont & CMD_DAT_CONT_WRITE)) {
 			/* should check fifo full before ? */
 			// if(full)do_transfer_write(sdhc);
-			SigNode_Set(sdhc->dmaReq,SIG_LOW);
-
-		} else {
-			dprintf("Do read transfer\n");
-			do_transfer_read(sdhc);
+			SigNode_Set(sdhc->nDmaReq,SIG_LOW);
 		}
 	}
 }
@@ -569,7 +663,9 @@ do_cmd_delayed(void *clientData)
 	 */
 	if(sdhc->cmd_dat_cont & CMD_DAT_CONT_DATA_ENABLE) {
 		sdhc->transfer_count = 0;
-		CycleTimer_Mod(&sdhc->data_delay_timer,MicrosecondsToCycles(UDELAY_DATA));
+		if((sdhc->cmd_dat_cont & CMD_DAT_CONT_WRITE)) {
+			CycleTimer_Mod(&sdhc->data_delay_timer,MicrosecondsToCycles(UDELAY_DATA));
+		} 
 	}
 }
 
@@ -583,14 +679,19 @@ cmd_dat_cont_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 		return;
 	}
 	if(sdhc->res_fifo_wp != sdhc->res_fifo_rp) {
+		fprintf(stderr,"SDHC: Response fifo not empty cmd %d, prev %d, rp %d, wp %d\n",sdhc->cmd,sdhc->prev_cmd,(int)sdhc->res_fifo_rp,(int)sdhc->res_fifo_wp);
 		sdhc->res_fifo_rp = sdhc->res_fifo_wp;
-		fprintf(stderr,"SDHC warning resp fifo not empty when starting cmd %d\n",sdhc->cmd);
 	}
 	/* Bad ! should wait until clock is started */ 
 	CycleTimer_Mod(&sdhc->cmd_delay_timer,MicrosecondsToCycles(UDELAY_CMD));
         return;
 }
 
+/* 
+ * -------------------------------------------------------------------------
+ * Response timeout has currently no effect
+ * -------------------------------------------------------------------------
+ */
 static uint32_t
 response_to_read(void *clientData,uint32_t address,int rqlen)
 {
@@ -721,6 +822,7 @@ static void
 cmd_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMX21Sdhc *sdhc = (IMX21Sdhc*) clientData;
+	sdhc->prev_cmd = sdhc->cmd;
 	sdhc->cmd = value & 0x3f;
         return;
 }
@@ -805,12 +907,9 @@ buffer_access_read(void *clientData,uint32_t address,int rqlen)
 		if(sdhc->buf_fifo_rp == sdhc->buf_fifo_wp) {
 			sdhc->status = sdhc->status | STATUS_APPL_BUFF_FE;
 			sdhc->int_status = sdhc->int_status | STATUS_APPL_BUFF_FE;
-			dprintf("Set DMA request to high\n");
-			SigNode_Set(sdhc->dmaReq,SIG_HIGH);
+			dprintf("Release DMA request\n");
+			SigNode_Set(sdhc->nDmaReq,SIG_HIGH);
 			update_interrupt(sdhc);
-			if((sdhc->cmd_dat_cont & CMD_DAT_CONT_DATA_ENABLE) && !(sdhc->cmd_dat_cont & CMD_DAT_CONT_WRITE)) {
-				do_transfer_read(sdhc); 
-			}
 		}
 	}
         return value;
@@ -834,7 +933,7 @@ buffer_access_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 			sdhc->status = sdhc->status | STATUS_APPL_BUFF_FF;
 			sdhc->int_status = sdhc->int_status | STATUS_APPL_BUFF_FF;
 			dprintf("Stop the DMA request because of full fifo\n");
-			SigNode_Set(sdhc->dmaReq,SIG_HIGH);
+			SigNode_Set(sdhc->nDmaReq,SIG_HIGH);
 			update_interrupt(sdhc);
 			CycleTimer_Mod(&sdhc->write_delay_timer,MicrosecondsToCycles(UDELAY_WRITE));
 		}
@@ -894,6 +993,7 @@ IMX21Sdhc_InsertCard(BusDevice *dev,MMCard *card)
 		return -1;
 	}
 	sdhc->card = card;
+	MMCard_AddListener(card,sdhc,16,receive_data);
 	return 0;
 }
 
@@ -905,6 +1005,7 @@ IMX21Sdhc_RemoveCard(BusDevice *dev,MMCard *card)
 		fprintf(stderr,"Error: SD Card Slot is already empty, can not remove\n");
 		return -1;
 	}
+	MMCard_RemoveListener(card,sdhc);
 	sdhc->card = NULL;
 	return 0;
 }
@@ -919,12 +1020,13 @@ IMX21_SdhcNew(const char *name) {
         }
         memset(sdhc,0,sizeof(IMX21Sdhc));
 	sdhc->irqNode = SigNode_New("%s.irq",name);
+	sdhc->name = strdup(name);
 	if(!sdhc->irqNode) {
 		fprintf(stderr,"i.MX21 SDHC: can not create interrupt line\n");
 		exit(1);
 	}
-	sdhc->dmaReq = SigNode_New("%s.dma_req",name);
-	if(!sdhc->dmaReq) {
+	sdhc->nDmaReq = SigNode_New("%s.dma_req",name);
+	if(!sdhc->nDmaReq) {
 		fprintf(stderr,"i.MX21 SDHC: can not create DMA request line\n");
 		exit(1);
 	}
@@ -935,6 +1037,10 @@ IMX21_SdhcNew(const char *name) {
         sdhc->bdev.UnMap=IMXSdhc_Unmap;
         sdhc->bdev.owner=sdhc;
         sdhc->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
+	sdhc->inclk = Clock_New("%s.inclk",name);
+	sdhc->sdclk = Clock_New("%s.sdclk",name);
+	/* Documentation (big piece of shit) has contradictions about clock source for SD-Clk */
+	Clock_SetFreq(sdhc->inclk,66500013.419/2); /* Should be connected to clock module ipg_perclk or perclk2  */
 	CycleTimer_Init(&sdhc->cmd_delay_timer,do_cmd_delayed,sdhc);
 	CycleTimer_Init(&sdhc->data_delay_timer,do_data_delayed,sdhc);
 	CycleTimer_Init(&sdhc->write_delay_timer,do_transfer_write,sdhc);
