@@ -5,8 +5,6 @@
  *   Author: Jochen Karrer
  *
  * state: basically working 
- *	configuring the terminal to correct baudrate is missing 
- *	Modem control signals are missing
  *
  *  This program is free software; you can distribute it and/or modify it
  *  under the terms of the GNU General Public License (Version 2) as
@@ -33,9 +31,11 @@
 #include <bus.h>
 #include <fio.h>
 #include <signode.h>
+#include <clock.h>
 #include <imx21_uart.h>
 #include <serial.h>
 #include <configfile.h>
+#include <senseless.h>
 
 #define URR(base)	((base) + 0x0)
 #define		URR_CHARRDY	(1<<15)
@@ -186,6 +186,9 @@
 
 typedef struct IMX_Uart {
 	BusDevice bdev;
+	Clock_t *in_clk;  /* from crm module */
+	Clock_t *refclk;	
+	char *name;
 	UartPort *port;
 	uint16_t ucr1,ucr2,ucr3,ucr4;
 	uint16_t ufcr;
@@ -200,12 +203,12 @@ typedef struct IMX_Uart {
 	uint16_t uts;	
 
 	uint8_t txfifo[TX_FIFO_SIZE];
-	uint64_t txfifo_wp;
-        uint64_t txfifo_rp;
+	int64_t txfifo_wp;
+        int64_t txfifo_rp;
 
 	uint16_t rxfifo[RX_FIFO_SIZE];
-	uint64_t rxfifo_wp;
-        uint64_t rxfifo_rp;
+	int64_t rxfifo_wp;
+        int64_t rxfifo_rp;
 
         int interrupt_posted;
         SigNode *irqNode;
@@ -226,33 +229,45 @@ update_interrupts(IMX_Uart *iuart)
 	int txtl = (iuart->ufcr & UFCR_TXTL_MASK) >> UFCR_TXTL_SHIFT;
 	int rxtl = (iuart->ufcr & UFCR_RXTL_MASK) >> UFCR_RXTL_SHIFT;
 	if(trdyen && (tx_room >= txtl)) {
-		//fprintf(stderr,"trdyen\n");
 		interrupt=1;
 	}
 	/* does it also trigger an interrupt when 0 characters are received  ? */
 	if(rrdyen && (rx_fill >= rxtl)) {
-		//fprintf(stderr,"rrdyen\n");
 		interrupt = 1;
 	}
 	if(txmptyen && (TX_FIFO_COUNT(iuart) == 0)) {
-		//fprintf(stderr,"txmptyen\n");
 		interrupt = 1;
 	}
 	if(interrupt) {
 		if(!iuart->interrupt_posted) {
-			//fprintf(stderr,"Post imxserial irq trp %lld, twp %lld rrp %lld rwp %lld\n",iuart->txfifo_rp,iuart->txfifo_wp,iuart->rxfifo_rp,iuart->rxfifo_wp);
 			SigNode_Set(iuart->irqNode,SIG_LOW);
 			iuart->interrupt_posted = 1;
 		}
 	} else {
 		if(iuart->interrupt_posted) {
-			//fprintf(stderr,"Unpost imxserial irq\n");
 			SigNode_Set(iuart->irqNode,SIG_HIGH);
 			iuart->interrupt_posted = 0;
 		}
 	}
 }
 
+static void
+update_refclk(IMX_Uart *iuart) 
+{
+	int rfdiv = (iuart->ufcr & UFCR_RFDIV_MASK) >> UFCR_RFDIV_SHIFT;
+	int divider = 1;
+	if(rfdiv < 6) {
+		divider = (rfdiv ^ 7) - 1;
+	} else if(rfdiv == 5) { 
+		divider = 7;
+	} else {
+		fprintf(stderr,"IMXUart: Illegal rfdiv value in UFCR register\n");
+	}
+	Clock_MakeDerived(iuart->refclk,iuart->in_clk,1,divider);
+	//fprintf(stderr,"Clock: %f, ref %f\n",Clock_Freq(iuart->in_clk),Clock_Freq(iuart->refclk));
+
+}
+
 /*
  * ------------------------------------------------------------------------
  * DMA request lines
@@ -294,6 +309,15 @@ serial_rx_char(IMX_Uart *iuart,uint8_t c) {
         return;
 }
 
+/*
+ * --------------------------------------------------------------------------------
+ * Serial Input has to accept the chars received from the underlying UART.
+ * It has the ability to stop the data when the input fifo is full. This is
+ * not very realistic because only UARTS with flow control enabled wait when
+ * fifo is full. But timing isn't good enough currently to guarantee that
+ * the cpu runs enough Cycles to empty the RX-Fifo. 
+ * --------------------------------------------------------------------------------
+ */
 
 static void 
 serial_input(void *cd) {
@@ -301,7 +325,7 @@ serial_input(void *cd) {
         int fifocount;
         while(1) {
                 uint8_t c;
-                int count=Uart_Read(iuart->port,&c,1);
+                int count=SerialDevice_Read(iuart->port,&c,1);
         	int room; 
                 if(count==1) {
                         serial_rx_char(iuart,c);
@@ -311,7 +335,7 @@ serial_input(void *cd) {
                 }
 		room = RX_FIFO_ROOM(iuart);
 		if(room<1) {
-			Uart_StopRx(iuart->port);
+			SerialDevice_StopRx(iuart->port);
 			break;
 		}
         }
@@ -323,6 +347,14 @@ serial_input(void *cd) {
         return;
 }
 
+/*
+ * --------------------------------------------------------------------------------
+ * Serial output is called whenever the underlying UART backend is able
+ * to accept one or more chars. If serial output has no data for sending
+ * (TX fifo is empty). Then the backend has to be stopped because the
+ * Event handler for fetching more data should not be invoked anymore. 
+ * --------------------------------------------------------------------------------
+ */
 static void 
 serial_output(void *cd) {
         IMX_Uart *iuart=cd;
@@ -334,13 +366,13 @@ serial_output(void *cd) {
                 if((TX_FIFO_RIDX(iuart)+fill) > TX_FIFO_SIZE) {
                         len = TX_FIFO_SIZE - TX_FIFO_RIDX(iuart);
                 }
-                count=Uart_Write(iuart->port,&iuart->txfifo[TX_FIFO_RIDX(iuart)],len);
+                count=SerialDevice_Write(iuart->port,&iuart->txfifo[TX_FIFO_RIDX(iuart)],len);
                 if(count>0) {
                         iuart->txfifo_rp = iuart->txfifo_rp + count;
 			update_txdma(iuart);
                 } 
         }
-	Uart_StopTx(iuart->port);
+	SerialDevice_StopTx(iuart->port);
 	update_interrupts(iuart);
         return;
 }
@@ -359,7 +391,7 @@ urr_read(void *clientData,uint32_t address,int rqlen)
 		update_rxdma(iuart);
 		/* Maybe rx was stopped because fifo was full */
 		if((count == RX_FIFO_SIZE) && (iuart->ucr2 & UCR2_RXEN)) { 
-			Uart_StartRx(iuart->port); 
+			SerialDevice_StartRx(iuart->port); 
 		}
 	} else {
 		urr = urr &~ URR_CHARRDY;
@@ -393,13 +425,12 @@ utr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 		fprintf(stderr,"IMX21-Uart TX-Fifo overflow\n");
 		return;
 	} 
-	//fprintf(stderr,"put char \"%c\" to fifo\n",value);
 	iuart->txfifo[TX_FIFO_WIDX(iuart)] = value;	
 	iuart->txfifo_wp += 1;
 	if(iuart->ucr2 & UCR2_TXEN) {
-		Uart_StartTx(iuart->port);
+		SerialDevice_StartTx(iuart->port);
 	} else {
-		fprintf(stderr,"tx not enabled\n" );
+		fprintf(stderr,"i.MX21 UART: Writing to TX register of %s while disabled\n",iuart->name );
 	}
 	update_interrupts(iuart);
 	update_txdma(iuart);
@@ -434,25 +465,75 @@ ucr2_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
 	uint16_t diff = iuart->ucr2 ^ value;
+	UartCmd cmd;
 	iuart->ucr2 = value | UCR2_SRST;
-	//fprintf(stderr,"UCR 2 now %08x\n",value);
 	if(diff & UCR2_TXEN) {
 		if(value & UCR2_TXEN) {
-			Uart_StartTx(iuart->port);
+			SerialDevice_StartTx(iuart->port);
 		} else {
-			Uart_StopTx(iuart->port);
+			SerialDevice_StopTx(iuart->port);
 		}
 	}
 	if(diff & UCR2_RXEN) {
 		if(value & UCR2_RXEN) {
 			int room = RX_FIFO_ROOM(iuart);
-			if(room>0) {
-				Uart_StartRx(iuart->port);
+			if(room > 0) {
+				if(!(iuart->ucr3 & UCR3_RXDMUXSEL)) {
+					fprintf(stderr,"Driver bug, starting rx with DMUXSEL 0\n");
+				} else {
+					SerialDevice_StartRx(iuart->port);
+				}
 			}
 		} else {
-			Uart_StopRx(iuart->port);
+			SerialDevice_StopRx(iuart->port);
 		}
 	}
+	if(!(value & UCR2_SRST)) {
+		iuart->txfifo_wp  =  iuart->txfifo_rp = 0; 
+		iuart->rxfifo_wp =   iuart->rxfifo_rp = 0; 
+		iuart->usr1 = USR1_RXDS | USR1_TRDY;
+		iuart->usr2 = USR2_TXFE | USR2_TXDC;
+	}
+	if(iuart->ucr2 & UCR2_CTSC) {
+		cmd.opcode = UART_OPC_CRTSCTS; 
+		cmd.arg = 1;
+		SerialDevice_Cmd(iuart->port,&cmd);
+	} else {
+		cmd.opcode = UART_OPC_CRTSCTS; 
+		cmd.arg = 0;
+		SerialDevice_Cmd(iuart->port,&cmd);
+
+		cmd.opcode = UART_OPC_SET_RTS;
+		if(iuart->ucr2 & UCR2_CTS) {
+			cmd.arg = UART_RTS_ACT;
+		} else {
+			cmd.arg = UART_RTS_INACT;
+		}
+		SerialDevice_Cmd(iuart->port,&cmd);
+	}
+	cmd.opcode = UART_OPC_PAREN; 
+	if(iuart->ucr2 & UCR2_PREN) {
+		cmd.arg = 1;
+	} else {
+		cmd.arg = 0;
+	}
+	SerialDevice_Cmd(iuart->port,&cmd);
+
+	cmd.opcode = UART_OPC_PARODD; 
+	if(iuart->ucr2 & UCR2_PROE) {
+		cmd.arg = 1;
+	} else {
+		cmd.arg = 0;
+	}
+	SerialDevice_Cmd(iuart->port,&cmd);
+
+	cmd.opcode = UART_OPC_SET_CSIZE;
+	if(iuart->ucr2 & UCR2_WS) {
+		cmd.arg = 8;
+	} else {
+		cmd.arg = 7;
+	}
+	SerialDevice_Cmd(iuart->port,&cmd);
         return;
 }
 
@@ -499,6 +580,7 @@ ufcr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
 	iuart->ufcr = value & 0xffff;
 	update_interrupts(iuart);
+	update_refclk(iuart);
         return;
 }
 
@@ -506,11 +588,13 @@ static uint32_t
 usr1_read(void *clientData,uint32_t address,int rqlen)
 {
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
+	UartCmd cmd;
 	int tx_room = TX_FIFO_ROOM(iuart); 
 	int rx_fill = RX_FIFO_COUNT(iuart);
 	int txtl = (iuart->ufcr & UFCR_TXTL_MASK) >> UFCR_TXTL_SHIFT;
 	int rxtl = (iuart->ufcr & UFCR_RXTL_MASK) >> UFCR_RXTL_SHIFT;
 	uint16_t usr1 = iuart->usr1;
+	uint16_t diff;
 
 	if(tx_room >= txtl) {
 		usr1 |= USR1_TRDY;	
@@ -522,6 +606,16 @@ usr1_read(void *clientData,uint32_t address,int rqlen)
 	} else {
 		usr1 |= USR1_RRDY;
 	}
+	cmd.opcode = UART_OPC_GET_CTS;	
+	SerialDevice_Cmd(iuart->port,&cmd);
+	if(cmd.retval) {
+		usr1 |= USR1_RTSS;
+	}
+	diff = usr1 ^ iuart->usr1;
+	if(diff & USR1_RTSS) {
+		usr1 |= USR1_RTSD;
+	}
+	iuart->usr1 = usr1;
 	return usr1; 
 }
 
@@ -538,18 +632,45 @@ static uint32_t
 usr2_read(void *clientData,uint32_t address,int rqlen)
 {
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
+	uint16_t diff;
+	uint16_t usr2 = iuart->usr2;
 	int rx_fill = RX_FIFO_COUNT(iuart);
 	int tx_fill = TX_FIFO_COUNT(iuart);
-	if(rx_fill) {
-		iuart->usr2 |= USR2_RDR;
+	UartCmd cmd;
+	
+	if(rx_fill > 0) {
+		usr2 |= USR2_RDR;
 	} else {
-		iuart->usr2 &= ~USR2_RDR;
+		usr2 &= ~USR2_RDR;
 	}
-	if(tx_fill) {
+	if(tx_fill > 0) {
 		//fprintf(stderr,"Not empty\n");
-		iuart->usr2 &= ~USR2_TXFE;
+		usr2 &= ~(USR2_TXFE | USR2_TXDC);
 	} else {
-		iuart->usr2 |= USR2_TXFE;
+		usr2 |= USR2_TXFE | USR2_TXDC;
+	}
+
+	cmd.opcode = UART_OPC_GET_RI;	
+	SerialDevice_Cmd(iuart->port,&cmd);
+	if(cmd.retval) {
+		usr2 |= USR2_RIIN;
+	}
+	cmd.opcode = UART_OPC_GET_DCD;
+	SerialDevice_Cmd(iuart->port,&cmd);
+	if(cmd.retval) {
+		usr2 |= USR2_DCDIN;
+	}
+
+	diff = usr2 ^ iuart->usr2;
+	if(diff & USR2_RIIN) {
+		usr2 |= USR2_RIDELT;
+	}
+	if(diff & USR2_DCDIN) {
+		usr2 |= USR2_DCDDELT;
+	}
+	iuart->usr2 = usr2;
+	if((usr2 & USR2_TXFE) && (~usr2 & USR2_RDR)) {
+		Senseless_Report(150);
 	}
 	return iuart->usr2;
 }
@@ -609,6 +730,7 @@ static void
 ubir_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
+	//fprintf(stderr,"UBIR write %d\n",iuart->ubir);
 	iuart->ubir = value & 0xffff;
         return;
 }
@@ -624,7 +746,14 @@ static void
 ubmr_write(void *clientData,uint32_t value,uint32_t address,int rqlen)
 {
 	IMX_Uart *iuart = (IMX_Uart *) clientData;
+	int baudrate;
+	UartCmd cmd;
 	iuart->ubmr = value & 0xffff;
+	baudrate = 1.0 *(iuart->ubir + 1)
+		* Clock_Freq(iuart->refclk)/(iuart->ubmr+1)/16;
+	cmd.opcode = UART_OPC_SET_BAUDRATE;
+	cmd.arg = baudrate;
+	SerialDevice_Cmd(iuart->port,&cmd);
         return;
 }
 
@@ -666,12 +795,12 @@ uts_read(void *clientData,uint32_t address,int rqlen)
 	int tx_fill = TX_FIFO_COUNT(iuart);
 	int rx_room = RX_FIFO_ROOM(iuart);
 	int tx_room = TX_FIFO_ROOM(iuart);
-	if(rx_fill) {
+	if(rx_fill > 0) {
 		iuart->uts &= ~UTS_RXEMPTY;
 	} else {
 		iuart->uts |= UTS_RXEMPTY;
 	}
-	if(tx_fill) {
+	if(tx_fill > 0) {
 		iuart->uts &= ~UTS_TXEMPTY;
 	} else {
 		iuart->uts |= UTS_TXEMPTY;
@@ -687,6 +816,9 @@ uts_read(void *clientData,uint32_t address,int rqlen)
 		iuart->uts |= UTS_TXFULL;
 	}
 	//fprintf(stderr,"rx_room %d tx_room %d,rx_fill %d tx_fill %d uts %08x\n",rx_room,tx_room,rx_fill,tx_fill,iuart->uts);
+	if((iuart->uts & UTS_TXEMPTY) && (iuart->uts & UTS_RXEMPTY)) {
+		Senseless_Report(100);
+	}
 	return iuart->uts;
 }
 
@@ -719,7 +851,7 @@ IMXUart_Map(void *owner,uint32_t base,uint32_t mask,uint32_t flags)
 {
        IMX_Uart *iuart=owner;
 	/* Partial read: Read all return part, Partial Write: Read modify write */
-	uint32_t mapflags = IOH_FLG_PRD_RARP | IOH_FLG_PWR_RMW;
+	uint32_t mapflags = IOH_FLG_PRD_RARP | IOH_FLG_PWR_RMW | IOH_FLG_HOST_ENDIAN;
 	IOH_New32f(URR(base),urr_read,urr_write,iuart,mapflags);
 	IOH_New32f(UTR(base),utr_read,utr_write,iuart,mapflags);
 	IOH_New32f(UCR1(base),ucr1_read,ucr1_write,iuart,mapflags);
@@ -779,14 +911,22 @@ IMXUart_New(const char *name)
 	iuart->usr1 = USR1_RXDS | USR1_TRDY;
 	iuart->usr2 = USR2_TXFE | USR2_TXDC;
 	iuart->ucr2 = UCR2_SRST;
+	/* Make it nonworking when user forgets to reset the fifo */
+	iuart->rxfifo_rp = 0x1000000;
+	iuart->ufcr = 0x00000801;
 	iuart->bdev.first_mapping=NULL;
         iuart->bdev.Map=IMXUart_Map;
         iuart->bdev.UnMap=IMXUart_UnMap;
         iuart->bdev.owner=iuart;
         iuart->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
+	iuart->name = strdup(name);
+	iuart->port = Uart_New(name,serial_input,serial_output,NULL,iuart);
+	iuart->in_clk = Clock_New("%s.clk",name);
+	/* Clock should come from perclk1 of CRM module */
+
+	iuart->refclk = Clock_New("%s.refclk",name);
+	update_refclk(iuart);
 	update_interrupts(iuart);
 	fprintf(stderr,"IMX21 Uart \"%s\" created\n",name);
-	//init_serial_fd(iuart,name);
-	iuart->port = Uart_New(name,serial_input,serial_output,NULL,iuart);
         return &iuart->bdev;
 }
