@@ -1259,11 +1259,6 @@ i_lsm() {
 	L=testbit(20,icode); // Load/Store
 	S=testbit(22,icode); // Modify Flags ?
 	if(S && L && (icode&(1<<15))) {
-		if(MODE_HAS_SPSR) {
-			SET_REG_CPSR(REG_SPSR);
-		} else {
-			fprintf(stderr,"RegBank %02x has no spsr in line %d\n",ARM_BANK,__LINE__);
-		}	
 		bank=ARM_BANK;
 	}  else if(S) {
 		bank=MODE_USER;	
@@ -1327,7 +1322,6 @@ i_lsm() {
 			for(i=0;i<8;i++) {
 				if(icode&(1<<i)) { 
 					value=ARM9_ReadRegNot15(i);
-					//value=ARM9_ReadReg(i);
 					MMU_Write32(value,start_address);
 					start_address+=4;
 				}
@@ -1357,6 +1351,13 @@ i_lsm() {
 		}
 		ARM9_WriteReg(Rn,rn);
 	}
+	if(S && L && (icode&(1<<15))) {
+		if(likely(MODE_HAS_SPSR)) {
+			SET_REG_CPSR(REG_SPSR);
+		} else {
+			fprintf(stderr,"RegBank %02x has no spsr in line %d\n",ARM_BANK,__LINE__);
+		}	
+        }
 	CycleCounter+=(ones<<1);
 	dprintf("Done LSM addr %08x L %d\n",start_address, L ? 1:0); 
 }
@@ -1626,10 +1627,22 @@ i_bkpt() {
 
 void 
 i_blx1() {
-	fprintf(stderr,"Thumb instruction Set not implemented\n");
-	fprintf(stderr,"Not implemented blx1\n"); 
-	ARM9_Exception(EX_UNDEFINED,4);
+	uint32_t icode = ICODE;
+	int32_t offset; 
+	uint32_t h = (icode >> 23) & 2;
+	if(icode & (1<<23)) {
+		offset = (icode & 0xffffff) | 0xff000000;
+	} else {
+		offset = (icode & 0xffffff);
+	}
+	offset = (offset << 2) | h;
+	REG_LR=GET_REG_PC-4;	
+	REG_CPSR |= FLAG_T;  
+	SET_REG_PC(GET_REG_PC+offset);
+	/* Enter Thumb */
+	ARM_RestartLoop();
 }
+
 void 
 i_blx2bx() 
 {
@@ -1642,10 +1655,10 @@ i_blx2bx()
 	}
 	rm = icode&0xf;
 	Rm=ARM9_ReadReg(rm);
-	if(Rm&1) {
+	if(Rm & 1) {
+		/* Enter Thumb mode */
 		REG_CPSR |= FLAG_T;  
-		fprintf(stderr,"Thumb instruction Set not implemented\n");
-		ARM9_Exception(EX_UNDEFINED,4);
+		ARM_RestartLoop();
 	}
 	L = testbit(5,icode);
 	if(L) {
@@ -1862,8 +1875,8 @@ i_strwub(){
 	if(!B) {
 		value=ARM9_ReadReg(rd);
 		if(unlikely(addr&3)) {
-			//fprintf(stderr,"Warning: unaligned str icode %08x, addr %08x, pc %08x\n",icode,addr,GET_REG_PC);
 			MMU_AlignmentException(addr);
+			fprintf(stderr,"Warning: unaligned str icode %08x, addr %08x, pc %08x\n",icode,addr,GET_REG_PC);
 			addr=addr&~3;
 		}
 		MMU_Write32(value,addr);
@@ -1895,8 +1908,8 @@ i_ldr() {
 	if(unlikely(addr&3)) {
 		unsigned int shift;
 		uint32_t memval;
-		/*fprintf(stderr,"Warning: Unaligned ldr  from %08x to reg %d at PC %08x\n",addr,rd,GET_REG_PC-8); */
 		MMU_AlignmentException(addr);
+		fprintf(stderr,"Warning: Unaligned ldr  from %08x to reg %d at PC %08x\n",addr,rd,GET_REG_PC-8); 
 		memval = MMU_Read32(addr&~3);
 		shift=(addr&3)<<3;
 		value = (memval >> shift) | (memval << (32-shift));
@@ -1906,10 +1919,14 @@ i_ldr() {
 	if(likely(rd!=15))  {
 		ARM9_WriteReg(value,rd);
 	} else {
-		SET_REG_PC(value&0xfffffffe);	
-		if(value&2) {
-			fprintf(stderr,"Thumb mode not implemented\n");
-			exit(2);
+		if(unlikely(value & 1)) {
+			/* Goto Thumb mode */
+			REG_CPSR |= FLAG_T;  
+			SET_REG_PC(value&0xfffffffe);	
+			ARM9_WriteReg(Rn_new,rn);
+			ARM_RestartLoop();
+		} else {
+			SET_REG_PC(value&0xfffffffc);	
 		}
 	}
 	if(Rn_new!=Rn) {
@@ -1961,8 +1978,8 @@ i_ldrh(){
 	rd=(icode>>12)&0xf;
 	addr = addr_mode3();
 	if(unlikely(addr&1)) {
-		fprintf(stderr,"Unaligned ldrh at 0x%08x rd %d gives unpredictable results !\n",GET_REG_PC-8,rd);
 		MMU_AlignmentException(addr);
+		fprintf(stderr,"Unaligned ldrh at 0x%08x rd %d gives unpredictable results !\n",GET_REG_PC-8,rd);
 		return;
 	}
 	value=MMU_Read16(addr);
@@ -2010,8 +2027,8 @@ i_ldrsh(){
 	rd=(icode>>12)&0xf;
 	addr = addr_mode3();
 	if(unlikely(addr&1)) {
-		fprintf(stderr,"Unaligned ldrh instruction gives unpredictable results\n");
 		MMU_AlignmentException(addr);
+		fprintf(stderr,"Unaligned ldrh instruction gives unpredictable results\n");
 		return;
 	}
 	value=MMU_Read16(addr);
@@ -2032,22 +2049,18 @@ i_ldrt() {
 void 
 i_mcr() {
 	uint32_t icode = ICODE;
-	int opcode_1,crn,rd,cp_num,opcode_2,crm;
+	int rd,cp_num;
 	uint32_t Rd;
 	ArmCoprocessor *copro;
 	if(!check_condition(icode)) {
 		return;
 	}
-	opcode_1=(icode>>21)&7;
-	crn=(icode>>16)&0xf;
-	rd=(icode>>12)&0xf;
 	cp_num=(icode>>8)&0xf;
-	opcode_2=(icode>>5)&0x7;
-	crm=icode&0xf;
+	rd=(icode>>12)&0xf;
 	copro=gcpu.copro[cp_num];
 	if(copro && copro->mcr) {
 		Rd=ARM9_ReadReg(rd);
-		copro->mcr(copro,opcode_1,Rd,crn,crm,opcode_2);
+		copro->mcr(copro,icode,Rd);
 	} else {		
 		dprintf("ArmCoprocessor %d MCR operation not found\n",cp_num); 
 		ARM9_Exception(EX_UNDEFINED,4);
@@ -2127,21 +2140,18 @@ i_mov() {
 void 
 i_mrc() {
 	uint32_t icode = ICODE;
-	int opcode_1,crn,rd,cp_num,opcode_2,crm;
+	int rd,cp_num;
 	uint32_t Rd;
 	ArmCoprocessor *copro;
 	if(!check_condition(icode)) {
 		return;
 	}
-	opcode_1=(icode>>21)&7;
-	crn=(icode>>16)&0xf;
+
 	rd=(icode>>12)&0xf;
 	cp_num=(icode>>8)&0xf;
-	opcode_2=(icode>>5)&0x7;
-	crm=icode&0xf;
 	copro=gcpu.copro[cp_num];
 	if(copro && copro->mrc) {
-		Rd=copro->mrc(copro,opcode_1,crn,crm,opcode_2);
+		Rd=copro->mrc(copro,icode);
 		if(rd==15) {
 			REG_CPSR=(REG_CPSR&0x0fffffff) | (Rd & 0xf0000000); 
 			dprintf("Rd %08x, new cpsr %08x at %08x\n",Rd,REG_CPSR,GET_REG_PC);
@@ -2664,8 +2674,8 @@ i_strh(){
 	rd = (icode>>12)&0xf;
 	addr = addr_mode3();
 	if(unlikely(addr&1)) {
-		fprintf(stderr,"Unaligned strh at %08x rd %d gives unpredictable results\n",GET_REG_PC-8,rd);
 		MMU_AlignmentException(addr);
+		fprintf(stderr,"Unaligned strh at %08x rd %d gives unpredictable results\n",GET_REG_PC-8,rd);
 		return;
 	}
 	Rd=ARM9_ReadReg(rd);
@@ -2676,12 +2686,14 @@ i_strh(){
 	}
 	dprintf("STRH to value %04x to %08x\n",Rd,addr);
 }
+
 void 
 i_strt(){
 	gcpu.signaling_mode=MODE_USER;
 	i_strwub();
 	gcpu.signaling_mode=REG_CPSR & 0x1f;
 }
+
 void 
 i_sub(){
 	uint32_t icode = ICODE;
@@ -3290,7 +3302,7 @@ i_strd() {
 	rd1=icode>>12&0xf;
 	rd2=rd1|1;
 	if(rd1&1) {
-		fprintf(stderr,"LDRD: only even numbered destination registers are allowed\n");
+		fprintf(stderr,"STRD: only even numbered destination registers are allowed\n");
 		return;
 	}
 	addr = addr_mode3();
