@@ -23,6 +23,4 @@
  */
 
 #include <stdint.h>
-int Load_AutoType(char *filename,uint32_t addr);
-int Load_Binary(char *filename,uint32_t addr);
-int Load_SRecords(char *filename);
+int64_t Load_AutoType(char *filename,uint32_t addr,uint64_t region_size);
