@@ -5,6 +5,7 @@
  * which is responsible for Memory Management
  *
  * (C) 2004  Lightmaze Solutions AG
+ * (C) 2006  Jochen Karrer
  *   Author: Jochen Karrer
  *   State: Working, MVA mode is missing
  *
@@ -41,6 +42,39 @@
 #include "debug.h"
 #include "mainloop_events.h"
 
+typedef void   McrProc(void *clientData,uint32_t icode,uint32_t value);
+typedef uint32_t MrcProc(void *clientData,uint32_t icode);
+
+
+typedef struct SystemCopro {
+	uint32_t id;
+	uint32_t ctrl;
+	uint32_t mtbase;
+	uint32_t mdac;
+	uint32_t dfsr;
+	uint32_t ifsr;
+	uint32_t far;
+	uint32_t mcctrl;
+	uint32_t mtlbctrl;
+	uint32_t dclck; // DCache lock
+	uint32_t iclck; // ICache lock
+	uint32_t mtlblck;
+	uint32_t mpid;
+
+	int64_t saved_cycles; /* Halt Instruction */
+	CycleCounter_t last_halt_cycles;
+	struct timespec last_halt;
+	SigNode *endianNode;
+	int debugmode;
+
+	/* The registers */
+	McrProc *mcrProc[16];
+	MrcProc  *mrcProc[16];
+	void *crn_clientData[16];
+	
+} SystemCopro;
+
+
 static uint32_t translation_enabled=0;
 static uint32_t ttbl_base=0; 
 static uint32_t sys_rom_protection=0;
@@ -158,6 +192,7 @@ update_byteorder(SystemCopro *mmu)
 		mmu_byte_addr_xor = 0x3;
 		mmu_word_addr_xor = 0x2;
 		SigNode_Set(mmu->endianNode,SIG_HIGH);
+		fprintf(stderr,"pc %08x\n",GET_REG_PC);
 	} else {
 		fprintf(stderr,"MMU: Byteorder is now LE\n");
 		mmu_byte_addr_xor = 0;
@@ -834,284 +869,424 @@ MMU_Write8(uint8_t value,uint32_t addr) {
 	IO_Write8(value,taddr);	
 }
 
+static void 
+CrnHandler_New(SystemCopro *mmu,int crn,MrcProc *rproc,McrProc *wproc,void *cd)
+{
+	crn = crn & 0xf;
+	mmu->mcrProc[crn] = wproc;
+	mmu->mrcProc[crn] = rproc;
+	mmu->crn_clientData[crn] = cd;
+}
+
+static void
+id_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	fprintf(stderr,"MMU:  CPU-ID is not writable\n");
+}
+
+static uint32_t 
+arm920t_id_read(void *clientData,uint32_t icode)
+{
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm!=0) {
+		fprintf(stderr,"MMU: CRM is not zero(%d) when reading ID\n",crm);
+		return 0;
+	}
+	if(opcode_2==1) {
+		/* 16/16 kB Harvard Instruction/Data Cache */
+		return 0x0d172172;
+	} else if(opcode_2==2) {
+		return 0; // TCM not yet ?
+	}
+	return 0x41129200 | 0;
+
+}
+
+static uint32_t 
+arm926ejs_id_read(void *clientData,uint32_t icode)
+{
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm!=0) {
+		fprintf(stderr,"MMU: CRM is not zero(%d) when reading ID\n",crm);
+		return 0;
+	}
+	if(opcode_2==1) {
+		/* 8/4 kB Harvard Instruction/Data Cache */
+		return 0x1d0d2112;
+	} else if(opcode_2==2) {
+			return 0; // TCM not yet ?
+	}
+	return 0x41069260 | 4; // CPU revision 4
+
+}
+
+static uint32_t
+ctrl_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+	if(crm!=0) {
+		fprintf(stderr,"MMU: CRM is not zero(%d) when reading CTRL\n",crm);
+			return 0;
+	}
+	return (mmu->ctrl & ~0xffff0c78UL)|0x00050078;
+}
+
+static void
+ctrl_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+	uint32_t diff;
+	dprintf("Setting MMU Control Register to %08x\n",value);
+	if(crm!=0) {
+		fprintf(stderr,"MMU: CRM is not zero(%d) when writing CTRL\n",crm);
+		return;
+	}
+	/* Bit 3 is write buffer enable */
+	diff=mmu->ctrl ^ value;
+	mmu->ctrl=value;
+	if(diff & MCTRL_BE) {
+		update_byteorder(mmu);
+	} 
+	translation_enabled= value & MCTRL_MMUEN;
+	mmu_high_vectors= value & MCTRL_V;
+	do_alignment_check= value & MCTRL_ALGNCHK;
+	sys_rom_protection=((value>>8)&0x3)<<SYS_ROM_SHIFT;
+	if(diff & (3<<8)) {
+		invalidate_tlb();
+	}
+#if 0
+	if(value & MCTRL_LABT) {
+		fprintf(stderr,"Late abort model\n");
+	} else {
+		fprintf(stderr,"Early abort model\n");
+	}
+#endif
+	return;
+
+}
+
+static uint32_t
+mtbase_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+	if(crm==0) {
+		return  mmu->mtbase;
+	} else {
+		fprintf(stderr,"mrc MTBASE with unknown crm %08x\n",crm);
+		return 0;
+	}
+}
+static void
+mtbase_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+	mmu->mtbase = value;
+	ttbl_base = value;
+	dprintf("Setting Page table base to %08x\n",value);
+	if(value & 0x3fff) {
+		fprintf(stderr,"Bad page table base 0x%08x\n",value);
+	}
+	invalidate_tlb();
+	return;
+}
+
+static uint32_t
+mdac_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+	return mmu->mdac;
+}
+static void
+mdac_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+	uint32_t diff;
+	dprintf("Load Domain Access Control Register with %08x\n",value);
+	diff = mmu->mdac ^ value;
+	mmu->mdac = value;
+	domain_access_reg = value;
+	if(diff) {
+		invalidate_tlb();
+	}
+	return;
+
+}
+
+static uint32_t
+mfstat_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm==0) {
+		if(opcode_2==0) {
+			return mmu->dfsr;
+		} else if(opcode_2==1) {
+			return mmu->ifsr;
+		} else {
+			fprintf(stderr,"FSR read with unknown opcode_2 %08x\n",opcode_2);
+			return 0;
+		}
+	} else {
+		fprintf(stderr,"mrc MFSTAT with unknown crm %08x\n",crm);
+		exit(1);
+	}
+}
+
+static void
+mfstat_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm!=0) {
+		fprintf(stderr,"write fsr with illegal crm %d\n",crm);
+		return;
+	}
+	if(opcode_2==0) {
+		mmu->dfsr = value;
+	} else if (opcode_2==1)  {
+		mmu->ifsr = value;	
+	} else {
+		fprintf(stderr,"illegal opcode_2 %d to MMU FSR\n",opcode_2);
+	}
+
+}
+static uint32_t
+mfaddr_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+	return mmu->far;
+}
+
+static void
+mfaddr_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(opcode_2==0) {
+		mmu->far = value;
+	} else {
+		fprintf(stderr,"illegal opcode_2 %d to MMU FAR\n",opcode_2);
+	}
+}
+static uint32_t
+mcctrl_read(void *clientData,uint32_t icode) 
+{
+	return FLAG_Z; // ????  
+}
+
+static void
+mcctrl_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	/* Wait for Interrupt instruction (halt) */
+	if((crm==0)&&(opcode_2==4)) {
+		uint64_t nsecs;
+		struct timespec *last_halt,tv_now;
+		last_halt=&mmu->last_halt;
+		clock_gettime(CLOCK_MONOTONIC,&tv_now);
+		nsecs = tv_now.tv_nsec-last_halt->tv_nsec + (int64_t)1000000000*(tv_now.tv_sec - last_halt->tv_sec);
+		mmu->saved_cycles=mmu->saved_cycles - (CycleTimerRate_Get()/1000000) * (nsecs/1000) +
+			(CycleCounter_Get() - mmu->last_halt_cycles);
+		*last_halt=tv_now;
+
+		/* 
+		 * Catch up a maximum of one half of a second if emulator is 
+		 * slower than real system 
+		 */
+		if(mmu->saved_cycles < -(int64_t)(CycleTimerRate_Get()/2)) {
+			mmu->saved_cycles=-(int64_t)(CycleTimerRate_Get()>>1);
+		} else if(mmu->saved_cycles > (int64_t)CycleTimerRate_Get()) {
+			mmu->saved_cycles=CycleTimerRate_Get();
+		}
+		/* Jump over CPU cycles if not enough saved for sleeping 1 timeslice */
+		while(mmu->saved_cycles<(CycleTimerRate_Get()/100) 
+			&& (firstCycleTimerTimeout != ~(uint64_t)0)) {
+
+			/* leave halt even if an IRQ arrives when irqs are disabled in CPSR */
+			if(mainloop_event_pending || (gcpu.signals_raw & (ARM_SIG_IRQ | ARM_SIG_FIQ))) {
+				mmu->last_halt_cycles = CycleCounter_Get();
+				return;
+			}
+			mmu->saved_cycles+= firstCycleTimerTimeout-CycleCounter_Get();
+			CycleCounter=firstCycleTimerTimeout;
+			CycleTimers_Check();
+		}
+		mmu->last_halt_cycles=CycleCounter_Get();
+		/* Jump over time (sleep) if we have saved enough CPU cycles for sleeping 1 timeslice */
+		while(mmu->saved_cycles>=(CycleTimerRate_Get()/100)) {
+			struct timespec tout;					
+			tout.tv_nsec=10000000; /* 10 ms */
+			tout.tv_sec=0;
+			if(mainloop_event_pending || (gcpu.signals_raw & (ARM_SIG_IRQ | ARM_SIG_FIQ))) {
+				return;
+			}
+			FIO_WaitEventTimeout(&tout);
+			clock_gettime(CLOCK_MONOTONIC,&tv_now);
+			nsecs=tv_now.tv_nsec-last_halt->tv_nsec + (int64_t)1000000000*(tv_now.tv_sec - last_halt->tv_sec);	
+			mmu->saved_cycles-=(CycleTimerRate_Get()/1000000) * (nsecs/1000);
+			*last_halt=tv_now;
+		} 
+	} else {
+		dprintf("Ignore Cache settings\n");
+	}
+	return;
+}
+
+static uint32_t
+mtlbctrl_read(void *clientData,uint32_t icode) 
+{
+	fprintf(stderr,"MMU: mtlbctrl register read not implemented\n");
+	return 0;
+	
+}
+
+static void
+mtlbctrl_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	//fprintf(stderr,"inv op2 %08x crm %08x,data %08x\n",opcode_2,crm,data);
+	invalidate_tlb();
+	dprintf("Invalidate TLB\n");
+	return;
+
+}
+
+/*
+ * ---------------------------------------------------
+ * Cache lockdown register
+ * ---------------------------------------------------
+ */
+static uint32_t
+mclck_read(void *clientData,uint32_t icode) 
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm==0) {
+		/* Read Cache lockdown registers */
+		if(opcode_2==0) {
+			return mmu->dclck;
+		} else if (opcode_2==1)  {
+			return mmu->iclck;	
+		} else {
+			fprintf(stderr,"illegal opcode_2 %d to MMU Cache lockdown\n",opcode_2);
+			return 0;
+		}
+	} else if(crm==1) {
+		fprintf(stderr,"MMU TCM not supported\n");	
+		exit(4334);
+	} else {
+		fprintf(stderr,"MMU MCLCK: Illegal value %d for crm\n",crm); 
+		exit(3243);
+	}
+}
+
+
+static void
+mclck_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	SystemCopro *mmu = (SystemCopro *) clientData;
+        uint32_t crm=icode&0xf;
+        uint32_t opcode_2=(icode>>5)&0x7;
+	if(crm==0) {
+		if((value & 0xfffffff0) != 0x0000fff0) {
+			fprintf(stderr,"MMU: illegal value for Cache Lockdown %08x\n",value);
+		}
+		if(opcode_2==0) {
+			mmu->dclck = value;
+		} else if (opcode_2==1)  {
+			mmu->iclck = value;	
+		} else {
+			fprintf(stderr,"illegal opcode_2 %d to MMU Cache lockdown\n",opcode_2);
+		}
+	} else if(crm==1) {
+		fprintf(stderr,"MMU TCM not supported\n");	
+			exit(4334);
+	} else {
+		fprintf(stderr,"MMU MCLCK: Illegal value %d for crm\n",crm); 
+		exit(3243);
+	}
+}
+
+static uint32_t
+mtlblck_read(void *clientData,uint32_t icode) 
+{
+	fprintf(stderr,"TLB locking not emulated\n");
+	return 0;
+}
+
+static void
+mtlblck_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	fprintf(stderr,"TLB locking not emulated\n");
+}
+
+static uint32_t
+mpid_read(void *clientData,uint32_t icode) 
+{
+	return 0;
+}
+
+static void
+mpid_write(void *clientData,uint32_t icode,uint32_t value)
+{
+	fprintf(stderr,"MMU: MVA addressing mode PID not implemented\n");
+}
+
 /* 
  * ---------------------------------------------
  * MMU MRC - Move from ArmCoprocessor to register
  * ---------------------------------------------
  */
 uint32_t 
-MMUmrc(ArmCoprocessor *copro,int opcode_1,int crn,int crm,int opcode_2) {
+MMUmrc(ArmCoprocessor *copro,uint32_t icode) {
 	SystemCopro *mmu = copro->owner;
+        uint32_t crn=(icode>>16)&0xf;
 	if(ARM_SIGNALING_MODE==MODE_USER) {
 		ARM9_Exception(EX_UNDEFINED,4);
 		return 0;
 	}
-	dprintf("Calling MMU to MRC\n");
-	dprintf("Opcode1 %d,crn %d crm %d,Opcode2 %d\n",opcode_1,crn,crm,opcode_2);
-	switch(crn) {
 
-		case SYSCPR_ID:
-			if(crm!=0) {
-				fprintf(stderr,"MMU: CRM is not zero(%d) when reading ID\n",crm);
-				return 0;
-			}
-			if(opcode_2==1) {
-				/* 8/4 kB Harvard Instruction/Data Cache */
-				return 0x1d0d2112;
-			} else if(opcode_2==2) {
-				return 0; // TCM not yet ?
-			}
-			return 0x41069260 | 4; // CPU revision 4
-
-		case  SYSCPR_CTRL:
-			if(crm!=0) {
-				fprintf(stderr,"MMU: CRM is not zero(%d) when reading CTRL\n",crm);
-				return 0;
-			}
-			return (mmu->ctrl & ~0xffff0c78UL)|0x00050078;
-
-		case SYSCPR_MTBASE:
-			if(crm==0) {
-				return  mmu->mtbase;
-			} else {
-				fprintf(stderr,"mrc MTBASE with unknown crm %08x\n",crm);
-				return 0;
-			}
-
-		case SYSCPR_MDAC:
-			return mmu->mdac;
-
-		case SYSCPR_MFSTAT:
-			if(crm==0) {
-				if(opcode_2==0) {
-					return mmu->dfsr;
-				} else if(opcode_2==1) {
-					return mmu->ifsr;
-				} else {
-					fprintf(stderr,"FSR read with unknown opcode_2 %08x\n",opcode_2);
-				}
-			} else {
-				fprintf(stderr,"mrc MFSTAT with unknown crm %08x\n",crm);
-				exit(1);
-			}
-
-		case SYSCPR_MFADDR:
-			return mmu->far;
-
-		case SYSCPR_MCCTRL:
-			return FLAG_Z; // ????  
-
-		case SYSCPR_MCLCK: 
-			if(crm==0) {
-				/* Read Cache lockdown registers */
-				if(opcode_2==0) {
-					return mmu->dclck;
-				} else if (opcode_2==1)  {
-					return mmu->iclck;	
-				} else {
-					fprintf(stderr,"illegal opcode_2 %d to MMU Cache lockdown\n",opcode_2);
-					return 0;
-				}
-			} else if(crm==1) {
-				fprintf(stderr,"MMU TCM not supported\n");	
-				exit(4334);
-			} else {
-				fprintf(stderr,"MMU MCLCK: Illegal value %d for crm\n",crm); 
-				exit(3243);
-			}
-			break;
-			
-		case SYSCPR_MTLBCTRL:
-		case SYSCPR_MTLBLCK: 
-		case SYSCPR_MPID:
-		default:
-			fprintf(stderr,"MMU mrc Register %d not implemented\n",crn);
-			exit(523);
+	if(mmu->mrcProc[crn]) {
+		return mmu->mrcProc[crn](mmu->crn_clientData[crn],icode);
+	} else {
+		fprintf(stderr,"MMU: register %d not implemented\n",crn);
+		return 0;
 	}
+
 }
+
+
+
 /*
  * ---------------------------------------------------------------------
  * MMU MCR - Move from Register to MMU-Coprocessor
  * ---------------------------------------------------------------------
  */
 void 
-MMUmcr(ArmCoprocessor *copro,int opcode_1,uint32_t data,int crn,int crm,int opcode_2) 
+MMUmcr(ArmCoprocessor *copro,uint32_t icode,uint32_t data) 
 {
 	SystemCopro *mmu = copro->owner;
-	uint32_t diff;
-	dprintf("Calling MMU to MCR ");
-	dprintf("Opcode1 %d,crn %d crm %d,Opcode2 %d, data 0x%08x\n",opcode_1,crn,crm,opcode_2,data);
-	if(debugflags) {
-		fprintf(stderr,"Opcode1 %d,crn %d crm %d,Opcode2 %d, data 0x%08x\n",opcode_1,crn,crm,opcode_2,data);
-	}
+        uint32_t crn=(icode>>16)&0xf;
+
 	if(ARM_SIGNALING_MODE==MODE_USER) {
 		ARM9_Exception(EX_UNDEFINED,4);
 		return;
 	}
-	switch(crn) {
-		case SYSCPR_ID:
-			fprintf(stderr,"MMU ID Register is not writable\n");
-			break;
-
-		case SYSCPR_CTRL:
-			dprintf("Setting MMU Control Register to %08x\n",data);
-			if(crm!=0) {
-				fprintf(stderr,"MMU: CRM is not zero(%d) when writing CTRL\n",crm);
-				return;
-			}
-			/* Bit 3 is write buffer enable */
-			diff=mmu->ctrl ^ data;
-			mmu->ctrl=data;
-			if(diff & MCTRL_BE) {
-				update_byteorder(mmu);
-			} 
-			translation_enabled= data & MCTRL_MMUEN;
-			mmu_high_vectors=data & MCTRL_V;
-			do_alignment_check=data & MCTRL_ALGNCHK;
-			sys_rom_protection=((data>>8)&0x3)<<SYS_ROM_SHIFT;
-			if(diff & (3<<8)) {
-				invalidate_tlb();
-			}
-#if 0
-			if(data & MCTRL_LABT) {
-				fprintf(stderr,"Late abort model\n");
-			} else {
-				fprintf(stderr,"Early abort model\n");
-			}
-#endif
-			return;
-
-		case SYSCPR_MTBASE:
-			mmu->mtbase=data;
-			ttbl_base=data;
-			dprintf("Setting Page table base to %08x\n",data);
-			//fprintf(stderr,"Setting Page table base to %08x\n",data);
-			if(data&0x3fff) {
-				fprintf(stderr,"Bad page table base %08x\n",data);
-				exit(3);
-			}
-			invalidate_tlb();
-			return;
-
-		case SYSCPR_MDAC:
-			dprintf("Load Domain Access Control Register with %08x\n",data);
-			//fprintf(stderr,"Load Domain Access Control Register with %08x\n",data);
-			diff=mmu->mdac ^ data;
-			mmu->mdac=data;
-			domain_access_reg = data;
-			if(diff) {
-				invalidate_tlb();
-			}
-			return;
-
-		case  SYSCPR_MFSTAT:
-			if(crm!=0) {
-				fprintf(stderr,"write fsr with illegal crm %d\n",crm);
-				return;
-			}
-			if(opcode_2==0) {
-				mmu->dfsr = data;
-			} else if (opcode_2==1)  {
-				mmu->ifsr = data;	
-			} else {
-				fprintf(stderr,"illegal opcode_2 %d to MMU FSR\n",opcode_2);
-			}
-			break;
-			
-		case  SYSCPR_MFADDR: 
-			if(opcode_2==0) {
-				mmu->far = data;
-			} else {
-				fprintf(stderr,"illegal opcode_2 %d to MMU FAR\n",opcode_2);
-			}
-			break;
-
-		case SYSCPR_MCCTRL:
-			/* Wait for Interrupt instruction (halt) */
-			if((crm==0)&&(opcode_2==4)) {
-				uint64_t nsecs;
-				struct timespec *last_halt,tv_now;
-				last_halt=&mmu->last_halt;
-				clock_gettime(CLOCK_MONOTONIC,&tv_now);
-				nsecs = tv_now.tv_nsec-last_halt->tv_nsec + (int64_t)1000000000*(tv_now.tv_sec - last_halt->tv_sec);
-				mmu->saved_cycles=mmu->saved_cycles - (CycleTimerRate_Get()/1000000) * (nsecs/1000) +
-					(CycleCounter_Get() - mmu->last_halt_cycles);
-				*last_halt=tv_now;
-
-				/* 
-			         * Catch up a maximum of one half of a second if emulator is 
-			         * slower than real system 
-				 */
-				if(mmu->saved_cycles < -(int64_t)(CycleTimerRate_Get()/2)) {
-					mmu->saved_cycles=-(int64_t)(CycleTimerRate_Get()>>1);
-				} else if(mmu->saved_cycles > (int64_t)CycleTimerRate_Get()) {
-					mmu->saved_cycles=CycleTimerRate_Get();
-				}
-				/* Jump over CPU cycles if not enough saved for sleeping 1 timeslice */
-				while(mmu->saved_cycles<(CycleTimerRate_Get()/100) 
-					&& (firstCycleTimerTimeout != ~(uint64_t)0)) {
-
-					/* leave halt even if an IRQ arrives when irqs are disabled in CPSR */
-					if(mainloop_event_pending || (gcpu.signals_raw & (ARM_SIG_IRQ | ARM_SIG_FIQ))) {
-						mmu->last_halt_cycles = CycleCounter_Get();
-						return;
-					}
-					mmu->saved_cycles+= firstCycleTimerTimeout-CycleCounter_Get();
-					CycleCounter=firstCycleTimerTimeout;
-					CycleTimers_Check();
-				}
-				mmu->last_halt_cycles=CycleCounter_Get();
-				/* Jump over time (sleep) if we have saved enough CPU cycles for sleeping 1 timeslice */
-				while(mmu->saved_cycles>=(CycleTimerRate_Get()/100)) {
-					struct timespec tout;					
-					tout.tv_nsec=10000000; /* 10 ms */
-					tout.tv_sec=0;
-					if(mainloop_event_pending || (gcpu.signals_raw & (ARM_SIG_IRQ | ARM_SIG_FIQ))) {
-						return;
-					}
-					FIO_WaitEventTimeout(&tout);
-					clock_gettime(CLOCK_MONOTONIC,&tv_now);
-					nsecs=tv_now.tv_nsec-last_halt->tv_nsec + (int64_t)1000000000*(tv_now.tv_sec - last_halt->tv_sec);	
-					mmu->saved_cycles-=(CycleTimerRate_Get()/1000000) * (nsecs/1000);
-					*last_halt=tv_now;
-				} 
-			} else {
-				dprintf("Ignore Cache settings\n");
-			}
-			return;
-		case SYSCPR_MTLBCTRL:
-			//fprintf(stderr,"inv op2 %08x crm %08x,data %08x\n",opcode_2,crm,data);
-			invalidate_tlb();
-			dprintf("Invalidate TLB\n");
-			return;
-
-		case SYSCPR_MCLCK: 
-			if(crm==0) {
-				if((data&0xfffffff0)!=0x0000fff0) {
-					fprintf(stderr,"MMU: illegal value for Cache Lockdown %08x\n",data);
-				}
-				if(opcode_2==0) {
-					mmu->dclck = data;
-				} else if (opcode_2==1)  {
-					mmu->iclck = data;	
-				} else {
-					fprintf(stderr,"illegal opcode_2 %d to MMU Cache lockdown\n",opcode_2);
-				}
-			} else if(crm==1) {
-				fprintf(stderr,"MMU TCM not supported\n");	
-				exit(4334);
-			} else {
-				fprintf(stderr,"MMU MCLCK: Illegal value %d for crm\n",crm); 
-				exit(3243);
-			}
-			break;
-			
-		case SYSCPR_MTLBLCK:
-		case SYSCPR_MPID:
-		default:
-			fprintf(stderr,"MMU mcr Register %d not implemented\n",crn);
-			exit(123);
+	if(mmu->mcrProc[crn]) {
+		return mmu->mcrProc[crn](mmu->crn_clientData[crn],icode,data);
+	} else {
+		fprintf(stderr,"MMU: mcr of register %d not implemented\n",crn);
 	}
+	
 }
 
 static ArmCoprocessor copro = {
@@ -1123,9 +1298,11 @@ static ArmCoprocessor copro = {
 };
 
 ArmCoprocessor *
-MMU_Create(const char *name,int endian) 
+MMU_Create(const char *name,int endian,uint32_t mmu_type) 
 {
 	SystemCopro * mmu;
+//	uint32_t variant = mmu_type & 0xffff;
+	uint32_t type = mmu_type & 0xffff0000;
 	char *nodename = alloca(strlen(name)+30);
 	gmmu=mmu=malloc(sizeof(SystemCopro));
 	if(!mmu) {
@@ -1150,6 +1327,28 @@ MMU_Create(const char *name,int endian)
 	update_byteorder(mmu);
 	copro.owner=mmu;
 	stlb_init();
+	switch(type) {
+		case MMU_ARM926EJS:
+			CrnHandler_New(mmu,SYSCPR_ID,arm926ejs_id_read,id_write,mmu);
+			break;
+		case MMU_ARM920T:
+			CrnHandler_New(mmu,SYSCPR_ID,arm920t_id_read,id_write,mmu);
+			break;
+		default:
+			fprintf(stderr,"Nonexisting MMU type %08x\n",type);
+			break;
+	}
+	CrnHandler_New(mmu,SYSCPR_CTRL,ctrl_read,ctrl_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MTBASE,mtbase_read,mtbase_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MDAC,mdac_read,mdac_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MFSTAT,mfstat_read,mfstat_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MFADDR,mfaddr_read,mfaddr_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MCCTRL,mcctrl_read,mcctrl_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MTLBCTRL,mtlbctrl_read,mtlbctrl_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MCLCK,mclck_read,mclck_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MTLBLCK,mtlblck_read,mtlblck_write,mmu);
+	CrnHandler_New(mmu,SYSCPR_MPID,mpid_read,mpid_write,mmu);
+
 	return &copro;
 }
 
