@@ -331,9 +331,8 @@ update_mii_state(NS9750eth *eth,int verbose) {
 	uint32_t clks = MCFG_CLKS(eth->mcfg);
 	int enable=1;
 	if(clks < 6) {
-		enable=0;
 		if(verbose) {
-			fprintf(stderr,"CLKS is <6\n");
+			fprintf(stderr,"NS9750Eth CLKS: %d < 6, Your MII is overclocked\n",clks);
 		}
 	} 
 	if(eth->mcfg & MCFG_RMIM) {
@@ -2305,7 +2304,7 @@ int
 NS9750_EthRegisterPhy(BusDevice *dev,PHY_Device *phy,unsigned int phy_addr) 
 {
 	NS9750eth *eth=dev->owner;
-	if(phy_addr>MAX_PHYS) {
+	if(phy_addr >= MAX_PHYS) {
 		fprintf(stderr,"NS9750 Eth:Illegal PHY address %d\n",phy_addr);
 		return -1;
 	} else {
