@@ -979,7 +979,7 @@ PCIBr_Read32(uint32_t addr) {
 }
 
 static void 
-PCIBr_Read(uint32_t addr,uint8_t *buf,uint32_t count) 
+PCIBr_Read(uint8_t *buf,uint32_t addr,uint32_t count) 
 {
 	if(gcpu_endian == TARGET_BYTEORDER) {
 		return Bus_Read(buf,addr,count);
@@ -988,8 +988,10 @@ PCIBr_Read(uint32_t addr,uint8_t *buf,uint32_t count)
 	}
 }
 
+// Bus *
 PCI_Function *
 NS9750_PciInit(const char *devname,int dev_nr) {
+//	Bus *bus = bla ;
 	PCI_Function *bridgefunc;
 	bridgefunc=BR_New(devname,dev_nr);
 	bridgefunc->bdev.first_mapping=NULL;
