@@ -34,6 +34,7 @@
 #include "ns9750_timer.h"
 #include "configfile.h"
 #include "signode.h"
+#include "senseless.h"
 #if 0 
 #define dprintf(x...) { fprintf(stderr,x); }
 #else
@@ -72,6 +73,7 @@ typedef struct Timer {
 	uint32_t trr;  // timer read
 	uint32_t tcr;  // timer control
 	CycleCounter_t last_update;
+	CycleCounter_t saved_cycles;
 	CycleTimer irq_timer;
 } Timer;
 
@@ -209,16 +211,10 @@ Sysco_UnPostIrq(int isource_nr) {
 		}
 	}
 }
-#if 0
-int 
-SyscoSetIrqLine(int isource_nr) {
-	if(isource_nr>=IRQ_EXTERNAL_0)
-}
-#endif
 
 /* 
  * --------------------------------
- * Timer Read Register
+ * Timer Read Register 
  * --------------------------------
  */
 void 
@@ -228,40 +224,42 @@ update_trr(Timer *timer) {
 	uint32_t tm;
 	uint32_t down;
 	uint32_t reload;
-	uint32_t divider;
+	unsigned int shift;
 	uint32_t timer32;
-	uint64_t cycles;
+	uint64_t count;
 	if(!(tcr & SYS_TCR_TEN)) {
 		return;
 	}
-	tlcs=((tcr&SYS_TCR_TLCS)>>SYS_TCR_TLCS_SHIFT);
+	tlcs = ((tcr & SYS_TCR_TLCS) >> SYS_TCR_TLCS_SHIFT);
 	if(tlcs==7) {
 		fprintf(stderr,"External Pulse Events not implemented\n");
 		return;
 	} else {
-		divider=1<<tlcs;
+		shift = tlcs;
 	}
-	tm=((tcr&SYS_TCR_TM)>>SYS_TCR_TM_SHIFT);
-	down=tcr&SYS_TCR_UDS;
-	timer32=tcr&SYS_TCR_TSZ;
-	reload=tcr&SYS_TCR_REN;
+	tm = ((tcr & SYS_TCR_TM)>>SYS_TCR_TM_SHIFT);
+	down = tcr & SYS_TCR_UDS;
+	timer32 = tcr & SYS_TCR_TSZ;
+	reload = tcr & SYS_TCR_REN;
 	if(!timer32) {
 		fprintf(stderr,"16 Bit Counters not implemented\n");
 	}
 	switch(tm) {
 		case 0:
-			cycles=(CycleCounter_Get()-timer->last_update)/divider;
+			timer->saved_cycles += CycleCounter_Get() - timer->last_update;
+			count = timer->saved_cycles >> shift;
+			timer->saved_cycles -= count << shift;
 			if(down) {
-				cycles=timer->trr-cycles;
+				count = timer->trr - count;
 			} else {
-				cycles=timer->trr+cycles;
+				count = timer->trr + count;
 			}
-			if(cycles&0xffffffff00000000LL) {
+			if(count & 0xffffffff00000000LL) {
 				if(!reload) {
 					if(down) {
-						cycles=0;
+						count = 0;
 					} else {
-						cycles=0xffffffff;
+						count =0xffffffff;
 					}
 
 				} else {
@@ -270,25 +268,25 @@ update_trr(Timer *timer) {
 						uint32_t div = timer->trcv;
 						uint32_t n;
 						if(div!=0) {
-							n=1+~cycles/div;
-							cycles=cycles+n*div;
+							n=1+~count/div;
+							count=count+n*div;
 						} else {
-							cycles=0;
+							count=0;
 						}
 					} else {
 						// overflow on up
 						uint32_t div = 0xffffffff-timer->trcv;	
 						int n;
 						if(div!=0) {
-							n=cycles/div;
-							cycles=-(cycles-n*div);
+							n=count/div;
+							count=-(count-n*div);
 						} else {
-							cycles=0xffffffff;
+							count=0xffffffff;
 						}
 					}
 				}	
 			}
-			timer->trr=cycles;
+			timer->trr = count;
 			break;
 			
 		default:
@@ -329,7 +327,6 @@ calculate_remaining(Timer *timer) {
 	uint32_t timer32=tcr&SYS_TCR_TSZ;
 	uint32_t ticks;	
 	uint32_t tlcs;
-	uint32_t divider;
 	if(timer32) {
 		if(down) {
 			ticks=timer->trr;
@@ -348,19 +345,17 @@ calculate_remaining(Timer *timer) {
 	if(tlcs==7) {
 		fprintf(stderr,"External Pulse Events not implemented\n");
 		return 0;
-	} else {
-		divider=1<<tlcs;
-		dprintf("divider is %d\n",divider);
 	}
 	ticks+=1;
 //	fprintf(stderr,"ticks is %d, trr is %d\n",ticks,timer->trr);
-	return ticks*divider;
+	return ticks << tlcs;
 }
 
 static uint32_t 
 timer_trr_read(void *cd,uint32_t address,int rqlen) {
 	Timer *timer=(Timer*)cd;
 	update_trr(timer);
+	Senseless_Report(200);
 	dprintf("Timer Read Register %08x\n",timer->trr);
 	return timer->trr;
 }
@@ -398,7 +393,8 @@ timer_tcr_write(void *cd,uint32_t new_tcr,uint32_t address,int rqlen) {
 	uint64_t remaining;
 	Timer *timer=(Timer*)cd;
 	uint32_t old_tcr=timer->tcr;
-	uint32_t diff_tcr=old_tcr ^ new_tcr;
+	uint32_t diff_tcr = old_tcr ^ new_tcr;
+	int new_tlcs;
 
 	dprintf("New Timer Control of %08x\n",new_tcr);
 
@@ -424,8 +420,11 @@ timer_tcr_write(void *cd,uint32_t new_tcr,uint32_t address,int rqlen) {
 		dprintf("Removed Timer, remaining\n");
 		CycleTimer_Remove(&timer->irq_timer);
 	}
-	timer->tcr=new_tcr;
-	
+	new_tlcs = ((new_tcr & SYS_TCR_TLCS) >> SYS_TCR_TLCS_SHIFT);
+	if(new_tlcs < 7) {
+		timer->saved_cycles  &= ((1 << new_tlcs) - 1);
+	}
+	timer->tcr = new_tcr;
 	return;
 }
 
@@ -639,13 +638,6 @@ sys_isra_read(void *cd,uint32_t address,int rqlen) {
 static void 
 sys_isra_write(void *cd,uint32_t value,uint32_t address,int rqlen) {
 	int irq=sysco->sys_intid;
-#ifdef DEBUG_KERNEL 
-	if(value!=irq) {
-		fprintf(stderr,"PC: %08x wrong irq acknowledged: %d instead %d irqmask %08x, old_irqmask %08x, intid %d oldintid %d\n",
-			REG_PC,value,irq,sysco->irqmask,sysco->old_irqmask[irq],sysco->sys_intid,sysco->old_intid[irq]);
-		exit(175);
-	}
-#endif
 	if(irq<32) {
 		if(!testbit(irq,sysco->irqmask)) {
 			sysco->irqmask=sysco->old_irqmask[irq];
