@@ -219,14 +219,14 @@ update_clock(NS_I2C *i2c,unsigned int clref,int speed)
 	uint32_t i2c_bus_clock;  
 	uint32_t period;
 	I2C_Timing *timing = &i2c->i2c_timing;
-	int scl_delay = clk * 50e-9;
+	int scl_delay = (int)(clk * 50e-9);
 	if(i2c->cfg & CFG_VSCD) {
 		clk = clk / 8;	
 	}
 	if(speed == I2C_SPEED_STD) {
 		i2c_bus_clock = clk / ((2*clref)+4+scl_delay);
 	} else if(speed == I2C_SPEED_FAST) {
-		i2c_bus_clock = (3./4.) * (clk/((clref * 2)+4+scl_delay));
+		i2c_bus_clock = 3 * (clk/((clref * 2)+4+scl_delay)) / 4;
 		i2c_bus_clock *= 2; /* Suspect bug in manual */
 	} else {
 		return;
@@ -254,6 +254,15 @@ update_clock(NS_I2C *i2c,unsigned int clref,int speed)
 		timing->t_susto = period * 4000 / 10000;
 		timing->t_buf = period * 4700 / 10000;
 	} else if(speed == I2C_SPEED_FAST) {
+		/* 
+ 		 * ----------------------------------------------------
+		 * Better use the controller in STD with 
+		 * faster clock if you want to have high speed 
+ 		 * Netsilicon has shortened high time because it
+		 * is shorter in the spec. But the shorter high
+		 * time in spec is because of longer rise times.
+ 		 * ----------------------------------------------------
+		 */ 
         	timing->t_hdsta = period * 600 / 2500;
         	timing->t_low = period * 1300 / 2500;
 		timing->t_high = period * 600 / 2500;
