@@ -416,10 +416,10 @@ i_mcrf(uint32_t icode) {
 	uint32_t mask;
 	int crfd = 7-((icode>>23)&7); // shit PPC bit counting 
 	int crfs = 7-((icode>>18)&7);
-	uint32_t or;
-	or=((CR>>(4*crfs))&0xf) << (4*crfd);
+	uint32_t setbits;
+	setbits=((CR>>(4*crfs))&0xf) << (4*crfd);
 	mask = ~(0xf << (4*crfd));
-	CR = (CR & mask) | or;
+	CR = (CR & mask) | setbits;
 	dprintf("instr i_mcrf(%08x)\n",icode);
 }
 
@@ -3783,13 +3783,13 @@ void
 i_mcrfs(uint32_t icode) {
 	uint32_t mask;
 	uint32_t clear;
-	uint32_t or;
+	uint32_t setbits;
 	int crfd = 7-((icode>>23)&7);
 	int crfs = 7-((icode>>18)&7);
 	clear = FPSCR & (0xf << (4*crfs)) & FPSCR_EXCEPTIONS;
-	or=((FPSCR>>(4*crfs))&0xf) << (4*crfd);
+	setbits=((FPSCR>>(4*crfs))&0xf) << (4*crfd);
 	mask = ~(0xf << (4*crfd));
-	CR = (CR & mask) | or;
+	CR = (CR & mask) | setbits;
 	/* Clear the exception bits except FEX and VX */
 	FPSCR = FPSCR ^ clear;	
 	//update_fex_vx(); /* Summary bits */
