@@ -43,6 +43,7 @@
 #include <configfile.h>
 #include <fbdisplay.h>
 #include <keyboard.h>
+#include <zlib.h>
 
 #if 0
 #define dbgprintf(x...) { fprintf(stderr,x); }
@@ -99,14 +100,19 @@ typedef struct FrameBufferInfo {
 	char name_string[MAX_NAME_LEN];
 } FrameBufferInfo;
 
-typedef struct UpdateRequest {
+typedef struct UpdateRectangle {
 	uint16_t x;
 	uint16_t y;
 	uint16_t width;
 	uint16_t height;
-	//struct UpdateRequest *next;
-} UpdateRequest;
+} UpdateRectangle;
 
+typedef struct RLEncoder {
+	uint32_t pixval;
+	int run_length;
+} RLEncoder;
+
+#define UDRECT_FIFOSIZE (256)
 /*
  * -----------------------------------------------------------------------------
  * RfbConnection structure contains the state information 
@@ -127,9 +133,16 @@ typedef struct RfbConnection {
 	uint8_t ibuf[IBUFSIZE];
 	int	ibuf_wp;
 	int 	ibuf_expected;
-	/* to lazy for a list currently */
-	UpdateRequest update_request; 
+	RLEncoder rle;
+	z_stream zs;	
 	RfbServer *rfbserv;
+	int update_outstanding;
+	UpdateRectangle udrect_fifo[UDRECT_FIFOSIZE];
+	uint64_t udrect_wp;
+	uint64_t udrect_rp;
+	uint8_t *obuf;
+	int obuf_wp;
+	int obuf_size;
 } RfbConnection;
 
 struct RfbServer {
@@ -156,6 +169,7 @@ onecount(uint16_t value) {
 	return ones;
 }
 
+
 /*
  * ----------------------------------------------------------------
  * Update the bit count fields when the maxval has changed 
@@ -202,6 +216,11 @@ rfbsrv_disconnect(RfbConnection *rcon)
 			}	
 		} 
 	}
+	deflateEnd(&rcon->zs);
+	if(rcon->obuf) {
+		free(rcon->obuf);
+	}
+	free(rcon);
 }
 
 /*
@@ -216,7 +235,7 @@ rfbsrv_disconnect(RfbConnection *rcon)
  * ----------------------------------------------------------
  */
 static int 
-rfbcon_send(RfbConnection *rcon,char *buf,int count) 
+rfbcon_send(RfbConnection *rcon,void *buf,int count) 
 {
 	int result;
 	int counter = count;
@@ -227,7 +246,7 @@ rfbcon_send(RfbConnection *rcon,char *buf,int count)
 		if(wsize>10)
 			wsize=10;
 #endif
-                result = write(rcon->sock_fd,buf,wsize);
+                result = write(rcon->sock_fd,(uint8_t *)buf,wsize);
                 if(result<=0) {
 			fprintf(stderr,"Connection to vncviewer broken, closing\n");
 			/* dangerous, maybe delayed is better */ 
@@ -349,117 +368,408 @@ Msg_ServerInitialisation(RfbConnection *rcon)
 #define write16(addr,value) (*(uint16_t*)(addr) = (value))
 #define write32(addr,value) (*(uint32_t*)(addr) = (value))
 
+/*
+ * ------------------------------------------------------------------------------
+ * Convert a single pixel stored in a src buffer from the framebuffer
+ * pixel format into a destination pixel format which is connection specific
+ * ------------------------------------------------------------------------------
+ */
+
+static inline uint32_t 
+encode_pixval(void *src,PixelFormat *pixf,PixelFormat *fbpixf) {
+	uint32_t pixval;
+	int red,green,blue;
+	switch(fbpixf->bits_per_pixel) {
+		case 32: 
+			pixval = le32_to_host(*(uint16_t*)src); 
+			break;
+		case 16: 
+			pixval = le16_to_host(*(uint16_t*)src); 
+			break;
+		case 8: 
+			pixval = *(uint8_t*)src; 
+			break;
+		default:
+			pixval = 0;
+	}
+	red = (pixval >> fbpixf->red_shift) & fbpixf->red_max;
+	green = (pixval >> fbpixf->green_shift) & fbpixf->green_max;
+	blue = (pixval >> fbpixf->blue_shift) & fbpixf->blue_max;
+	if(fbpixf->red_bits > pixf->red_bits) {
+		red = red >> (fbpixf->red_bits - pixf->red_bits); 
+	} else {
+		red = red << (pixf->red_bits - fbpixf->red_bits); 
+	}
+	if(fbpixf->green_bits > pixf->green_bits) {
+		green = green >> (fbpixf->green_bits - pixf->green_bits);
+	} else {
+		green = green << (pixf->green_bits - fbpixf->green_bits);
+	}
+	if(fbpixf->blue_bits > pixf->blue_bits) {
+		blue = blue >> (fbpixf->blue_bits - pixf->blue_bits);
+	} else {
+		blue = blue << (pixf->blue_bits - fbpixf->blue_bits);
+	}
+	pixval = (red << pixf->red_shift) | (green <<  pixf->green_shift) |
+		(blue << pixf->blue_shift);
+
+	return pixval;
+}
+
+static inline int 
+write_pixel(int con_bypp,int swap,uint8_t *dst,uint32_t pixval) 
+{
+	switch(con_bypp) {
+		case 1:
+			dst[0]=pixval;
+			return 1;
+		case 2:
+			if(swap) {
+				write16(dst,swap16(pixval));
+			} else {
+				write16(dst,pixval);
+			}
+			return 2;
+		case 3:
+			if(swap) {
+				uint8_t *pix = (uint8_t *) &pixval;	
+				dst[0] = pix[3];
+				dst[1] = pix[2];
+				dst[2] = pix[1];
+			} else {
+				/* Do not use write32 because of alignment traps */
+				uint8_t *pix = (uint8_t *) &pixval;	
+				dst[0] = pix[0];
+				dst[1] = pix[1];
+				dst[2] = pix[2];
+			}
+			return 3;
+		case 4:
+			if(swap) {
+				write32(dst,swap32(pixval));
+			} else {
+				write32(dst,pixval);
+			}
+			return 4;
+		default:
+			fprintf(stderr,"%d bytes per pixel not implemented\n",con_bypp);
+			return 0;
+	}
+}
+
+static inline int 
+fill_update_header(uint8_t *hdr,int rects) 
+{
+	hdr[0] = SRV_FB_UPDATE;
+	hdr[1] = 0;
+	write16be(hdr+2,rects); /* Number Rectangles */
+	return 4;
+}
 
 static inline void 
-srv_fb_update_raw(RfbConnection *rcon,UpdateRequest *udrq) 
+rle_init(RLEncoder *rle)
+{
+	rle->run_length = 0;
+}
+
+static inline int
+rle_add_pixval(RLEncoder *rle,int con_bypp,int swap,uint8_t *dst,uint32_t pixval) 
+{
+	if(rle->run_length == 0) {
+		rle->pixval = pixval;	
+		rle->run_length++;
+		return 0;
+	} else {
+		if(rle->pixval == pixval) {
+			rle->run_length++;
+			return 0;
+		} else {
+			int count;
+			count = write_pixel(con_bypp,swap,dst,rle->pixval);
+			dst+=count;
+			//fprintf(stderr,"Pixel %08x, rl %d cnt ",rle->pixval,rle->run_length);
+			rle->run_length-=1;
+			while(rle->run_length >= 255) {
+				*dst++ = 255;
+				//fprintf(stderr,"%d %02x ",count,255);
+				count++;
+				rle->run_length-=255;
+			}
+			*dst = rle->run_length;
+			//fprintf(stderr,"%d %02x ",count,*dst);
+			count++;
+			rle->pixval = pixval;
+			rle->run_length = 1;
+			return count; 
+		}
+	}
+}
+
+static inline int
+rle_flush(RLEncoder *rle,int con_bypp,int swap,uint8_t *dst) 
+{
+	int retval;
+	retval = rle_add_pixval(rle,con_bypp,swap,dst,rle->pixval ^ 0xffffffff);
+	rle->run_length = 0;
+	//fprintf(stderr,"Flushed RLE\n");
+	return retval;
+}
+
+static inline void 
+srv_fb_encode_update_raw(RfbConnection *rcon,UpdateRectangle *udrect) 
 {
 	FrameBufferInfo *fbi = rcon->fbi;
 	PixelFormat *pixf = &rcon->pixfmt;
 	PixelFormat *fbpixf = &fbi->pixfmt;
 	int con_bypp = pixf->bits_per_pixel >> 3;
 	int fb_bypp = fbi->pixfmt.bits_per_pixel >> 3;
-	int size = udrq->width * udrq->height * con_bypp;
-	char *reply = alloca(size + 16);
-	char *data;
+	int memsize = udrect->width * udrect->height * con_bypp + 16;
+	uint8_t *reply; 
+	uint8_t *data;
 	int x,y;
-	int red,green,blue;
-	reply[0] = SRV_FB_UPDATE;
-	reply[1] = 0;
-	write16be(reply+2,1);
-	write16be(reply+4,udrq->x);
-	write16be(reply+6,udrq->y);
-	write16be(reply+8,udrq->width);
-	write16be(reply+10,udrq->height);
-	write32(reply+12,ENC_RAW);
-	data = reply+16;
-	int swap =  (pixf->big_endian_flag != fbpixf->big_endian_flag);
-	for(y = udrq->y; y < (udrq->y+udrq->height);y++) {
-		int ofs = (y * fbi->fb_width + udrq->x) * fb_bypp;
-		for(x=udrq->x; x < (udrq->x + udrq->width);x++) {
-			uint32_t pixval = *(uint16_t*)(&fbi->framebuffer[ofs]); 
-			red = (pixval >> fbpixf->red_shift) & fbpixf->red_max;
-			green = (pixval >> fbpixf->green_shift) & fbpixf->green_max;
-			blue = (pixval >> fbpixf->blue_shift) & fbpixf->blue_max;
-			if(fbpixf->red_bits > pixf->red_bits) {
-				red = red >> (fbpixf->red_bits - pixf->red_bits); 
-			} else {
-				red = red << (pixf->red_bits - fbpixf->red_bits); 
+	if(rcon->obuf_size < memsize) {
+		rcon->obuf_size = memsize;
+		rcon->obuf = realloc(rcon->obuf,rcon->obuf_size);
+	}
+	reply = rcon->obuf;
+	data = reply;
+	data += fill_update_header(data,rcon->udrect_wp-rcon->udrect_rp);
+	while(rcon->udrect_rp < rcon->udrect_wp) {	
+		udrect = &rcon->udrect_fifo[rcon->udrect_rp % UDRECT_FIFOSIZE];
+		rcon->udrect_rp++;
+		write16be(data,udrect->x);
+		write16be(data+2,udrect->y);
+		write16be(data+4,udrect->width);
+		write16be(data+6,udrect->height);
+		write32be(data+8,ENC_RAW);
+		data += 12;
+
+		int swap =  (pixf->big_endian_flag != fbpixf->big_endian_flag);
+		for(y = udrect->y; y < (udrect->y+udrect->height);y++) {
+			int ofs = (y * fbi->fb_width + udrect->x) * fb_bypp;
+			for(x=udrect->x; x < (udrect->x + udrect->width);x++) {
+				uint32_t pixval = encode_pixval(&fbi->framebuffer[ofs],pixf,fbpixf); 
+				data += write_pixel(con_bypp,swap,data,pixval);
+				ofs += fb_bypp;
 			}
-			if(fbpixf->green_bits > pixf->green_bits) {
-				green = green >> (fbpixf->green_bits - pixf->green_bits);
-			} else {
-				green = green << (pixf->green_bits - fbpixf->green_bits);
-			}
-			if(fbpixf->blue_bits > pixf->blue_bits) {
-				blue = blue >> (fbpixf->blue_bits - pixf->blue_bits);
-			} else {
-				blue = blue << (pixf->blue_bits - fbpixf->blue_bits);
-			}
-			pixval = (red << pixf->red_shift) | (green <<  pixf->green_shift) |
-				(blue << pixf->blue_shift);
-			switch(con_bypp) {
-				case 1:
-					data[0]=pixval;
-					break;
-				case 2:
-					if(swap) {
-						write16(data,swap16(pixval));
-					} else {
-						//fprintf(stderr,"%d(%d).%d(%d).%d(%d) %04x\n",red,green,blue,pixval);
-						write16(data,pixval);
-					}
-					break;
-				case 4:
-					if(swap) {
-						write32(data,swap32(pixval));
-					} else {
-						write32(data,pixval);
-						//fprintf(stderr,"Send pixval %08x\n",pixval);
-					}
-					break;
-				default:
-					fprintf(stderr,"%d bytes per pixel not implemented\n",con_bypp);
-			}
-			ofs += fb_bypp;
-			data += con_bypp;
 		}
+		rfbcon_send(rcon,reply,data-reply);
+		data = reply;
 	}
-	rfbcon_send(rcon,reply,size+16);
 	return;
 }
 
+#if 0
+static void
+zrle_tile(RfbConnection *rcon,UpdateRectangle *udrect,int x0,int y0) {
+	RLEncoder *rle = &rcon->rle;
+	uint8_t *tileBuf = alloca(1 + 64*64*5);
+	uint8_t *tileBufEnd;
+	int x1,y1,y,ofs;
+	int fb_bypp = rcon->fbi->pixfmt.bits_per_pixel >> 3;
+	tileBufEnd = tileBuf;
+	*tileBufEnd++ = 128;  /* Plain RLE */
+	rle_init(rle);
+	for(y1=0;(y1 < 64) && ((y0+y1) < udrect->height);y1++) { 
+		y = y0 + y1 + udrect->y;
+		ofs = (y * rcon->fbi->fb_width + (x0+udrect->x)) * fb_bypp;
+		for(x1=0;(x1 < 64) && ((x1+x0) < udrect->width);x1++) {
+			uint32_t pixval = encode_pixval(&rcon->fbi->framebuffer[ofs],pixf,fbpixf); 
+			tileBufEnd += rle_add_pixval(rle,con_bypp,swap,tileBufEnd,pixval);
+			ofs += fb_bypp;
+		}	
+	}
+	//fprintf(stderr,"Tile at %d %d, %d %d\n",x0+udrect->x,y0+udrect->y,x1,y1);
+	tileBufEnd += rle_flush(rle,con_bypp,swap,tileBufEnd);
+	zs->next_in = tileBuf;  
+	zs->avail_in = tileBufEnd - tileBuf;  
+	deflate(zs,Z_SYNC_FLUSH);
+	if(zs->avail_out < 32768) {
+		rcon->obuf_size *= 2;
+		rcon->obuf = realloc(rcon->obuf,rcon->obuf_size);
+		zs->next_out = rcon->obuf + rcon->obuf_wp + zs->total_out;
+		zs->avail_out = rcon->obuf_size - (rcon->obuf_wp + zs->total_out);
+		deflate(zs,Z_SYNC_FLUSH);
+	}
+}
+#endif
+
+/*
+ * ----------------------------------------------------------------------------
+ * srv_fb_encode_update_zrle
+ * 	create zrle encoded update request. Currently not working because 
+ * 	compression is missing
+ * ----------------------------------------------------------------------------
+ */
 static inline void 
-srv_fb_update(RfbConnection *rcon,UpdateRequest *udrq) 
+srv_fb_encode_update_zrle(RfbConnection *rcon,UpdateRectangle *udrect) 
 {
-	if(rcon->state != CONSTAT_IDLE) {
+	FrameBufferInfo *fbi = rcon->fbi;
+	PixelFormat *pixf = &rcon->pixfmt;
+	PixelFormat *fbpixf = &fbi->pixfmt;
+	int con_bypp = pixf->bits_per_pixel >> 3;
+	int fb_bypp = fbi->pixfmt.bits_per_pixel >> 3;
+	uint8_t *tileBuf = alloca(1 + 64*64*5);
+	uint8_t *tileBufEnd;
+	int lengthP;
+	int ofs;
+	int x0,y0,x1,y1,y;
+	RLEncoder *rle = &rcon->rle;
+	z_stream *zs = &rcon->zs;
+	rcon->obuf_wp = fill_update_header(rcon->obuf,rcon->udrect_wp - rcon->udrect_rp);
+	while(rcon->udrect_rp < rcon->udrect_wp) {	
+		udrect = &rcon->udrect_fifo[rcon->udrect_rp % UDRECT_FIFOSIZE];
+		rcon->udrect_rp++;
+		write16be(rcon->obuf+rcon->obuf_wp,udrect->x);
+		write16be(rcon->obuf+rcon->obuf_wp+2,udrect->y);
+		write16be(rcon->obuf+rcon->obuf_wp+4,udrect->width);
+		write16be(rcon->obuf+rcon->obuf_wp+6,udrect->height);
+		write32be(rcon->obuf+rcon->obuf_wp+8,ENC_ZRLE);
+		rcon->obuf_wp += 12;
+		if(con_bypp == 4) {
+			con_bypp = 3;
+		}
+		lengthP = rcon->obuf_wp; rcon->obuf_wp+=4;
+		zs->next_out = rcon->obuf+rcon->obuf_wp;
+		zs->avail_out = rcon->obuf_size - rcon->obuf_wp;
+		zs->total_out = 0;
+		//fprintf(stderr,"con bypp %d, bits per pixel %d\n",con_bypp,pixf->bits_per_pixel);
+		int swap =  (pixf->big_endian_flag != fbpixf->big_endian_flag);
+		for(y0 = 0; y0 < udrect->height; y0+=64) {
+			for(x0 = 0;x0 < udrect->width;x0+=64) {
+				/* One Tile */
+				tileBufEnd = tileBuf;
+				*tileBufEnd++ = 128;  /* Plain RLE */
+				rle_init(rle);
+				for(y1=0;(y1 < 64) && ((y0+y1) < udrect->height);y1++) { 
+					y = y0 + y1 + udrect->y;
+					ofs = (y * fbi->fb_width + (x0+udrect->x)) * fb_bypp;
+					for(x1=0;(x1 < 64) && ((x1+x0) < udrect->width);x1++) {
+						uint32_t pixval = encode_pixval(&fbi->framebuffer[ofs],pixf,fbpixf); 
+						tileBufEnd += rle_add_pixval(rle,con_bypp,swap,tileBufEnd,pixval);
+						ofs += fb_bypp;
+					}	
+				}
+				//fprintf(stderr,"Tile at %d %d, %d %d\n",x0+udrect->x,y0+udrect->y,x1,y1);
+				tileBufEnd += rle_flush(rle,con_bypp,swap,tileBufEnd);
+				zs->next_in = tileBuf;  
+				zs->avail_in = tileBufEnd - tileBuf;  
+				deflate(zs,Z_SYNC_FLUSH);
+				if(zs->avail_out < 32768) {
+					rcon->obuf_size *= 2;
+					rcon->obuf = realloc(rcon->obuf,rcon->obuf_size);
+					zs->next_out = rcon->obuf + rcon->obuf_wp + zs->total_out;
+					zs->avail_out = rcon->obuf_size - (rcon->obuf_wp + zs->total_out);
+				}
+			}
+		}
+		//fprintf(stderr,"total out %lu av out %lu bpp %d bytes %d\n",zs->total_out,zs->avail_out,fbpixf->bits_per_pixel,con_bypp);
+		write32be(rcon->obuf+lengthP,zs->total_out);
+		rcon->obuf_wp+=zs->total_out;
+		rfbcon_send(rcon,rcon->obuf,rcon->obuf_wp);
+		rcon->obuf_wp = 0;
+	}
+	return;
+}
+
+
+static inline void 
+srv_fb_update(RfbConnection *rcon,UpdateRectangle *udrect) 
+{
+	FrameBufferInfo *fbi = rcon->fbi;
+	if((rcon->state != CONSTAT_IDLE)) {
 		return; 
 	}
+	if(((udrect->y + udrect->height) > fbi->fb_height) || 
+	   ((udrect->x + udrect->width) > fbi->fb_width)) {
+	   fprintf(stderr,"RFB-Server received update outside of window\n");
+	   return;
+	}  
+	//fprintf(stderr,"SRV update display\n"); // jk
 	switch(rcon->current_encoding) {
 		case ENC_RAW:
-			srv_fb_update_raw(rcon,udrq);
+			srv_fb_encode_update_raw(rcon,udrect);
+			break;
+		case ENC_ZRLE:
+			srv_fb_encode_update_zrle(rcon,udrect);
 			break;
 		case ENC_HEXTILE:
 		default:
 			fprintf(stderr,"encoding %d not implemented\n",rcon->current_encoding);
 	}	
 }
-#if 0
-static void 
-srv_set_colour_map_entries() 
-{
 
-}
-static void 
-srv_bell() 
+static inline void 
+trigger_fb_update(RfbConnection *rcon) 
 {
-
+	if((rcon->udrect_wp != rcon->udrect_rp) && rcon->update_outstanding) {
+		UpdateRectangle * udrect;
+		udrect = &rcon->udrect_fifo[rcon->udrect_rp % UDRECT_FIFOSIZE];
+		rcon->update_outstanding = 0;
+		srv_fb_update(rcon,udrect);
+	}
 }
-static void 
-src_cut_text() 
+
+/*
+ * ----------------------------------------------------------------------
+ * Expand a color with less than 16 bit to a color with 16 Bit by
+ * Repeating the bit pattern
+ * ----------------------------------------------------------------------
+ */
+static int
+expand_color_to_16(int color, int bits)
 {
-
+        int result=0;
+        int bitcount = 16;
+        if(!bits) {
+                return 0;
+        }
+        while(bitcount>0) {
+                if(bitcount-bits > 0) {
+                        result = result | (color << (bitcount-bits));
+                } else {
+                        result = result | (color >> (bits -bitcount));
+                }
+                bitcount-=bits;
+        }
+        return result;
 }
-#endif
+
+/*
+ * For 8 Bit Pseudo color
+ */
+static void
+srv_set_8bit_color_map_entries(RfbConnection *rcon,PixelFormat *pixf)
+{
+        uint8_t reply[256*6+100];
+        uint8_t *wp;
+        int i;
+        pixf->red_max = 7;
+        pixf->red_shift = 5;
+        pixf->green_max = 7;
+        pixf->green_shift = 2;
+        pixf->blue_max = 3;
+        pixf->blue_shift = 0;
+        pixfmt_update_bits(pixf);
+        wp = reply;
+        *wp++=SRV_SET_COLOUR_MAP_ENTRIES;
+        *wp++=0;
+        *wp++=0; *wp++=0; // first color
+        write16be(wp,256);wp+=2;
+        for(i=0;i<256;i++) {
+                uint16_t red = (i >> pixf->red_shift) & pixf->red_max;
+                uint16_t green = (i >> pixf->green_shift) & pixf->green_max;
+                uint16_t blue = (i >> pixf->blue_shift) & pixf->blue_max;
+                red = expand_color_to_16(red,pixf->red_bits);
+                green = expand_color_to_16(green,pixf->green_bits);
+                blue = expand_color_to_16(blue,pixf->blue_bits);
+                write16be(wp,red);wp+=2;
+                write16be(wp,green);wp+=2;
+                write16be(wp,blue);wp+=2;
+        }
+        rfbcon_send(rcon,reply,wp-reply);
+}
+
 static void
 clnt_set_pixel_format(RfbConnection *rcon,uint8_t *data,int len) 
 {
@@ -477,7 +787,9 @@ clnt_set_pixel_format(RfbConnection *rcon,uint8_t *data,int len)
 	pixf->green_shift = data[15];
 	pixf->blue_shift = data[16];
 	pixfmt_update_bits(pixf);
-
+        if((!pixf->true_color_flag) && (pixf->bits_per_pixel == 8)) {
+                srv_set_8bit_color_map_entries(rcon,pixf);
+        }
 #if 1
 	dbgprintf("got set pixelformat message\n");
 	dbgprintf("msgtype %02x\n",data[0]);
@@ -493,6 +805,7 @@ clnt_set_pixel_format(RfbConnection *rcon,uint8_t *data,int len)
 	dbgprintf("blueshift  %d\n",data[16]);
 #endif
 }
+
 static void
 clnt_set_encodings(RfbConnection *rcon,uint8_t *data,int len) 
 {
@@ -504,7 +817,15 @@ clnt_set_encodings(RfbConnection *rcon,uint8_t *data,int len)
 		enc = ntohl(encodings[i]);
 		dbgprintf("enc %d\n",enc);
 		switch(enc) {
+			case ENC_ZRLE:
+				rcon->current_encoding = ENC_ZRLE;
+				dbgprintf("Switched to encoding %d\n",rcon->current_encoding);
+				return;
+
 			case ENC_RAW:	
+				rcon->current_encoding = ENC_RAW;
+				return;
+
 			case ENC_COPYRECT:
 			case ENC_RRE:
 			case ENC_CORRE:
@@ -512,7 +833,7 @@ clnt_set_encodings(RfbConnection *rcon,uint8_t *data,int len)
 			case ENC_ZLIB:
 			case ENC_TIGHT:
 			case ENC_ZLIBHEX:
-			case ENC_ZRLE:
+				
 			case ENC_CURSOR_PSEUDO:
 			case ENC_DESKTOP_SIZE:
 			default:
@@ -522,19 +843,46 @@ clnt_set_encodings(RfbConnection *rcon,uint8_t *data,int len)
 	}
 }
 
+static void
+write_udrect_to_fifo(RfbConnection *rcon,UpdateRectangle *udrect) 
+{
+	UpdateRectangle *u;
+	if((rcon->udrect_wp - rcon->udrect_rp + 1) >= UDRECT_FIFOSIZE) {
+		/* Fifo full, queue a complete screen update instead */
+		FrameBufferInfo *fbi = rcon->fbi;
+		rcon->udrect_wp = rcon->udrect_rp;
+		u = &rcon->udrect_fifo[rcon->udrect_wp % UDRECT_FIFOSIZE];
+		u->x = 0; u->y = 0;
+		u->width = fbi->fb_width; u->height = fbi->fb_height;
+		rcon->udrect_wp++;
+	} else {
+		u = &rcon->udrect_fifo[rcon->udrect_wp % UDRECT_FIFOSIZE];
+		*u = *udrect;
+		rcon->udrect_wp++;
+	}
+}
+/*
+ * Update request from Client
+ */
 static void 
 clnt_fb_update_req(RfbConnection *rcon,uint8_t *data,int len) 
 {
 	uint8_t incremental;
-	UpdateRequest udrq;
+	UpdateRectangle udrect;
 	incremental = data[1];
-	udrq.x=read16be(data+2);
-	udrq.y=read16be(data+4);
-	udrq.width = read16be(data+6);
-	udrq.height = read16be(data+8);
+	udrect.x=read16be(data+2);
+	udrect.y=read16be(data+4);
+	udrect.width = read16be(data+6);
+	udrect.height = read16be(data+8);
+	//fprintf(stderr,"Got updaterequest from Client\n");
 	if(!incremental) {
-		dbgprintf("UDRQ x %d y %d w %d h %d\n",udrq.x,udrq.y,udrq.width,udrq.height);
-		srv_fb_update(rcon,&udrq) ;
+		write_udrect_to_fifo(rcon,&udrect);
+		rcon->update_outstanding = 1;
+		trigger_fb_update(rcon);
+	} else {
+		//fprintf(stderr,"INCREMENTAL UDRQ x %d y %d w %d h %d\n",udrect.x,udrect.y,udrect.width,udrect.height);
+		rcon->update_outstanding = 1;
+		trigger_fb_update(rcon);
 	}
 }
 
@@ -661,6 +1009,10 @@ rfbcon_handle_message(RfbConnection *rcon)
 		case CONSTAT_SHARED_WAIT:
 			// handle shared flag is missing here (close other cons)
 			Msg_ServerInitialisation(rcon);
+			/* 
+		         * Clear Incomming buffer before going to idle to remove
+			 * outstanding requests with wrong depth ????
+			 */
 			rcon->state = CONSTAT_IDLE;
 			rcon->ibuf_expected = 4;
 			break;
@@ -750,6 +1102,20 @@ rfbsrv_accept(int fd,char *host,unsigned short port, void *clientData)
 	rcon->rfbserv = rfbserv;
 	rcon->next = rfbserv->con_head;
 	rcon->fbi = &rfbserv->fbi;
+	rcon->obuf_size = 65536;
+	rcon->obuf = malloc(rcon->obuf_size);
+	if(!rcon->obuf) {
+		fprintf(stderr,"RFB: Out of memory allocating output buffer\n");
+		exit(1);
+	}
+
+	rcon->zs.zalloc = Z_NULL;
+	rcon->zs.zfree = Z_NULL;
+	rcon->zs.opaque = Z_NULL;
+	 if (deflateInit(&rcon->zs, Z_DEFAULT_COMPRESSION) != Z_OK) {
+		fprintf(stderr,"Decompressor init failed\n");
+	}
+
 	/* Default Pixel format is framebuffers pixelformat */
 	memcpy(&rcon->pixfmt,&rfbserv->fbi.pixfmt,sizeof(PixelFormat));
 	rfbserv->con_head = rcon;
@@ -779,6 +1145,7 @@ rfbsrv_accept(int fd,char *host,unsigned short port, void *clientData)
 static void 
 rfbserv_set_fbformat(struct FbDisplay *fbdisp,FbFormat *fbf) 
 {
+	int fb_size;
 	RfbServer *rfbserv = fbdisp->owner;
 	FrameBufferInfo *fbi = &rfbserv->fbi;	
 	PixelFormat *pixf = &fbi->pixfmt;
@@ -795,6 +1162,21 @@ rfbserv_set_fbformat(struct FbDisplay *fbdisp,FbFormat *fbf)
 	pixf->blue_bits = fbf->blue_bits;
 	pixf->blue_shift = fbf->blue_shift;
 
+	/* Should come from the LCD controller */
+	pixf->bits_per_pixel = fbf->bits_per_pixel;
+	pixf->bypp = (pixf->bits_per_pixel + 7) / 8; 
+	pixf->depth = fbf->depth;
+	pixf->true_color_flag = 1;
+	fb_size = fbi->fb_width*fbi->fb_height*pixf->bypp;
+	if(fb_size != fbi->fb_size) {
+		fbi->fb_size = fb_size;
+		fbi->framebuffer = realloc(fbi->framebuffer,fbi->fb_size);
+		if(!fbi->framebuffer) {
+			fprintf(stderr,"OOM: realloc of framebuffer %d bytes failed\n",fbi->fb_size);
+			exit(1);
+		}
+	}
+
 }
 
 /*
@@ -805,35 +1187,48 @@ rfbserv_set_fbformat(struct FbDisplay *fbdisp,FbFormat *fbf)
  * ----------------------------------------------------------------
  */
 static int 
-rfbserv_update_display(struct FbDisplay *fbdisp,FbUpdateRequest *fbudrq) 
+rfbserv_update_display(struct FbDisplay *fbdisp,FbUpdateRequest *fbudrect) 
 {
 	RfbServer *rfbserv = fbdisp->owner;
 	RfbConnection *rcon;
-	UpdateRequest udrq;
+	UpdateRectangle udrect;
 	FrameBufferInfo *fbi = &rfbserv->fbi;
-	unsigned int start = fbudrq->offset;
-	unsigned int count = fbudrq->count;
+	unsigned int start = fbudrect->offset;
+	unsigned int count = fbudrect->count;
 	if(start > fbi->fb_size) {
 		return -1;
 	} 
 	if(start + count > fbi->fb_size) {
 		count = fbi->fb_size - start;
 	}
-	memcpy(fbi->framebuffer+start,fbudrq->fbdata,count);
-	//fprintf(stderr,"RFBserver received update request\n");
-	udrq.x = 0;
-	udrq.y = ((start / fbi->fb_width)>>1);
-	udrq.width = fbi->fb_width; 
-	udrq.height = (count+fbi->fb_linebytes-1)/(fbi->fb_linebytes); 
-	if(udrq.height+udrq.y > fbi->fb_height) {
-		udrq.height =  fbi->fb_height - udrq.y;	
+	/* 
+ 	 * Would be better to check the differences before overwriting 
+	 * also comparing differences should be delayed to time of sending the 
+	 * the data in a fileevent handler
+ 	 */
+	memcpy(fbi->framebuffer+start,fbudrect->fbdata,count);
+	udrect.x = 0;
+	udrect.y = ((start / fbi->fb_width)>>1);
+	udrect.width = fbi->fb_width; 
+	udrect.height = (count+fbi->fb_linebytes-1)/(fbi->fb_linebytes); 
+	if(udrect.height+udrect.y > fbi->fb_height) {
+		udrect.height =  fbi->fb_height - udrect.y;	
 	}
+	//fprintf(stderr,"Got update request from LCD controller\n");
 	for(rcon = rfbserv->con_head;rcon;rcon=rcon->next) {
-		srv_fb_update(rcon,&udrq); 
+		write_udrect_to_fifo(rcon,&udrect);
+		trigger_fb_update(rcon);
 	}
 	return -1;
 }
 
+/*
+ * ------------------------------------------------------------------------------------------
+ * RfbServer_New
+ * 	Create a new rfbserver. The rfbserver implements a display, a keyboard and a mouse.
+ * ------------------------------------------------------------------------------------------
+ */ 
+
 void
 RfbServer_New(const char *name,FbDisplay **display,Keyboard **keyboard,void **mouse)
 {
@@ -882,7 +1277,15 @@ RfbServer_New(const char *name,FbDisplay **display,Keyboard **keyboard,void **mo
 	pixf->bits_per_pixel = 16;
 	pixf->bypp = 2; 
 	pixf->depth = 12;
-	pixf->big_endian_flag = 0;
+		
+	if(HOST_BYTEORDER == en_BIG_ENDIAN) {
+		pixf->big_endian_flag = 1;
+	} else if(HOST_BYTEORDER == en_LITTLE_ENDIAN) {
+		pixf->big_endian_flag = 0;
+	} else {
+		fprintf(stderr,"No byteorder detected\n");
+		exit(1);
+	}
 	pixf->true_color_flag = 1;
 	pixf->red_max = 15;
 	pixf->blue_max = 15;
@@ -894,6 +1297,7 @@ RfbServer_New(const char *name,FbDisplay **display,Keyboard **keyboard,void **mo
 
         if((fd=FIO_InitTcpServer(&rfbserv->tcpserv,rfbsrv_accept,rfbserv,host,port))<0) {
                 free(rfbserv);
+		fprintf(stderr,"Can not create RFB server\n");
                 return;
         }
         fprintf(stderr,"RFB Server Listening on host \"%s\" port %d\n",host,port);
