@@ -196,6 +196,7 @@ typedef struct STE10_100 {
 	int ether_fd;	// file descriptor for the data
 	FIO_FileHandler input_fh;
 	int receiver_is_enabled;
+	SigNode *irqNode[4];
 	int interrupt_posted;
 
 	uint16_t device_id;
@@ -1571,6 +1572,18 @@ STE_ConfigWrite(uint32_t value,uint32_t mask,uint8_t reg,PCI_Function *pcifunc)
                         if(mask&0xff) {
                                 ste->irq_line=value&0xff;
                         }
+			if(mask & 0xff00) {
+				int pin = (value >> 8) & 3;
+				if(pin != ste->irq_pin) {
+					SigNode_Set(ste->irqNode[ste->irq_pin],SIG_OPEN);
+					if(ste->interrupt_posted) {
+						SigNode_Set(ste->irqNode[pin],SIG_LOW);
+					} else {
+						SigNode_Set(ste->irqNode[pin],SIG_PULLUP);
+					}	
+					ste->irq_pin = pin;
+				}
+			}
                         break;
 		case 0x10: 
 			MODIFY_MASKED(ste->driver_space,value,mask & 0xffff);
@@ -1782,6 +1795,7 @@ read_eeprom(STE10_100 *ste)
 PCI_Function *
 STE_New(const char *devname,PCI_Function *bridge,int dev_nr,int bus_irq) {
         STE10_100 *ste;
+	int i;
 	char *eepromname=alloca(strlen(devname)+20);
 	SigNode *mw_sclk;
 	SigNode *mw_sdi;
@@ -1826,6 +1840,13 @@ STE_New(const char *devname,PCI_Function *bridge,int dev_nr,int bus_irq) {
 		fprintf(stderr,"STE: can not connect to microwire eeprom\n");
 		exit(1);
 	}
+	for(i=0;i<4;i++) {
+		ste->irqNode[i] = SigNode_New("%s.int%c",devname,'A'+i);
+		if(!ste->irqNode[i]) {
+			fprintf(stderr,"STE: Can not create irqline \n");
+			exit(1);
+		}
+	}
 	SigNode_Link(ste->edo,mw_sdo);
 	SigNode_Link(ste->edi,mw_sdi);
 	SigNode_Link(ste->eck,mw_sclk);
