@@ -45,14 +45,14 @@
 
 struct OhciHC {
 	BusDevice bdev;		// Own function
-	BusDevice *bus; 	// Bus	
+	Bus *bus; 		// Bus	
 
+	SigNode *irqNode;
 	SigNode *endianNode;
         SigTrace *endianTrace;
         int endian;
 
 	//UsbDevice roothub;
-	int bus_irq;
 	int hold_reset; 	// if true all registers have default values and write is ignored 
 
 	uint32_t HcControl;
@@ -155,12 +155,12 @@ update_interrupts(OhciHC *hc) {
 	uint32_t ist=hc->HcInterruptStatus;
 	if((ist & ien) && (ien & HCInterrupt_MIE)) {
 		if(!hc->interrupt_posted) {
-			hc->bus->postIRQ(hc->bus,hc->bus_irq);
+			SigNode_Set(hc->irqNode,SIG_LOW);
 			hc->interrupt_posted=1;
 		}
 	} else {
 		if(hc->interrupt_posted) {
-			hc->bus->unPostIRQ(hc->bus,hc->bus_irq);
+			SigNode_Set(hc->irqNode,SIG_PULLUP);
 			hc->interrupt_posted=0;
 		}
 	}
@@ -711,7 +711,7 @@ HcControl_write(void *clientData,uint32_t value,uint32_t address,int rqlen) {
 	if(oldFState==HcFunctionalStateUsbReset && ((oldFState!=fState))) {
 		int ms = CyclesToMilliseconds(CycleCounter_Get() - hc->UsbEnterResetTime);
 		if(ms<10) {
-			fprintf(stderr,"OHCI Warning: UsbResetState was not kept 10ms CPU-Cycles: %lld\n",CycleCounter_Get() - hc->UsbEnterResetTime);
+			fprintf(stderr,"OHCI Warning: UsbResetState was not kept 10ms CPU-Cycles: %lld\n",(unsigned long long)(CycleCounter_Get() - hc->UsbEnterResetTime));
 		}
 	}
 	/* 
@@ -1215,20 +1215,17 @@ change_endian(SigNode *node,int value,void *clientData)
  *---------------------------------------------------------------
  */
 BusDevice *
-OhciHC_New(char *name,BusDevice *bus,int irq) {
+OhciHC_New(char *name,Bus *bus) {
 	OhciHC *hc=malloc(sizeof(OhciHC));	
 	if(!hc) {
 		fprintf(stderr,"Out of memory\n");
 		exit(3456);
 	}
 	memset(hc,0,sizeof(OhciHC));
-	if(!bus->postIRQ || !bus->unPostIRQ) {
-		fprintf(stderr,"Error: OHCI connected to bus without Interrupts\n");
-		exit(4378);
-	}
-	if(!bus->read32 || !bus->write32) {
-		fprintf(stderr,"Error: OHCI connected to bus without Master read/write access\n");
-		exit(4379);
+	hc->irqNode = SigNode_New("%s.irq",name);
+	if(!hc->irqNode) {
+		fprintf(stderr,"Can not create irq signal\n");
+		exit(1);
 	}
 
 	hc->endian = en_LITTLE_ENDIAN;
@@ -1245,7 +1242,7 @@ OhciHC_New(char *name,BusDevice *bus,int irq) {
         hc->bdev.owner=hc;
         hc->bdev.hw_flags=MEM_FLAG_WRITABLE|MEM_FLAG_READABLE;
 	hc->bus=bus;
-	hc->bus_irq=irq;
 	HcHardReset(hc);
+	fprintf(stderr,"USB-OHCI module created\n");
 	return &hc->bdev;
 }
